/* algebra.c */
#include <math.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "globals.h"

float epsilon=0.005;

void
print_x(float*x,int nx,int n)
{
  int j,k;
  for(j=0;j<nx;j++)
  {
    float*_x=x+n*j;
    fprintf(logfile,"(");
    for(k=0;k<n;k++)
      fprintf(logfile,"%.1f,",_x[k]);
    fprintf(logfile,")");
  }
  fprintf(logfile,"\n");
}  

void
add( float*a,float*b,float*c)
{
  int i;
  for(i=0;i<3;i++)
    c[i]=a[i]+b[i];
}
 
void minus(
  float*a,
  float*b,
  float*c)
{
  int i;
  for(i=0;i<3;i++)
    c[i]=a[i]-b[i];
}

float dot3(
  float*a,
  float*b)
{
  int i;
  float d=0.0;
  for(i=0;i<3;i++)
    d+=a[i]*b[i];
  return d;
}

float metric3(
  float*a,
  float*b)
{
  float res;
  res=
    fabs(a[0]-b[0])+
    fabs(a[1]-b[1])+
    fabs(a[2]-b[2]);
#ifdef DEBUG_ALGEBRA
  fprintf(logfile,"[[%.1f,%.1f,%.1f|%.1f,%.1f,%.1f|res=%.1f]]",
    a[0],a[1],a[2],b[0],b[1],b[2],res);
#endif
  return res;
}

float
xmetric3(float*a,float*b,int nx)
{
  int i;
  float res=0.0;
  for(i=0;i<nx;i++)
  {
    res+=metric3(a+3*i,b+3*i);
#ifdef DEBUG_ALGEBRA
    fprintf(logfile,"{res=%.2f}",res);
#endif
  }
  return res;
}

int imetric3(
  int*a,int*b)
{
  return abs(a[0]-b[0])+abs(a[1]-b[1])+abs(a[2]-b[2]);
}

float dmetric2(float*a,float*b)
{
  return sqrt((a[1]-b[1])*(a[1]-b[1])+(a[0]-b[0])*(a[0]-b[0]));
}

void print_rot(float*r)
{
  int i,j;
  for(i=0;i<3;i++)
    for(j=0;j<3;j++)
    {
      fprintf(logfile,"%.2f ",r[i*3+j]);
      if(j==2) 
        fprintf(logfile,", ");
    }
  fprintf(logfile,"\n");
  fflush(logfile);
}

void rotate(
  float *r, /* float r[9]; */
  float*x,  /* float x[3]; */
  float*_x /* float _x[3]; */ /* return */ )
{
  /* x==_x okay */
  int i,j;
  float rx[3];
  for(i=0;i<3;i++)
  {
    rx[i]=0;
    for(j=0;j<3;j++)
      rx[i]+=r[i*3+j]*x[j];
  }
  for(i=0;i<3;i++)
  {
    _x[i]=rx[i];
  }
}

void lrotate(
  float*x,  /* float x[3]; */
  float *r, /* float r[9]; */
  float*_x /* float _x[3]; */ /* return */ )
{
  /* x==_x okay */
  int i,j;
  float rx[3];
  for(i=0;i<3;i++)
  {
    rx[i]=0;
    for(j=0;j<3;j++)
      rx[i]+=r[i+j*3]*x[j];
  }
  for(i=0;i<3;i++)
  {
    _x[i]=rx[i];
  }
}

void mult(float *a,float*b,float*c)
{
  /* float a[9], b[9], c[9]; */
  /* return ab in c */
  /* a==c || b==c okay */
  int i,j,k;
  float rx[3][3];
  for(i=0;i<3;i++)
    for(j=0;j<3;j++)
    {
      rx[i][j]= 0.0;
      for(k=0;k<3;k++)
        rx[i][j]+=a[3*i+k]*b[3*k+j];
    }
  for(i=0;i<3;i++)
    for(j=0;j<3;j++)
      c[3*i+j]=rx[i][j];
}

void conj(float *a,float*b,float*c)
{
  /* float a[9],b[9],c[9]; */
  /* return trans(b)ab in c */
  /* a==c || b==c okay */
  float rx[9],_rx[9];
  int i,j,k;
  mult(a,b,rx);
  for(i=0;i<3;i++)
    for(j=0;j<3;j++)
    {
      _rx[3*i+j]=0;
      for(k=0;k<3;k++)
        _rx[3*i+j]+=b[3*k+i]*a[3*k+j];
    }
  memcpy(c,_rx,sizeof(float)*9);
}

void trans(float *a,float*b)
{
  /* float a[9],float b[9]; */
  /* put transpose of a into b */
  /* a==b okay */
  int i,j;
  for(i=0;i<3;i++)
    for(j=0;j<3;j++)
      b[3*i+j]=a[i+3*j];
}

void
ortho(float*a)
{
  /* float a[9]; */
  /* orthogonalize a, we use Gram-Schmidt */
  float r,_r;
  int i;

  r=0.0;
  for(i=0;i<3;i++) r+=a[i]*a[i];
  r=sqrt(r);
  assert((r>epsilon)||(r<-epsilon));
  for(i=0;i<3;i++) a[i]/=r;

  r=0.0;
  for(i=0;i<3;i++) r+=a[3+i]*a[i];
  for(i=0;i<3;i++) a[3+i]-=a[i]*r;
  r=0.0;
  for(i=0;i<3;i++) r+=a[3+i]*a[3+i];
  r=sqrt(r);
  assert((r>epsilon)||(r<-epsilon));
  for(i=0;i<3;i++) a[3+i]/=r;

  r=0.0; _r=0.0;
  for(i=0;i<3;i++) { r+=a[6+i]*a[i]; _r+=a[6+i]*a[3+i]; }
  for(i=0;i<3;i++) { a[6+i]-=r*a[i]+_r*a[3+i]; }
  r=0.0;
  for(i=0;i<3;i++) r+=a[6+i]*a[6+i];
  r=sqrt(r);
  assert((r>epsilon)||(r<-epsilon));
  for(i=0;i<3;i++) a[6+i]/=r;
}

void z_rot(float*r,float theta)
{
  /* float r[9] */
  assert(r);
  r[0]=cos(theta);
  r[1]=-sin(theta);
  r[3]=sin(theta);
  r[4]=cos(theta);
  r[2]=0.0;
  r[5]=0.0;
  r[6]=0.0;
  r[7]=0.0;
  r[8]=1.0;
}

void y_rot(float*r,float theta)
{
  /* float r[9] */
  assert(r);
  r[0]=cos(theta);
  r[6]=-sin(theta);
  r[2]=sin(theta);
  r[8]=cos(theta);
  r[1]=0.0;
  r[5]=0.0;
  r[3]=0.0;
  r[7]=0.0;
  r[4]=1.0;
}

void x_rot(float*r,float theta)
{
  /* float r[9] */
  assert(r);
  r[4]=cos(theta);
  r[5]=-sin(theta);
  r[7]=sin(theta);
  r[8]=cos(theta);
  r[2]=0.0;
  r[1]=0.0;
  r[3]=0.0;
  r[6]=0.0;
  r[0]=1.0;
}

