#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <string.h>
#include "architect.h"
#include "scene.h"
#include "people.h"
#include "map.h"
#include "cam.h"
#include "cursor.h"
#include "command.h"
#include "main.h"

void
architect_init(struct architect*ar,int edit,char*filename)
{
  int bi=30, bj=20, bk=11;
  FILE*stream;
  int c;
  char s[256];
  int count;

  assert(filename);
  strncpy(ar->filename,filename,255);
  filename[255]=0;
  ar->edit=edit;

  /** open the file */
  if(filename[0])
    stream=fopen(ar->filename,"r");
  else
  {
    fprintf(logfile,"no file to edit...\n");
    ar->filename[0]=0;
    return;
  }

  if(stream==NULL&&edit)
  {
    /* make new file */
    stream=fopen(ar->filename,"w+");
    fprintf(logfile,"new file \"%s\"\n",ar->filename);
  }

  if(stream==NULL)
  {
    fprintf(logfile,"could not open \"%s\"\n",ar->filename);
    main_set_quit();
    return;
  }
  else
    printf( 
      " architect_init: opened file \"%s\" \n", 
      ar->filename );

  /* read map size */
  fgets(s,256,stream);
  count=sscanf(s,"%d %d %d",&bi,&bj,&bk);
  if(count!=3)
  {
    printf("default map size\n");
  }
  printf("architect_init: map size %d,%d,%d\n",
    bi,bj,bk);

  map_init(&ar->m,&ar->s,bi,bj,bk);
  cam_init(&ar->c);
  cam_center(&ar->c,ar->m.bi,ar->m.bj,ar->m.bk);
  scene_init(&ar->s,&ar->m,&ar->c);
  people_init(&ar->pl,&ar->s,&ar->m);
  ar->cr=NULL;
  ar->cr=(struct cursor*)xmalloc(sizeof(struct cursor));
  cursor_init(ar->cr,&ar->s,&ar->m,&ar->pl,bi/2,bj/2,bk/2);
  command_init(&ar->cmd,ar);

  if(edit)
  {
    /** edit mode */
    command_set_record(&ar->cmd);
  }

  /* read commands */
  for(c=getc(stream);c!=EOF;c=getc(stream))
  {
#ifdef DEBUG_ARCHITECT
    printf("dispatch %c\n",c);
#endif
    command_dispatch(&ar->cmd,c);
  }
  fclose(stream);

  if(!edit)
  {
    /** game play mode */
    map_compile(&ar->m);
    cursor_finalize(ar->cr);
    xfree(ar->cr);
    ar->cr=NULL;
  }
  main_set_draw();
}

void
architect_finalize(struct architect*ar)
{
  if(ar->edit)
    if(ar->filename[0])
    {
      /* write to file */
      int fd;
      char s[256];
      if(unlink(ar->filename)<0)
        assert(0);
      fd=open(ar->filename,O_WRONLY|O_CREAT,0644);
      if(fd<0)
        assert(0);
      /* write map size */
      snprintf(s,256,"%d %d %d\n",
        ar->m.bi,ar->m.bj,ar->m.bk);
      write(fd,s,strlen(s)); 
      command_write(&ar->cmd,fd);
      close(fd);
    }
  command_finalize(&ar->cmd);
  if(ar->cr)
  {
    cursor_finalize(ar->cr);
    xfree(ar->cr);
    ar->cr=NULL;
  }
  people_finalize(&ar->pl);
  map_finalize(&ar->m);
  scene_finalize(&ar->s);
}

