#include <stdio.h>
#include <math.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>

#include <SDL/SDL.h>

#include "memory.h"

static int _audio; /* if audio is alive */

static int _dig,_fall,_caught,_parcel,_success,_finali;
static Sint16* adat=NULL;
static int volume=1000;
int fd=0;

void
_dig_packet(short*stream,int len,int bsz)
{
  float amp=0,damp = (float)volume/(float)len;
  int i;
  for(i=0;i<len;i++)
  {
    stream[i]=((short)amp)*(2*(i%bsz)-1)/16;
    amp+=damp;
    if(amp>volume)
      damp=-damp;
  }
}

void
_parcel_packet(short*stream,int len,int bsz)
{
  float amp=0,damp = (float)volume/(float)len;
  int i;
  for(i=0;i<len;i++)
  {
    stream[i]=((short)amp)*(2*(i%bsz)-1)/16;
    amp+=damp;
    if(amp>volume)
      damp=-damp;
  }
}

void
_fd_packet(short*stream,int len)
{
  int _len=0,i,rd=0;
  if(fd==0)
  {
    printf("no fd!\n");
    return;
  }
  for(i=0;i<len;i++)
    stream[i]=0;
#define AUDIO
#ifdef AUDIO
  for(i=0;i<len;i++)
  {
    rd=read(fd,stream+i,2);
    if(rd==-1)rd=0;
    if(rd<2) break;
    _len+=rd;
  }
#endif
  if(rd==0 && fd)
  {
    close(fd);
    printf("_fd_packet close i%d\n",fd);
    fd=0;
  }
}

void
_fall_packet(short*stream,int len,int bsz)
{
  int i;
  float amp=0.0,damp;
  float x=0,dx;
  dx=volume/bsz;
  damp=2.0/(float)len;
  for(i=0;i<len;i++)
  {
    stream[i]=(short)(x*amp);
    amp+=damp;
    x+=dx;
    if(amp>1.0)
    {
      damp=-damp;
      amp=1.0;
    }
    else if(amp<-1.0)
    {
      damp=-damp;
      amp=-1.0;
    }
    if(x>volume)
    {
      dx=-dx;
    }
    else if (x<-volume) 
      dx=-dx;
  }
}

void
audio_fill(void *udata, Uint8 *stream, int len)
{
#ifdef DEBUG_AUDIO
#endif
  int i;
  /*printf("audio_fill()\n");*/

  if(_dig)
  {
    /*printf("audio_fill(): dig\n");*/
    if (adat==NULL)
      adat=(short*)xmalloc(len);
    for(i=0;i<len/2;i++)adat[i]=0;
    _dig_packet(adat,len/8,20*_dig);
    _dig++;
    
    SDL_LockAudio(); 
    memcpy(stream,adat,len);
    SDL_UnlockAudio();  
  }
  else if(_fall)
  {
    /*printf("audio_fill(): fall\n");*/
    if (adat==NULL)
      adat=(short*)xmalloc(len);
    for(i=0;i<len/2;i++)adat[i]=0;
    _fall_packet(adat,len/2,_fall);
    _fall++;
    
    SDL_LockAudio(); 
    memcpy(stream,adat,len);
    SDL_UnlockAudio();  
  }
  else if(_caught||_success)
  {
    SDL_LockAudio(); 
    _fd_packet((short*)stream,len/2);
    SDL_UnlockAudio();  
  }
  else if(_parcel)
  {
    printf("audio_fill(): parcel\n");
    if (adat==NULL)
      adat=(short*)xmalloc(len);
    for(i=0;i<len/2;i++)adat[i]=0;
    _parcel_packet(adat,len/2,10);
    
    SDL_LockAudio(); 
    memcpy(stream,adat,len);
    SDL_UnlockAudio(); 
    _parcel=0;
  }
}

void
audio_start_dig()
{
  _dig=1;
  SDL_PauseAudio(0);
  printf("audio_start_dig()\n");
#ifdef DEBUG_AUDIO
#endif
}

void
audio_stop_dig()
{
  _dig=0;
  SDL_PauseAudio(1);
  printf("audio_stop_dig()\n");
}

void
audio_start_fall()
{
  if(_fall) return;
  _fall=1;
  SDL_PauseAudio(0);
  printf("audio_start_fall()\n");
#ifdef DEBUG_AUDIO
#endif
}

void
audio_stop_fall()
{
  _fall=0;
  SDL_PauseAudio(1);
  printf("audio_stop_fall()\n");
}

void
audio_start_caught()
{
#ifdef AUDIO
  if(fd)
    close(fd);
  fd=open("strav.raw",O_RDONLY|O_NONBLOCK);
  printf("audio_start_caught() fd=%d\n",fd);
  _caught=1;  
  SDL_PauseAudio(0);
#endif
}

void
audio_start_success()
{
#ifdef AUDIO
  if(fd)
    close(fd);
  fd=open("success.raw",O_RDONLY|O_NONBLOCK);
  printf("audio_start_success() fd=%d\n",fd);
  _success=1;  
  SDL_PauseAudio(0);
#endif
}

void
audio_start_finali()
{
#ifdef AUDIO
  if(fd)
    close(fd);
  fd=open("finali.raw",O_RDONLY|O_NONBLOCK);
  printf("audio_start_finali() fd=%d\n",fd);
  _finali=1;
  SDL_PauseAudio(0);
#endif
}

void
audio_start_parcel()
{
  printf("audio_start_parcel()\n");
  _parcel=1;
  SDL_PauseAudio(0);
}

void
audio_stop()
{
  printf("audio_stop() fd %d\n",fd);
  _dig=0;
  _fall=0;
  _success=0;
  _finali=0;
  SDL_PauseAudio(1);
  if(fd)
  {
    printf("close %d\n",fd);
    close(fd);
    fd=0;
  }
}

void
audio_init()
{
  SDL_AudioSpec a_spec;
  SDL_AudioSpec _a_spec;
  _a_spec.freq = 22050;
  _a_spec.format = AUDIO_S16;
  _a_spec.channels = 1; 
  _a_spec.samples = 4096;  
  _a_spec.samples = 1024;  
  _a_spec.callback = audio_fill;
  _a_spec.userdata = NULL;

  _audio=0;
  if ( SDL_OpenAudio(&_a_spec, &a_spec) < 0 )
  {
     fprintf(
       stderr, "Couldn't open audio: %s\n", SDL_GetError());
     return;
  }
#ifdef DEBUG_AUDIO
  printf("audio_init()\n");
  printf("  freq %d\n  format %d\n  channels %d\n",
    a_spec.freq,
    a_spec.format|AUDIO_S16,
    a_spec.channels);
  printf("  bufsiz %d \n", a_spec.samples);
#endif
  if(a_spec.channels!=1)
  {
    printf("  can't get mono: closing device\n");
    SDL_CloseAudio();
  }
  else if(!(a_spec.format|AUDIO_S16LSB))
  {
    printf("  can't get 16 bit audio: closing device\n");
    SDL_CloseAudio();
  }
  _dig=0;
  _fall=0;
  SDL_PauseAudio(1);
  _audio=1;
  printf("audio_init(): %d\n",_audio);

}

void
audio_finalize()
{
  SDL_CloseAudio();
  if(adat)
    xfree(adat);
  printf("audio_finalize()\n");
#ifdef DEBUG_AUDIO
#endif
}


