#include <stdio.h>
#include "memory.h"
#include "bsp.h"
#include "globals.h"
#include "scene.h"
#include "render.h"
#include "cam.h"

struct bsp*
_bsp_alloc()
{
  return (struct bsp*)xcalloc(sizeof(struct bsp));
}

struct bsp*
bsp_alloc_init(int l,int a)
{
  struct bsp*p;
  assert(l==0||l==1||l==2);
  p=_bsp_alloc();
  p->l=l;
  p->a=a;
  p->_dl=dlink_alloc_init();
  p->dl=dlink_alloc_init();
  p->dl_=dlink_alloc_init();
  assert(dlink_is_empty(p->_dl));
  assert(dlink_is_empty(p->dl));
  assert(dlink_is_empty(p->dl_));
  assert(p->_p==NULL);
  assert(p->p_==NULL);
#ifdef DEBUG_BSP
  fprintf(logfile,"bsp_alloc_init( l=%d, a=%d)=%p\n", l,a,p );
#endif
  return p;
}

void
bsp_finalize_free(struct bsp*);

void
_bsp_finalize(struct bsp*p)
{
  /** return to state after _bsp_alloc */
  assert(p);
  p->l=0;
  p->a=0;
  if(p->_p)
    bsp_finalize_free(p->_p);
  p->_p=NULL;
  if(p->p_)
    bsp_finalize_free(p->p_);
  p->p_=NULL;
  dlink_free(p->_dl);
  p->_dl=NULL;
  dlink_free(p->dl);
  p->dl=NULL;
  dlink_free(p->dl_);
  p->dl_=NULL;
}

void
_bsp_free(struct bsp*p)
{
  assert(p);
  xfree(p);
}

void
bsp_finalize_free(struct bsp*p)
{
  assert(p);
  _bsp_finalize(p);
  _bsp_free(p);
}

float
_bsp_test(struct bsp*p,float x[3])
{
  assert(p);
  assert(x);
  return x[p->l]-((float)p->a);
}

#define IS_BEHIND(x)	((x)<-0.3)
#define IS_INFRONT(x)	((x)>0.3)

struct bsp*
__bsp_separate(struct draw_item*d,struct draw_item*_d)
{
  float x[3],_x[3];
  float a;
  int i;
  assert(d);
  assert(_d);
#ifdef DEBUG_BSP
  fprintf(logfile,"_bsp_separate %p, %p\n", d,_d);
#endif
  for(i=0;i<3;i++)
  {
    x[i]=0;
    _x[i]=0;
  }

  for(i=0;i<d->nx;i++)
  {
    x[0]+=d->so->x[3*i];
    x[1]+=d->so->x[3*i+1];
    x[2]+=d->so->x[3*i+2];
  }
  for(i=0;i<3;i++)
    x[i]/=d->nx;

  for(i=0;i<_d->nx;i++)
  {
    _x[0]+=_d->so->x[3*i];
    _x[1]+=_d->so->x[3*i+1];
    _x[2]+=_d->so->x[3*i+2];
  }
  for(i=0;i<3;i++)
    _x[i]/=_d->nx;

  {
    float dx=fabs(x[0]-_x[0]);
    int l=0;
    struct bsp*p;
    float test, _test;
    if(fabs(x[1]-_x[1])>dx)
    {
      l=1;
      dx=fabs(x[1]-_x[1]);
    }
    if(fabs(x[2]-_x[2])>dx)
      l=2;
    a=(x[l]+_x[l])/2.0;
    p=bsp_alloc_init(l,a);
    test=_bsp_test(p,x);
    _test=_bsp_test(p,_x);
#ifdef DEBUG_BSP
    fprintf(logfile," separation=%.2f, test(%p)=%.2f, test(%p)=%.2f\n",
      dx, d, test, _d, _test);
#endif
    if( IS_BEHIND(test)!=IS_BEHIND(_test) 
      || IS_INFRONT(test)!=IS_INFRONT(_test) )
      return p;
    bsp_finalize_free(p);
    return NULL;
  }
}

struct bsp*
_bsp_separate(struct dlink*dl)
{
  struct dlink*_dl,*__dl;
  assert(dl);
  assert(dl->next);
  assert(dl->next->next);
  /** check each pair */
  for(_dl=dl;_dl->next;_dl=_dl->next)
    for(__dl=_dl->next;__dl->next;__dl=__dl->next)
    {
      struct bsp*p;
      p=__bsp_separate(_dl->next->d,__dl->next->d);
      if(p)
        return p;
    }
  return NULL;
}

int _insert_depth=0;
void
bsp_insert_draw_item(struct bsp*p,struct draw_item*d)
{
  float x=0.0;
  int i;
  assert(p);
  assert(d);
  if(_insert_depth>100)
  { fflush(logfile); exit(-1); }
  _insert_depth++;
#ifdef DEBUG_BSP
  fprintf(logfile,"<%d>",_insert_depth);
  fprintf(logfile,"bsp_insert_draw_item %p, %p\n", p, d);
#endif
  for(i=0;i<d->nx;i++)
    x+=_bsp_test(p,d->so->x+3*i);
  x/=(float)d->nx;
  if(fabs(x)<0.3)
    /** item is on */
    dlink_push(p->dl,d);
  else
  if(x<-0.3)
  {
    /** item is behind */
    if(p->_p)
      bsp_insert_draw_item(p->_p,d);
    else
    if(dlink_is_empty(p->_dl))
      dlink_push(p->_dl,d);
    else
    {
      struct bsp*_p;
      dlink_push(p->_dl,d);
      _p=_bsp_separate(p->_dl);
      if(_p)
      {
        p->_p=_p;
        while(!dlink_is_empty(p->_dl))
          bsp_insert_draw_item(p->_p,dlink_pop(p->_dl));
      }
    }
  }
  else /* x>0.3 */
  {
    /** item is in front */
    if(p->p_)
      bsp_insert_draw_item(p->p_,d);
    else
    if(dlink_is_empty(p->dl_))
      dlink_push(p->dl_,d);
    else
    {
      struct bsp*p_;
      dlink_push(p->dl_,d);
      p_=_bsp_separate(p->dl_);
      if(p_)
      {
        p->p_=p_;
        while(!dlink_is_empty(p->dl_))
          bsp_insert_draw_item(p->p_,dlink_pop(p->dl_));
      }
    }
  }
  _insert_depth--;
}

int _remove_depth=0;
int
bsp_remove_draw_item(struct bsp*p,struct draw_item*d)
{
  float x=0.0;
  int i;
  int found=0;
  assert(p);
  assert(d);
  _remove_depth++;
#ifdef DEBUG_BSP
  fprintf(logfile,"<%d>",_remove_depth);
  fprintf(logfile,"bsp_remove_draw_item %p, %p\n", p, d);
#endif
  for(i=0;i<d->nx;i++)
    x+=_bsp_test(p,d->so->x+3*i);
  x/=(float)d->nx;
  if(fabs(x)<0.3)
    /** item is on */
    found=dlink_remove(p->dl,d);
  else
  if(x<-0.3)
  {
    /** item is behind */
    if(p->_p)
      found=bsp_remove_draw_item(p->_p,d);
    else
      found=dlink_remove(p->_dl,d);
  }
  else /* x>0.3 */
  {
    /** item is in front */
    if(p->p_)
      found=bsp_remove_draw_item(p->p_,d);
    else
      found=dlink_remove(p->dl_,d);
  }
  _remove_depth--;
  return found;
}

#define RENDER_IT(dl)  \
        if(!(dl)->next->d->im->skip) \
          (*(dl)->next->d->so->render) \
            ((dl)->next->d->im->x,(dl)->next->d->so->c)

void
bsp_draw_scene(struct bsp*p, struct cam*c)
{
  float x;
  struct dlink*dl;
  assert(p);
  assert(c);
  x=_bsp_test(p,c->s);
  if(x>0.0)
  {
    /** behind */
    if(p->_p)
      bsp_draw_scene(p->_p,c);
    else
      for(dl=p->_dl;dl->next;dl=dl->next)
        RENDER_IT(dl);
    /** on */
    for(dl=p->dl;dl->next;dl=dl->next)
      RENDER_IT(dl);
    /** in front */
    if(p->p_)
      bsp_draw_scene(p->p_,c);
    else
      for(dl=p->dl_;dl->next;dl=dl->next)
        RENDER_IT(dl);
  }
  else
  {
    /** in front */
    if(p->p_)
      bsp_draw_scene(p->p_,c);
    else
      for(dl=p->dl_;dl->next;dl=dl->next)
        RENDER_IT(dl);
    /** on */
    for(dl=p->dl;dl->next;dl=dl->next)
      RENDER_IT(dl);
    /** behind */
    if(p->_p)
      bsp_draw_scene(p->_p,c);
    else
      for(dl=p->_dl;dl->next;dl=dl->next)
        RENDER_IT(dl);
  }
}







