#ifndef __CAM_H
#define __CAM_H

#include "map.h"
#include "algebra.h"
#include "render.h"
#include "memory.h"
#include "globals.h"

struct source_item;
struct image_item;

struct cam
{
  /* camera position */
  float s[3];
  /* camera orientation */
  float a[9];
  /* zoom */
  float r;
};

void
cam_init(struct cam*c);

void
cam_print(struct cam*c);

void
cam_image_source_item(
  struct cam*c,
  struct image_item*ii,
  struct source_item*si,
  int nx);

void
cam_look_at(
  struct cam*c,
  float*t, /* target */
  float*up /* up vector */ );

void
cam_center(struct cam*c,int bx,int by,int bz);
void
cam_flip(struct cam*,float*);
void
cam_up(struct cam*c);
void
cam_down(struct cam*c);
void
cam_left(struct cam*c);
void
cam_right(struct cam*c);
void
cam_in(struct cam*c);
void
cam_out(struct cam*c);
void
cam_rotate_up(struct cam*c);
void
cam_rotate_down(struct cam*c);
void
cam_rotate_left(struct cam*c);
void
cam_rotate_right(struct cam*c);
void
cam_track_up(struct cam*c,float*,float);
void
cam_track_down(struct cam*c,float*,float);
void
cam_track_left(struct cam*c,float*,float);
void
cam_track_right(struct cam*c,float*,float);
void
cam_zoom_in(struct cam*c);
void
cam_zoom_out(struct cam*c);

#endif

