#include <unistd.h>
#include <stdlib.h>
#include <assert.h>
#include <SDL/SDL.h>
#include "colour.h"

extern SDL_Surface*screen;

struct cmap
{
  Uint8 r,g,b;
  Uint32 c;
};

#define MAX_CMAP 16
struct cinfo
{
  struct cmap cm[MAX_CMAP]; /* array */
  int n;
};

/*
struct cinfo border_cinfo;
struct cinfo brick_cinfo;
struct cinfo solid_cinfo;
struct cinfo parcel_cinfo;
struct cinfo robot_cinfo;
struct cinfo hero_cinfo;
struct cinfo robot_portal_cinfo;
struct cinfo portal_cinfo;
*/

static struct cinfo cin_list[MAX_CID+1];
/* =
{
  border_cinfo,
  brick_cinfo,
  solid_cinfo,
  parcel_cinfo,
  robot_cinfo,
  hero_cinfo,
  robot_portal_cinfo,
  portal_cinfo,
  NULL
};
*/


static
void
colour_err()
{
  printf("colour file error \n");
  exit(0);
}

void
colour_init()
{
  int i,j;
  FILE*stream;

  stream=fopen("colours.ascii","r");
  if(stream==NULL)
    colour_err();
  for(i=0;i<MAX_CID;i++)
  {
    char *s=NULL;
    cin_list[i].n=0;
    for(j=0;j<MAX_CMAP;j++)
    {
      char line[256];
      int r,g,b;
      int d;
#ifdef DEBUG_COLOURS
      printf("[%d,%d]\n",i,j);
#endif
      s=fgets(line,256,stream);
      if(s==NULL) break;
      if(s!=line) colour_err();
      if(line[0]=='%') break;
      d=sscanf(line,"%d %d %d ",&r,&g,&b);
      if(d<3) colour_err();
      if(d<3) break;
      cin_list[i].cm[j].r=r;
      cin_list[i].cm[j].g=g;
      cin_list[i].cm[j].b=b;
      cin_list[i].n++;
#ifdef DEBUG_COLOURS
      printf("  read %d %d %d\n",
        cin_list[i].cm[j].r,
        cin_list[i].cm[j].g,
        cin_list[i].cm[j].b);
      printf("  c=%d\n",cin_list[i].cm[j].c);
#endif
      cin_list[i].cm[j].c
        =SDL_MapRGB( screen->format,
          cin_list[i].cm[j].r,
          cin_list[i].cm[j].g,
          cin_list[i].cm[j].b);
    }
    assert(j<MAX_CMAP);
    if(s==NULL)
      break;
  }
  assert(i==MAX_CID);
  fclose(stream);
  fflush(stdout);
}

Uint32
colour_lookup(int cid,int idx)
{
  assert(cid>=0);
  assert(cid<=MAX_CID);
  assert(0<=idx);
  if(idx>=cin_list[cid].n)
    assert(0);
#ifdef DEBUG_COLOURS
  printf("colour %d\n",cin_list[cid].cm[idx].c);
#endif
  return cin_list[cid].cm[idx].c;
}

void
colour_finalize()
{
}


