#include "command.h"
#include "architect.h"
#include "cursor.h"
#include "map.h"
#include "algebra.h"
#include "main.h"
#include "people.h"
#include "cam.h"

#define CURSOR_UP 		'i'
#define CURSOR_DOWN	 	'm'
#define CURSOR_LEFT		'j'
#define CURSOR_RIGHT		'k'
#define CURSOR_IN		'.'
#define CURSOR_OUT		','
#define CURSOR_TURN_DOWN	'='
#define CURSOR_TURN_LEFT	'['
#define CURSOR_TURN_UP		'\''
#define CURSOR_TURN_RIGHT	']'
#define CURSOR_MODE_SOLID	'`'
#define CURSOR_MODE_BRICK	'1'
#define CURSOR_MODE_PARCEL	'2'
#define CURSOR_MODE_ROBOT	'3'
#define CURSOR_MODE_HERO	'4'
#define CURSOR_MODE_ROBOT_PORTAL '5'
#define CURSOR_MODE_EXIT_PORTAL	'6'
#define CURSOR_TOGGLE_PEN	'e'
#define CAM_UP 			'y'
#define CAM_DOWN		'b'
#define CAM_LEFT		'g'
#define CAM_RIGHT		'h'
#define CAM_IN			'{'
#define CAM_OUT			'}'
#define CAM_ROTATE_UP 		'r'
#define CAM_ROTATE_DOWN		'c'
#define CAM_ROTATE_LEFT		'd'
#define CAM_ROTATE_RIGHT	'f'
#define CAM_FLIP		'E'
#define CAM_ZOOM_IN		')'
#define CAM_ZOOM_OUT		'('
#define CAM_CENTER		'x'
#define QUIT			'Q'

void
command_init(struct command*cmd,struct architect*ar)
{
  cmd->ar=ar;
  cmd->ch=NULL;
  cmd->idx=0;
  cmd->nch=0;
}

void
command_set_record(struct command*cmd)
{
  assert(cmd->ch==NULL);
  cmd->nch=16; /** change this to 4096 */
  /* dynamic array */
  cmd->ch=(char*)xmalloc(sizeof(char)*cmd->nch);
  cmd->idx=0;
}

void
command_write(struct command*cmd,int fd)
{
  int i;
  for(i=0;i<cmd->idx;i++)
    if( write(fd,cmd->ch+i,1) < 1 )
      assert(0);
}

void
command_finalize(struct command*cmd)
{
  if(cmd->ch)
    xfree(cmd->ch);
  cmd->nch=0;
  cmd->ch=NULL;
  cmd->idx=0;
  cmd->ar=NULL;
}

void
_push_command(struct command*cmd,char ch)
{
  if(cmd->nch==0)
    return;
  if(cmd->idx>=cmd->nch-1)
  {
    /* double size */
    cmd->nch*=2;
    cmd->ch=(char*)xrealloc(cmd->ch,cmd->nch);
  }
  assert(cmd->idx<cmd->nch);
  cmd->ch[cmd->idx]=ch;
  cmd->idx++;
  cmd->ch[cmd->idx]=0;
}

int
command_dispatch(struct command*cmd,char ch)
{
  struct architect*ar=cmd->ar;
  switch(ch)
  {
    case PERSON_WALK_FORWARD:
      if(ar->pl.pn[0])
        person_walk_forward(ar->pl.pn[0]);
      break;
    case PERSON_WALK_LEFT:
      if(ar->pl.pn[0])
        person_walk_left(ar->pl.pn[0]);
      break;
    case PERSON_WALK_RIGHT:
      if(ar->pl.pn[0])
        person_walk_right(ar->pl.pn[0]);
      break;
    case PERSON_WALK_BACK:
      if(ar->pl.pn[0])
        person_walk_back(ar->pl.pn[0]);
      break;
    case PERSON_STOP:
      if(ar->pl.pn[0])
        person_stop(ar->pl.pn[0]);
      break;
    case SPACE_BAR:
      if(ar->cr)
      {
        _push_command(cmd,ch);
        cursor_click(ar->cr);
        main_set_draw();
        break;
      }
      if(ar->pl.pn[0])
        person_dig(ar->pl.pn[0]);
      break;
    case CURSOR_UP:
      _push_command(cmd,ch);
      cursor_move(ar->cr,1,0,0);
      main_set_draw();
      break;
    case CURSOR_DOWN:
      _push_command(cmd,ch);
      cursor_move(ar->cr,-1,0,0);
      main_set_draw();
      break;
    case CURSOR_RIGHT:
      _push_command(cmd,ch);
      cursor_move(ar->cr,0,1,0);
      main_set_draw();
      break;
    case CURSOR_LEFT:
      _push_command(cmd,ch);
      cursor_move(ar->cr,0,-1,0);
      main_set_draw();
      break;
    case CURSOR_OUT:
      _push_command(cmd,ch);
      cursor_move(ar->cr,0,0,-1);
      main_set_draw();
      break;
    case CURSOR_IN:
      _push_command(cmd,ch);
      cursor_move(ar->cr,0,0,1);
      main_set_draw();
      break;
    case CURSOR_TURN_DOWN:
      _push_command(cmd,ch);
      cursor_turn_down(ar->cr);
      main_set_draw();
      break;
    case CURSOR_TURN_LEFT:
      _push_command(cmd,ch);
      cursor_turn_left(ar->cr);
      main_set_draw();
      break;
    case CURSOR_TURN_UP:
      _push_command(cmd,ch);
      cursor_turn_up(ar->cr);
      main_set_draw();
      break;
    case CURSOR_TURN_RIGHT:
      _push_command(cmd,ch);
      cursor_turn_right(ar->cr);
      main_set_draw();
      break;
    case CURSOR_MODE_SOLID:
      _push_command(cmd,ch);
      cursor_mode_solid(ar->cr);
      main_set_draw();
      break;
    case CURSOR_MODE_BRICK:
      _push_command(cmd,ch);
      cursor_mode_brick(ar->cr);
      main_set_draw();
      break;
    case CURSOR_MODE_PARCEL:
      _push_command(cmd,ch);
      cursor_mode_parcel(ar->cr);
      main_set_draw();
      break;
    case CURSOR_MODE_ROBOT:
      _push_command(cmd,ch);
      cursor_mode_robot(ar->cr);
      main_set_draw();
      break;
    case CURSOR_MODE_HERO:
      _push_command(cmd,ch);
      cursor_mode_hero(ar->cr);
      main_set_draw();
      break;
    case CURSOR_MODE_EXIT_PORTAL:
      _push_command(cmd,ch);
      cursor_mode_exit_portal(ar->cr);
      main_set_draw();
      break;
    case CURSOR_MODE_ROBOT_PORTAL:
      _push_command(cmd,ch);
      cursor_mode_robot_portal(ar->cr);
      main_set_draw();
      break;
    case CURSOR_TOGGLE_PEN: /** toggle erase mode */
      _push_command(cmd,ch);
      cursor_toggle_pen(ar->cr);
      main_set_draw();
      break;
    case CURSOR_CMODE_DOT:
      _push_command(cmd,ch);
      cursor_cmode_dot(ar->cr);
      main_set_draw();
      break;
    case CURSOR_CMODE_IWALL:
      _push_command(cmd,ch);
      cursor_cmode_iwall(ar->cr);
      main_set_draw();
      break;
    case CURSOR_CMODE_JWALL:
      _push_command(cmd,ch);
      cursor_cmode_jwall(ar->cr);
      main_set_draw();
      break;
    case CURSOR_CMODE_KWALL:
      _push_command(cmd,ch);
      cursor_cmode_kwall(ar->cr);
      main_set_draw();
      break;
    case CURSOR_CMODE_IPIPE:
      _push_command(cmd,ch);
      cursor_cmode_ipipe(ar->cr);
      main_set_draw();
      break;
    case CURSOR_CMODE_JPIPE:
      _push_command(cmd,ch);
      cursor_cmode_jpipe(ar->cr);
      main_set_draw();
      break;
    case CURSOR_CMODE_KPIPE:
      _push_command(cmd,ch);
      cursor_cmode_kpipe(ar->cr);
      main_set_draw();
      break;
/***** see SPACE_BAR ********
    case CURSOR_CLICK:
      _push_command(cmd,ch);
      cursor_click(ar->cr);
      main_set_draw();
      break;
*****************************/
    case CAM_TRACK_UP:
      _push_command(cmd,ch);
      cam_track_up(&ar->c,map_get_center(&ar->m),5);
      main_set_cam();
      break;
    case CAM_TRACK_DOWN:
      _push_command(cmd,ch);
      cam_track_down(&ar->c,map_get_center(&ar->m),5);
      main_set_cam();
      break;
    case CAM_TRACK_LEFT:
      _push_command(cmd,ch);
      cam_track_left(&ar->c,map_get_center(&ar->m),5);
      main_set_cam();
      break;
    case CAM_TRACK_RIGHT:
      _push_command(cmd,ch);
      cam_track_right(&ar->c,map_get_center(&ar->m),5);
      main_set_cam();
      break;
#ifdef ADVANCED_COMMANDS
    case CAM_UP:
      _push_command(cmd,ch);
      cam_up(&ar->c);
      main_set_cam();
      break;
    case CAM_DOWN:
      _push_command(cmd,ch);
      cam_down(&ar->c);
      main_set_cam();
      break;
    case CAM_LEFT:
      _push_command(cmd,ch);
      cam_left(&ar->c);
      main_set_cam();
      break;
    case CAM_RIGHT:
      _push_command(cmd,ch);
      cam_right(&ar->c);
      main_set_cam();
      break;
    case CAM_IN:
      _push_command(cmd,ch);
      cam_in(&ar->c);
      main_set_cam();
      break;
    case CAM_OUT:
      _push_command(cmd,ch);
      cam_out(&ar->c);
      main_set_cam();
      break;
    case CAM_ROTATE_UP: 
      _push_command(cmd,ch);
      cam_rotate_up(&ar->c);
      main_set_cam();
      break;
    case CAM_ROTATE_DOWN:
      _push_command(cmd,ch);
      cam_rotate_down(&ar->c);
      main_set_cam();
      break;
    case CAM_ROTATE_LEFT:
      _push_command(cmd,ch);
      cam_rotate_left(&ar->c);
      main_set_cam();
      break;
    case CAM_ROTATE_RIGHT:
      _push_command(cmd,ch);
      cam_rotate_right(&ar->c);
      main_set_cam();
      break;
    case CAM_FLIP:
      _push_command(cmd,ch);
      cam_flip(&ar->c,map_get_center(&ar->m));
      main_set_cam();
      break;
    case CAM_ZOOM_IN:
      _push_command(cmd,ch);
      cam_zoom_in(&ar->c);
      main_set_cam();
      break;
    case CAM_ZOOM_OUT:
      _push_command(cmd,ch);
      cam_zoom_out(&ar->c);
      main_set_cam();
      break;
    case CAM_CENTER:
      _push_command(cmd,ch);
      cam_center(&ar->c,ar->m.bi,ar->m.bj,ar->m.bk);
      main_set_cam();
      break;
#endif
    case QUIT:
      main_set_quit();
      break;
    default :
      return 0;
  }
  return 1;
}
