#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <assert.h>
#include "cam.h"
#include "cursor.h"
#include "render.h"
#include "globals.h"
#include "scene.h"
#include "colour.h"

#define CURSOR_SOLID_MODE 		1
#define CURSOR_BRICK_MODE 		2
#define CURSOR_PARCEL_MODE		3
#define CURSOR_ROBOT_MODE		4
#define CURSOR_HERO_MODE		5
#define CURSOR_ROBOT_PORTAL_MODE	6
#define CURSOR_EXIT_PORTAL_MODE		7

const float _cr_x[8*3]=
{
  0.0,0.0,0.0,
  0.0,1.0,0.0,
  0.0,0.0,1.0,
  0.0,1.0,1.0,

  1.0,0.0,0.0,
  1.0,1.0,0.0,
  1.0,0.0,1.0,
  1.0,1.0,1.0,
};

void
__make_cursor_coords(struct cursor*cr,float x[8*3])
{
  int i;
  memcpy(x,_cr_x,sizeof(float)*3*8);
  for(i=0;i<8;i++)
    ivec_operate(cr->iv,x+3*i);
}

/** first thing to do with a xmalloced cursor */
void
cursor_init(
  struct cursor*cr,
  struct scene*s,
  struct map*m,
  struct people*pl,
  int i, int j, int k)
{
  int idx;
  float x[12*2];
  cr->s=s;
  cr->m=m;
  cr->pl=pl;
  cr->iv.x[0]=i;
  cr->iv.x[1]=j;
  cr->iv.x[2]=k;
  cr->mode=CURSOR_BRICK_MODE;
  cr->cmode=CURSOR_CMODE_DOT; /** one brick at a time */
  cr->erase=0;
  for(idx=0;idx<9;idx++)
    if((idx%3)==(idx/3))
      cr->iv.d[idx]=1;
    else
      cr->iv.d[idx]=0;
  __make_cursor_coords(cr,x);
  cr->render=render_box;
  cr->d = scene_new_draw_item(
    s,x,8,colour_lookup(CURSOR_CID,0),render_box,IS_DYNAMIC);
}

/** synchronize the draw_item with the cursor state */
void
_cursor_update(struct cursor*cr)
{
  float x[12*2];
  __make_cursor_coords(cr,x);
  scene_update_draw_item(
    cr->s,cr->d,x,colour_lookup(CURSOR_CID,0),cr->render);
}

void
cursor_set(struct cursor*cr, struct ivec iv)
{
  cr->iv=iv;
  _cursor_update(cr);
}

/** a relative movement ***/
int
cursor_move(struct cursor*cr,int di, int dj, int dk)
{
  if(cr==NULL)
    return 1;
  /** an instruction from user land! */
  if( cr->iv.x[0]+di < 0 ) return -1;
  if( cr->iv.x[1]+dj < 0 ) return -2;
  if( cr->iv.x[2]+dk < 0 ) return -3;
  if( cr->iv.x[0]+di >= cr->m->bi ) return -4;
  if( cr->iv.x[1]+dj >= cr->m->bj ) return -5;
  if( cr->iv.x[2]+dk >= cr->m->bk ) return -6;
#ifdef PRINT_STUFF
  fprintf(logfile,"moving cursor\n");
#endif
  cr->iv.x[0]+=di;
  cr->iv.x[1]+=dj;
  cr->iv.x[2]+=dk;
  _cursor_update(cr);
  return 1;
}

void
cursor_turn_up(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->iv=ivec_turn_up(cr->iv);
  _cursor_update(cr);
}

void
cursor_turn_down(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->iv=ivec_turn_down(cr->iv);
  _cursor_update(cr);
}

void
cursor_turn_left(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->iv=ivec_turn_left(cr->iv);
  _cursor_update(cr);
}

void
cursor_turn_right(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->iv=ivec_turn_right(cr->iv);
  _cursor_update(cr);
}

void
cursor_pan1(struct cursor*cr)
{
  int i,a;
  if(cr==NULL)
    return;
  for(i=0;i<3;i++)
  {
    a=cr->iv.d[3*i];
    cr->iv.d[3*i]=cr->iv.d[1+3*i];
    cr->iv.d[1+3*i]=a;
  }
  _cursor_update(cr);
}

void
cursor_pan2(struct cursor*cr)
{
  int i,a;
  if(cr==NULL)
    return;
  for(i=0;i<3;i++)
  {
    a=cr->iv.d[3*i+1];
    cr->iv.d[3*i+1]=cr->iv.d[2+3*i];
    cr->iv.d[2+3*i]=a;
  }
  _cursor_update(cr);
}

void
cursor_print(struct cursor*cr)
{
  if(cr==NULL)
    return;
  fprintf(logfile,"cursor = {\n");
  print_x(cr->iv.x,1,3);
  print_x(cr->iv.d,3,3);
  fprintf(logfile,"};\n");
}

void
cursor_toggle_pen(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->erase=!cr->erase;
}

void
cursor_cmode_dot(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->cmode=CURSOR_CMODE_DOT;
}

void
cursor_cmode_ipipe(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->cmode=CURSOR_CMODE_IPIPE;
}

void
cursor_cmode_jpipe(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->cmode=CURSOR_CMODE_JPIPE;
}

void
cursor_cmode_kpipe(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->cmode=CURSOR_CMODE_KPIPE;
}

void
cursor_cmode_iwall(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->cmode=CURSOR_CMODE_IWALL;
}

void
cursor_cmode_jwall(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->cmode=CURSOR_CMODE_JWALL;
}

void
cursor_cmode_kwall(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->cmode=CURSOR_CMODE_KWALL;
}

void
cursor_click(struct cursor*cr)
{
  if(cr==NULL)
    return;
  if ( cr->mode==CURSOR_BRICK_MODE ||
    cr->mode==CURSOR_SOLID_MODE )
  {
      int i=cr->iv.x[0];
      int j=cr->iv.x[1];
      int k=cr->iv.x[2];
      int _i, _j, _k, i0, j0, k0, i1, j1, k1;
      i0=i; i1=i+1;
      j0=j; j1=j+1;
      k0=k; k1=k+1;
      switch(cr->cmode)
      {
        case CURSOR_CMODE_DOT:
          break;
        case CURSOR_CMODE_IPIPE:
          i0=0; i1=cr->m->bi;
          break;
        case CURSOR_CMODE_JPIPE:
          j0=0; j1=cr->m->bj;
          break;
        case CURSOR_CMODE_KPIPE:
          k0=0; k1=cr->m->bk;
          break;
        case CURSOR_CMODE_IWALL:
          j0=0; j1=cr->m->bj;
          k0=0; k1=cr->m->bk;
          break;
        case CURSOR_CMODE_JWALL:
          k0=0; k1=cr->m->bk;
          i0=0; i1=cr->m->bi;
          break;
        case CURSOR_CMODE_KWALL:
          j0=0; j1=cr->m->bj;
          i0=0; i1=cr->m->bi;
          break;
        default:
          assert(0);
      }
      for(_i=i0;_i<i1;_i++)
      for(_j=j0;_j<j1;_j++)
      for(_k=k0;_k<k1;_k++)
        if(cr->erase)
        {
          if(map_is_brick(cr->m,_i,_j,_k))
            map_unset_brick(cr->m,_i,_j,_k);
        }
        else
        {
          if(map_is_brick(cr->m,_i,_j,_k))
            map_unset_brick(cr->m,_i,_j,_k);
          map_set_brick(cr->m,_i,_j,_k,
             cr->mode&CURSOR_SOLID_MODE);
        }
/********
          if(!map_is_brick(cr->m,_i,_j,_k))
            map_set_brick(cr->m,_i,_j,_k,
              cr->mode&CURSOR_SOLID_MODE);
********/
  }
  else
  switch(cr->mode)
  {
    case CURSOR_PARCEL_MODE:
      map_toggle_parcel(cr->m,cr->iv.x[0],cr->iv.x[1],cr->iv.x[2]);
      break;
    case CURSOR_ROBOT_MODE:
      people_toggle_robot(cr->pl,cr->iv);
      break;
    case CURSOR_HERO_MODE:
      people_set_hero(cr->pl,cr->iv);
      break;
    case CURSOR_EXIT_PORTAL_MODE:
      map_toggle_portal(cr->m,cr->iv);
      break;
    case CURSOR_ROBOT_PORTAL_MODE:
      people_set_robot_portal(cr->pl,cr->iv);
      break;
    default:
      assert(0);
  }
}

void
cursor_mode_solid(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->mode=CURSOR_SOLID_MODE;
  cr->render=render_box;
  _cursor_update(cr);
}

void
cursor_mode_brick(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->mode=CURSOR_BRICK_MODE;
  cr->render=render_box;
  _cursor_update(cr);
}

void
cursor_mode_parcel(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->mode=CURSOR_PARCEL_MODE;
  cr->render=render_box;
  _cursor_update(cr);
}

void
cursor_mode_robot(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->mode=CURSOR_ROBOT_MODE;
  cr->render=render_dalek;
  _cursor_update(cr);
}

void
cursor_mode_hero(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->mode=CURSOR_HERO_MODE;
  cr->render=render_dalek;
  _cursor_update(cr);
}

void
cursor_mode_exit_portal(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->mode=CURSOR_EXIT_PORTAL_MODE;
  cr->render=render_dalek;
  _cursor_update(cr);
}

void
cursor_mode_robot_portal(struct cursor*cr)
{
  if(cr==NULL)
    return;
  cr->mode=CURSOR_ROBOT_PORTAL_MODE;
  cr->render=render_dalek;
  _cursor_update(cr);
}

void
cursor_finalize(struct cursor*cr)
{
  if(cr==NULL)
    return;
  scene_del_draw_item(cr->s,cr->d);
}

