#ifndef __CURSOR_H
#define __CURSOR_H

#include "ivec.h"

struct scene;
struct draw_item;

#define CURSOR_COLOUR 52

#define CURSOR_BRICK_MODE 		1
#define CURSOR_PARCEL_MODE		2
#define CURSOR_ROBOT_MODE		3
#define CURSOR_HERO_MODE		4
#define CURSOR_ROBOT_PORTAL_MODE	5
#define CURSOR_EXIT_PORTAL_MODE		6

#define CURSOR_CMODE_DOT		0
#define CURSOR_CMODE_IPIPE		1
#define CURSOR_CMODE_JPIPE		2
#define CURSOR_CMODE_KPIPE		3
#define CURSOR_CMODE_IWALL		4
#define CURSOR_CMODE_JWALL		5
#define CURSOR_CMODE_KWALL		6


struct cursor
{
  struct scene*s;
  struct map*m;  /* cursor needs to interact with map */
  struct people*pl; /* and with people */
  int mode;
  int cmode;
  int erase; /** 0 or 1 */
  struct ivec iv; /* location, direction */
  struct draw_item*d;
  void (*render)(float*,int);
};

void
cursor_init(
  struct cursor*cr,
  struct scene*s,
  struct map*,
  struct people*,
  int i, int j, int k);

void
cursor_click(struct cursor*cr);

void
cursor_toggle_pen(struct cursor*cr);

void
cursor_cmode_dot(struct cursor*);

void
cursor_cmode_ipipe(struct cursor*cr);

void
cursor_cmode_jpipe(struct cursor*cr);

void
cursor_cmode_kpipe(struct cursor*cr);

void
cursor_cmode_iwall(struct cursor*cr);

void
cursor_cmode_jwall(struct cursor*cr);

void
cursor_cmode_kwall(struct cursor*cr);

void
cursor_mode_brick(struct cursor*cr);

void
cursor_mode_solid(struct cursor*cr);

void
cursor_mode_parcel(struct cursor*cr);

void
cursor_mode_robot(struct cursor*cr);

void
cursor_mode_hero(struct cursor*cr);

void
cursor_mode_robot_portal(struct cursor*cr);

void
cursor_mode_exit_portal(struct cursor*cr);

int
cursor_move(struct cursor*,int di, int dj, int dk); /* relative movement */

void
cursor_set(struct cursor*cr, struct ivec);

void
cursor_turn_up(struct cursor*cr);

void
cursor_turn_down(struct cursor*cr);

void
cursor_turn_left(struct cursor*cr);

void
cursor_turn_right(struct cursor*cr);

void
cursor_pan1(struct cursor*cr);

void
cursor_pan2(struct cursor*cr);

void
cursor_finalize(struct cursor*);

#endif
