#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>

#include <SDL/SDL.h>

SDL_Surface*screen;
int width,height;
Uint32 _black;
Uint32 _red,_white,_blue,_green;

#define SGN(x) ((x)>0?1:((x)==0?0:(-1)))
#define ABS(x) ((x)>0?(x):(-x))

#define FG_COLOR colors[0]
#define BG_COLOR colors[nlevels-1];
#define NUM_LEVELS 256
#define DRAW 

static unsigned char gamma_table[256];

static double Draw_gamma = 2.35;  /* looks good enough */

static void
draw_abort(void)
{
    fprintf(stderr, "problem: %s\n", SDL_GetError());
    exit(1);
}

extern int found_joystick;

void
draw_init()
{
  int i;
  Uint32 vmod,xres,yres;
  /* Initialize SDL */
  if ( SDL_Init(SDL_INIT_VIDEO|SDL_INIT_JOYSTICK) < 0 )
  {
    found_joystick=0;
    printf("Could not init joystick\n");
    if ( SDL_Init(SDL_INIT_VIDEO) < 0 )
      draw_abort();
  }
  else 
  {
    found_joystick=1;
    printf("draw_init: init joystick OK\n");
  }

  atexit(SDL_Quit);
  vmod=SDL_HWSURFACE|SDL_DOUBLEBUF;
#define _FULL_SCREEN
#ifdef FULL_SCREEN
  vmod|=SDL_FULLSCREEN;
#endif
  xres=800;
  yres=600;
  screen=NULL;
  screen = SDL_SetVideoMode( xres, yres, 0, vmod );
  if ( screen == NULL ) 
      draw_abort();
  else
  {
    printf( "render_init(): new screen, %d bytes pp,", 
      screen->format->BytesPerPixel );
    printf( " %dx%d\n", 
      screen->w, screen->h );
  }
  height=screen->h;
  width=screen->w;
#ifdef FULL_SCREEN
  SDL_ShowCursor(SDL_DISABLE);
#endif

  /* generate gamma correction table */
  for (i=0; i<NUM_LEVELS; i++)
    gamma_table[i] = (int) (NUM_LEVELS-1)
      *pow((double)i/((double)(NUM_LEVELS-1)), 
        1.0/Draw_gamma);
  _black=SDL_MapRGB(screen->format,0,0,0);
  _white=SDL_MapRGB(screen->format,255,255,255);
  _red=SDL_MapRGB(screen->format,255,0,0);
  _green=SDL_MapRGB(screen->format,0,255,0);
  _blue=SDL_MapRGB(screen->format,0,0,255);
}

void draw_finalize()
{
#ifdef DRAW
  SDL_FreeSurface(screen);
  SDL_Quit();
#endif
}

void
draw_update()
{
#ifdef DRAW
  SDL_Flip(screen);
#endif
}

void
draw_pixel(int x,int y,int c)
{
  Uint32 bpp, ofs;

  bpp = screen->format->BytesPerPixel;
  ofs = screen->pitch*y + x*bpp;

  memcpy(screen->pixels + ofs, &c, bpp);
}

void 
draw_wu_line(Uint32 x0, Uint32 y0, Uint32 x1, Uint32 y1,
  Uint32 fgc, Uint32 bgc, int nlevels, int nbits)
{
  Uint32 intshift, erracc,erradj;
  Uint32 erracctmp, wgt, wgtcompmask;
  int dx, dy, tmp, xdir, i;
  Uint32 colors[nlevels];
  SDL_Rect r;
  SDL_Color fg, bg;

  SDL_GetRGB(fgc, screen->format, &fg.r, &fg.g, &fg.b);
  SDL_GetRGB(bgc, screen->format, &bg.r, &bg.g, &bg.b);

  /* generate the colors by linear interpolation,
    applying gamma correction */
  for (i=0; i<nlevels; i++) {
    Uint8 r, g, b;

    r = gamma_table[fg.r - (i*(fg.r - bg.r))/(nlevels-1)];
    g = gamma_table[fg.g - (i*(fg.g - bg.g))/(nlevels-1)];
    b = gamma_table[fg.b - (i*(fg.b - bg.b))/(nlevels-1)];
    colors[i] = SDL_MapRGB(screen->format, r, g, b);
  }
  if (y0 > y1) {
    tmp = y0; y0 = y1; y1 = tmp;
    tmp = x0; x0 = x1; x1 = tmp;
  }
  SDL_LockSurface(screen);
  /* draw the initial pixel in the foreground color */
  draw_pixel(x0, y0, FG_COLOR);

  dx = x1 - x0;
  xdir = (dx >= 0) ? 1 : -1;
  dx = (dx >= 0) ? dx : -dx;

  /* special-case horizontal, vertical, 
    and diagonal lines which need no
    weighting because they go right through 
    the center of every pixel. */
  if ((dy = y1 - y0) == 0) {
    /* horizontal line */
    r.x = (x0 < x1) ? x0 : x1;
    r.y = y0;
    r.w = dx;
    r.h = 1;
    SDL_FillRect(screen, &r, FG_COLOR);
    SDL_UnlockSurface(screen);
    return;
  }

  if (dx == 0) {
    /* vertical line */
    r.x = x0;
    r.y = y0;
    r.w = 1;
    r.h = dy;
    SDL_FillRect(screen, &r, FG_COLOR);
    SDL_UnlockSurface(screen);
    return;
  }

  if (dx == dy) {
    for (; dy != 0; dy--) {
      x0 += xdir;
      y0++;
     draw_pixel(x0, y0, FG_COLOR);
    }
    SDL_UnlockSurface(screen);
    return;
  }
  /* line is not horizontal, vertical, or diagonal */
  erracc = 0;	/* err. acc. is initially zero */
  /* # of bits by which to shift erracc 
    to get intensity level */
  intshift = 32 - nbits;
  /* mask used to flip all bits in an intensity weighting */
  wgtcompmask = nlevels - 1;
  /* x-major or y-major? */
  if (dy > dx) {
    /* y-major.  
       Calculate 16-bit fixed point 
       fractional part of a pixel that
       X advances every time Y advances 1 pixel, 
       truncating the result so that
       we won't overrun the endpoint along the X axis */
    erradj = ((Uint64)dx << 32) / (Uint64)dy;
    /* draw all pixels other than the first and last */
    while (--dy) {
      erracctmp = erracc;
      erracc += erradj;
      if (erracc <= erracctmp) {
	/* rollover in error accumulator, x coord advances */
	x0 += xdir;
      }
      y0++; /* y-major so always advance Y */
      /* the nbits most significant bits 
         of erracc give us the intensity
	 weighting for this pixel, 
         and the complement of the weighting for
	 the paired pixel. */
      wgt = erracc >> intshift;
      draw_pixel(x0, y0, colors[wgt]);
      draw_pixel(x0+xdir, y0, colors[wgt^wgtcompmask]);
    }
    /* draw the final pixel,
      which is always exactly intersected by the line
       and so needs no weighting */
    draw_pixel(x1, y1, FG_COLOR);
    SDL_UnlockSurface(screen);
    return;
  }
  /* x-major line.  Calculate 16-bit 
     fixed-point fractional part of a pixel
     that Y advances each time X advances 1 pixel,
     truncating the result so
     that we won't overrun the endpoint along the X axis. */
  erradj = ((Uint64)dy << 32) / (Uint64)dx;
  /* draw all pixels other than the first and last */
  while (--dx) {
    erracctmp = erracc;
    erracc += erradj;
    if (erracc <= erracctmp) {
      /* accumulator turned over, advance y */
      y0++;
    }
    x0 += xdir;	/* x-major so always advance X */
    /* the nbits most significant bits 
       of erracc give us the intensity
       weighting for this pixel, 
       and the complement of the weighting for
       the paired pixel. */
    wgt = erracc >> intshift;
    draw_pixel(x0, y0, colors[wgt]);
    draw_pixel(x0, y0+1, colors[wgt^wgtcompmask]);
  }
  /* draw final pixel, 
     always exactly intersected by the line and doesn't
     need to be weighted. */
  draw_pixel(x1, y1, FG_COLOR);
  SDL_UnlockSurface(screen);
}

void
draw_line(int x1,int y1,int x2,int y2,int color)
{
  int lg_delta, sh_delta, cycle, lg_step, sh_step;

  lg_delta = x2 - x1;
  sh_delta = y2 - y1;
  lg_step = SGN(lg_delta);
  lg_delta = ABS(lg_delta);
  sh_step = SGN(sh_delta);
  sh_delta = ABS(sh_delta);
  SDL_LockSurface(screen);
  if (sh_delta < lg_delta) {
    cycle = lg_delta >> 1;
    while (x1 != x2) {
      draw_pixel(x1, y1, color);
      cycle += sh_delta;
      if (cycle > lg_delta) {
	cycle -= lg_delta;
	y1 += sh_step;
      }
      x1 += lg_step;
     }
    draw_pixel(x1, y1, color);
  }
  cycle = sh_delta >> 1;
  while (y1 != y2) {
    draw_pixel(x1, y1, color);
    cycle += lg_delta;
    if (cycle > sh_delta) {
      cycle -= sh_delta;
      x1 += lg_step;
    }
    y1 += sh_step;
  }
  draw_pixel(x1, y1, color);
  SDL_UnlockSurface(screen);
}

void
draw_fillbox(int x0,int y0,int x1,int y1,int c)
{
  SDL_Rect rect={x0,y0,x1-x0,y1-y0};
  SDL_LockSurface(screen);
  SDL_FillRect(screen,&rect,c);
  SDL_UnlockSurface(screen);
}

void
draw_hline(int x0,int y,int x1,int c)
{
  int bpp=screen->format->BytesPerPixel;
  if(x0>x1) ;
  else
  if(bpp==2)
  {
    int i;
    Uint16* mem=screen->pixels
      +y*screen->pitch
      +x0*bpp;
    Uint16 col=c;
    for(i=x1-x0;i;i--,mem++)
      *mem=col;
  }
  else
  {
    int i,j;
    for(j=0;j<bpp;j++)
    {
      Uint8* mem=(Uint8*)screen->pixels
        +y*screen->pitch
        +x0*bpp+j;
      Uint8 col=((Uint8*)&c)[j];
      for(i=x1-x0;i;i--,mem+=bpp)
        *mem=col;
    }
  }
}


