#include <string.h>
#include <unistd.h>
#include <assert.h>

#include "command.h"
#include "mtime.h"

#include <SDL/SDL.h>

int found_joystick=0;
static SDL_Joystick*joystick=NULL;
static int joynum=0;

#define MAX_KEYS 512
static char keys[MAX_KEYS];
#define MAX_AXIS 5
static int values[MAX_AXIS];
#define MAX_BUTTONS 10
static char buttons[MAX_BUTTONS];

void
event_init()
{
  int i;

  for(i=0;i<MAX_KEYS;keys[i++]=0);
  for(i=0;i<MAX_AXIS;values[i++]=0);
  for(i=0;i<MAX_BUTTONS;buttons[i++]=0);

  found_joystick=0; /* no go */
  if (found_joystick)
  {
    int n;
    n=SDL_NumJoysticks();
    if(n==0)
      printf("event_init: could not find any joysticks\n");
    else
    {
      joynum=0;
      printf("event_init: using joystick %s\n",
        SDL_JoystickName(joynum));
      SDL_JoystickEventState(SDL_ENABLE);
      joystick = SDL_JoystickOpen(joynum);
      if(joystick==NULL)
        printf("event_init: NULL joystick error\n");
    }
  }
}

void
event_finalize()
{
  if(joystick)
    SDL_JoystickClose(joystick);
}

static
int mkchar(int key, int mod)
{
  switch(key)
  {
    case 27 : return 'Q'; /* escape */
    case 13 : return '\\'; /* return */
    case SDLK_UP : return PERSON_WALK_FORWARD;
    case SDLK_DOWN : return PERSON_WALK_BACK;
    case SDLK_LEFT : return PERSON_WALK_LEFT;
    case SDLK_RIGHT : return PERSON_WALK_RIGHT;
  }
  if(!(mod&KMOD_SHIFT))
    return key;
  if('a'<=key && key<='z' )
    return key-'a'+'A';
  if('1'<=key && key<='5' )
    return key-'1'+'!';
  switch(key)
  {
    case '6': return '^';
    case '7': return '&';
    case '8': return '*';
    case '9': return '(';
    case '0': return ')';
    case '-': return '_';
    case '=': return '+';
    case '[': return '{';
    case ']': return '}';
    case '\\': return '|';
    case ';': return ':';
    case '\'': return '"';
    case ',': return '<';
    case '.': return '>';
    case '/': return '?';
    case '`': return '~';
  }
  return 0;
}

int event_wait()
{
  /* wait for key or QUIT event */
  SDL_Event event;
#ifdef DEBUG_GETCH
  printf("render_getch...\n");
#endif
  while ( SDL_WaitEvent(&event) >= 0 )
  {
    if ( event.type==SDL_KEYDOWN )
      return mkchar( event.key.keysym.sym,
        event.key.keysym.mod );
    if ( event.type==SDL_QUIT )
      return 'Q';
  }
  return 0;
}

static
void
set_key(int key,char flag)
{
  if(0<key && key<MAX_KEYS)
    keys[key]=flag;
}

int event_poll(int max,int*poll_msec)
{
  /* non blocking poll */
  /* return 0 if no event */
  SDL_Event event;
  int key=0;
  int i;
  for(i=0; (i<max || !max) && SDL_PollEvent(&event) > 0;i++ )
  {
    /* we got an event */
    /* int _msec=mtime_get(); */
    switch(event.type)
    {
#ifdef UNUSED
      int value;
      int axis;
      int button;
#endif
      case SDL_QUIT:
        return 'Q';
      case SDL_KEYDOWN: 
        key = mkchar( event.key.keysym.sym,
          event.key.keysym.mod );
        if(poll_msec)*poll_msec=40;
        set_key(key,1);
        break;
      case SDL_KEYUP:
        set_key(key,0);
        break;
#ifdef JOYSTICK
      case SDL_JOYAXISMOTION:
        if(poll_msec)*poll_msec=400;
        value=event.jaxis.value;
        axis=event.jaxis.axis;
        if (value<-8200||value>8200) 
        {
          printf("joystick axis %d value %d\n", axis,value );
          if(axis==0)
          {
/******
            if(value<0) key=PERSON_WALK_LEFT;
            else key=PERSON_WALK_RIGHT;
******/
            if(value<0) key=CAM_TRACK_LEFT;
            else key=CAM_TRACK_RIGHT;
          }
          else if(axis==1)
          {
/******
            if(value>0) key=PERSON_WALK_FORWARD;
            else key=PERSON_WALK_BACK;
******/
            if(value<0) key=CAM_TRACK_DOWN;
            else key=CAM_TRACK_UP;
          }
        }
        break;
      case SDL_JOYBUTTONDOWN:
        button=event.jbutton.button;
        if(button<MAX_BUTTONS)
        {
          printf("button %d down\n",button);
          buttons[button]=1;
        }
        break;
      case SDL_JOYBUTTONUP:
        button=event.jbutton.button;
        if(button<MAX_BUTTONS)
        {
          printf("button %d up\n",button);
          buttons[button]=0;
        }
        break;
#endif
    }/* switch */
  }/*for*/ 
#define _DEBUG_POLL
#ifdef DEBUG_POLL
  if(i) printf("event_poll : %d events, key=%d\n",i,key);
#endif
  return key;
}

void
event_keyrepeat(int on,int rpt)
{
  if ( SDL_EnableKeyRepeat(on,rpt) < 0 )
    printf("event_keyrepeat error\n");
}

void
event_flush()
{
  SDL_Event event;
  while(SDL_PollEvent(&event));
}


