#include <string.h>
#include <assert.h>
#include <stdio.h>
#include "ivec.h"
#include "algebra.h"
#include "globals.h"

void
ivec_print(struct ivec iv)
{
  int i;
  fprintf(logfile,"   <x=[");
  for(i=0;i<3;i++)
    fprintf(logfile,"%.1f,",iv.x[i]);
  fprintf(logfile,"],d=[");
  for(i=0;i<9;i++)
    fprintf(logfile,"%.1f,",iv.d[i]);
  fprintf(logfile,"]>\n");
}

/* comparison */
int
ivec_cmp(struct ivec iv,struct ivec _iv)
{
  float res;
  res=xmetric3(iv.x,_iv.x,1)+xmetric3(iv.d,_iv.d,3);
#define _DEBUG_IVEC
#ifdef DEBUG_IVEC
  fprintf(logfile,"ivec_cmp\niv=\n");
  ivec_print(iv);
  fprintf(logfile,"_iv=\n");
  ivec_print(_iv);
  fprintf(logfile,"    res=%f\n", res );
#endif
  return res>epsilon;
  return memcmp(&iv,&_iv,sizeof(struct ivec));
}

int
ivec_ucmp(struct ivec iv,struct ivec _iv)
{
  int di=_iv.d[2];
  int dj=_iv.d[5];
  int dk=_iv.d[8];
  memcpy(_iv.d,iv.d,sizeof(float)*9);
  _iv.d[2]=di;
  _iv.d[5]=dj;
  _iv.d[8]=dk;
  return ivec_cmp(iv,_iv);
}

float
ivec_xmetric(struct ivec iv,struct ivec _iv)
{
  return xmetric3(iv.x,_iv.x,1);
}

/** only compare the position */
int
ivec_xcmp(struct ivec iv,struct ivec _iv)
{
  float res;
  return ivec_xmetric(iv,_iv)>epsilon;
  res=xmetric3(iv.x,_iv.x,1);
  return res>epsilon;
  return memcmp(&iv.x,&_iv.x,sizeof(float)*3);
}

/* only compare direction */
int
ivec_dcmp(struct ivec iv,struct ivec _iv)
{
  return xmetric3(iv.d,_iv.d,3)>epsilon;
  return memcmp(&iv.d,&_iv.d,sizeof(float)*9);
}

void
ivec_xcpy(struct ivec*iv,struct ivec*_iv)
{
  memcpy(iv,_iv,sizeof(float)*3);
}

void
ivec_dcpy(struct ivec*iv,struct ivec*_iv)
{
  memcpy((float*)iv+3,
    (float*)_iv+3,sizeof(float)*9);
}

/** orient and translate x vector */
void
ivec_operate(struct ivec iv,float*_x)
{
  float x[3];
  int j,k;
  assert(_x);
  for(j=0;j<3;j++)
    _x[j]-=0.5;
  for(j=0;j<3;j++)
  {
    x[j]=0.0;
    for(k=0;k<3;k++)
      x[j]+=iv.d[3*j+k]*_x[k];
  }
  for(j=0;j<3;j++)
    x[j]+=iv.x[j];
  memcpy(_x,x,sizeof(float)*3);
  for(j=0;j<3;j++)
    _x[j]+=0.5;
}

struct ivec
ivec_interpolate(struct ivec iv,struct ivec _iv,float t)
{
  struct ivec __iv;
  int i;
  for(i=0;i<3;i++)
    __iv.x[i]=(1.0-t)*iv.x[i]+t*_iv.x[i];
  /** this is crude but it should work for acute angles */
  for(i=0;i<9;i++)
    __iv.d[i]=(1.0-t)*iv.d[i]+t*_iv.d[i];
/*
  ortho(__iv.d);
*/
  return __iv;
}

struct ivec
ivec_round(struct ivec iv)
{
  int i;
  for(i=0;i<3;i++)
    iv.x[i]=rint(iv.x[i]);
  for(i=0;i<9;i++)
    iv.d[i]=rint(iv.d[i]);
  return iv;
}

struct ivec
ivec_forward(struct ivec iv)
{
  int i;
  for(i=0;i<3;i++)
    iv.x[i]+=iv.d[3*i];
  iv=ivec_round(iv);
  return iv;
}

struct ivec
ivec_below(struct ivec iv)
{
  int i;
  for(i=0;i<3;i++)
    iv.x[i]-=iv.d[3*i+2];
  iv=ivec_round(iv);
  return iv;
}

struct ivec
ivec_above(struct ivec iv)
{
  int i;
  for(i=0;i<3;i++)
    iv.x[i]+=iv.d[3*i+2];
  iv=ivec_round(iv);
  return iv;
}

struct ivec
ivec_turn_back(struct ivec iv)
{
  int i;
  for(i=0;i<3;i++)
  {
    iv.d[3*i]=-iv.d[3*i];
    iv.d[3*i+1]=-iv.d[3*i+1];
  }
  iv=ivec_round(iv);
  return iv;
}

struct ivec
ivec_turn_left(struct ivec iv)
{
  struct ivec _iv;
  int i;
  _iv=iv;
  for(i=0;i<3;i++)
  {
    _iv.d[3*i]=iv.d[3*i+1];
    _iv.d[3*i+1]=-iv.d[3*i];
  }
  iv=ivec_round(iv);
  return _iv;
}

struct ivec
ivec_turn_right(struct ivec iv)
{
  struct ivec _iv;
  int i;
  _iv=iv;
  for(i=0;i<3;i++)
  {
    _iv.d[3*i]=-iv.d[3*i+1];
    _iv.d[3*i+1]=iv.d[3*i];
  }
  iv=ivec_round(iv);
  return _iv;
}

struct ivec
ivec_turn_up(struct ivec iv)
{
  struct ivec _iv;
  int i;
  _iv=iv;
  for(i=0;i<3;i++)
  {
    _iv.d[3*i]=iv.d[3*i+2];
    _iv.d[3*i+2]=-iv.d[3*i];
  }
  iv=ivec_round(iv);
  return _iv;
}

struct ivec
ivec_turn_down(struct ivec iv)
{
  struct ivec _iv;
  int i;
  _iv=iv;
  for(i=0;i<3;i++)
  {
    _iv.d[3*i]=-iv.d[3*i+2];
    _iv.d[3*i+2]=iv.d[3*i];
  }
  iv=ivec_round(iv);
  return _iv;
}


