#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>

#include "map.h"
#include "algebra.h"
#include "render.h"
#include "cam.h"
#include "scene.h"
#include "memory.h"
#include "globals.h"
#include "architect.h"
#include "command.h"
#include "audio.h"
#include "colour.h"
#include "mtime.h"

int msec_per_frame=80;
int usec_per_frame=80000; /*  */

struct state
{
  int cam;
  int draw;
  int people;

  int caught;
  int level_done;
  int quit;
};

static struct state st={0,0,0,0,0,0};

static
void
state_init()
{
  st.cam=0;
  st.draw=0;
  st.people=0;
  st.caught=0;
  st.level_done=0;
  st.quit=0;
}

void main_set_cam() { st.cam=1; }
void main_set_draw() { st.draw=1; }
void main_set_people() { st.people=1; }

void main_set_caught() { st.caught=1; }
void main_set_level_done() { st.level_done=1; }
void main_set_quit() { st.quit=1; }

FILE*logfile=NULL;

void
open_logfile()
{
  logfile=NULL;
#ifdef NAAAAAAAAHHH
  if(logfile==NULL)
    logfile=fopen("/dev/pts/0","a");
#endif
  if(logfile==NULL)
    logfile=fopen("logfile","w");
  if(logfile==NULL)
  {
    fprintf(stderr,"Can't open logfile.\n");
    abort();
  }
  fprintf(logfile,"THIS IS AXIS RUNNER V1.0\n");
}

void event_init();
void event_finalize();
int event_poll(int max,int*);
int event_wait();
void event_flush();
void event_keyrepeat(int on,int rpt);

static
void
edit_level(struct architect*ar)
{
  render_clear_all();
  while( !st.quit)
  {
    int i;
    char ch;
    if(st.cam)
    {
      scene_update_cam(&ar->s,&ar->c);
      main_set_draw();
      st.cam=0;
    }
    if(st.draw)
    {
      {
        char left[256]={0},
          center[256]={0},
          right[256]={0};
        static int counter=0;
        {
          int i=ar->cr->iv.x[0];
          int j=ar->cr->iv.x[1];
          int k=ar->cr->iv.x[2];
          snprintf(center,256,"%d %d %d",i,j,k);
        }
        render_status(left,center,right);
        counter++;
      }
      st.draw=0;

      if(scene_draw(&ar->s))
        render_update();
    }

    /******* process key events ***********/
    for(i=0;(ch=event_poll(0,NULL)) &&i<3 ;i++)
    {
      command_dispatch(&ar->cmd,ch);
      fflush(logfile);
    }
    if(i==3)
      event_flush();
  }
}

static struct architect ar; /* a singleton */
int level=1;
int lives=5;

static void
_main_instructions()
{
  /** show instructions **/
  FILE*ifile=NULL;
#define N_LINES 10
  char *lines[N_LINES];
  int n=0;

  ifile=fopen("GAMEPLAY","r");
  for(;n<N_LINES;n++)
  {
    lines[n%N_LINES]=xmalloc(256);
    fgets(lines[n%N_LINES],256,ifile);
  }

  while(ifile)
  {
    render_instructions(lines,N_LINES,n);
    if(event_wait()=='Q')
      break;

    if(fgets(lines[n%N_LINES],256,ifile)==NULL)
      break;
    n++;
  }
  for(n=0;n<N_LINES;n++)
    xfree(lines[n]);
  if(ifile)
    fclose(ifile);
}

static int
_do_flashing(int msec,int finish)
{
  int inc=msec;
  int flashing=1;
  struct person*hero=NULL;
  static int skip=0;
  hero=people_get_hero(&ar.pl);

  if (inc>3000||finish)
  {
    flashing=0;
#define _COLOUR_FLASH
#ifdef COLOUR_FLASH
    scene_set_draw_item_colour(
      &ar.s,
      hero->d, 
      colour_lookup(HERO_CID,0) );
#else
    scene_set_draw_item_skip(&ar.s,hero->d,0);
#endif
    main_set_draw();
    skip=0;
  }
  else
  {
    inc%=300;
    /* printf("inc = %d\n",inc); */
    if (!skip&&inc<150)
    {
#ifdef COLOUR_FLASH
      scene_set_draw_item_colour(
        &ar.s,
        hero->d, _black );
#else
      scene_set_draw_item_skip(&ar.s,hero->d,1);
#endif
      main_set_draw();
      skip=1;
      /* printf("skip=%d\n", skip ); */
    }
    else if (skip&&inc>150)
    {
#ifdef COLOUR_FLASH
      scene_set_draw_item_colour(
        &ar.s,
        hero->d, 
        colour_lookup(HERO_CID,0) );
#else
      scene_set_draw_item_skip(&ar.s,hero->d,0);
#endif
      main_set_draw();
      skip=0;
      /* printf("skip=%d\n", skip ); */
    }
  }
  /* printf("skip=%d\n", skip ); */
  return flashing;
}


static
void
_main_render()
{
  char left[256]={0},
    center[256]={0},
    right[256]={0};
  static int counter=0;
  snprintf(left,256,"%d",map_n_parcels(&ar.m));
  snprintf(center,256,"level %d",level);
  snprintf(right,256,"lives %d",lives);
  render_status(left,center,right);
  counter++;

  st.draw=0;

  if(scene_draw(&ar.s))
    render_update();
}

#define NUM_LEVELS 11
static int
_main_tail()
{
  if(st.caught)
  {
    lives--;
    if (!lives)
    {
      audio_start_caught();
      render_banner("GAME OVER");
      event_wait();
      return 1;
    }
    else
    {
      audio_start_caught();
      render_banner("TRY AGAIN");
      sleep(1);
    }
  }
  else if(st.quit)
  {
    audio_start_caught();
    render_banner("GAME QUIT");
    sleep(1);
    return 1;
  }
  else if(st.level_done)
  {
    level=level+1;
    if(level==NUM_LEVELS+1)
    {
      render_banner("~I AM FREE~");
      audio_start_success();
      event_wait();
      return 1;
    }
    render_banner("YO");
    audio_start_success();
    sleep(1);
    lives++;
  }
  return 0;
}

extern Uint32 _black;

static
void
play_game()
{
  level=1;
  lives=5;

  _main_instructions();
  event_poll(0,NULL);

  state_init();
  event_keyrepeat(120,40);

  while(
      !st.caught
      &&!st.level_done
      &&!st.quit)
  {
    int msec,_msec,__msec;
    int flashing=1;
    char filename[256];
    int poll_msec=40;
    char ch=0;

    snprintf(filename,256,"levels/level.%d.axs",level);
    printf("initializing ...\n");
    architect_init(&ar,0,filename);
    printf("done!\n");
    audio_stop();

    msec=mtime_get();
    _msec=msec;
    __msec=msec; /* constant */

    while(flashing&&
      !st.caught
      &&!st.level_done
      &&!st.quit)
    {
      flashing=_do_flashing(msec-__msec,0);

      if(st.cam)
      {
        scene_update_cam(&ar.s,&ar.c);
        main_set_draw();
        st.cam=0;
      }

      if(st.draw)
      {
        _main_render();
        st.draw=0;
      }

      msec=mtime_get();

      /******* process events ***********/
      if(msec-_msec>poll_msec)
      {
        ch=event_poll(0,&poll_msec);
        /*printf("  at %d ch=%d\n",msec-_msec,ch);*/
        _msec=msec;
        if(ch)
        {
          /* rotations only */
          if ( ch==CAM_TRACK_UP
            && ch== CAM_TRACK_DOWN
            && ch== CAM_TRACK_LEFT
            && ch== CAM_TRACK_RIGHT)
            continue;
          else event_flush();
          command_dispatch(&ar.cmd,ch);
        }
      }
    } /* end while flashing */
    event_poll(0,NULL);

    while(
      !st.caught
      &&!st.level_done
      &&!st.quit)
    {
      if(st.cam)
      {
        scene_update_cam(&ar.s,&ar.c);
        main_set_draw();
        st.cam=0;
      }

      if(st.draw)
      {
        _main_render();
        st.draw=0;
      }

      msec=mtime_get();

      /* game play */
      update_people_state(&ar.pl,mtime_get());

      while(update_people_draw(&ar.pl));
      st.draw=1;

      /******* process events ***********/
      if(msec-_msec>poll_msec)
      {
        char ch;
        ch=event_poll(0,&poll_msec);
        /*printf("  at %d ch=%d\n",msec-_msec,ch);*/
        _msec=msec;
        if (ch=='#')
          st.level_done=1;
        else if(ch)
          command_dispatch(&ar.cmd,ch);
      }
    } /* end while ... */
    audio_stop();

    architect_finalize(&ar);
    if( _main_tail() ) 
      break;
    audio_stop();
/*
    event_wait();
*/
    state_init();
  }
  audio_stop();
}

int main(int argc, char**argv)
{
  open_logfile();
  xmark(); /** memory checking */
  if (argc==1)
  {
    audio_init();
    render_init();
    event_init(); /* joystick */
    mtime_init();
    play_game();
    event_finalize();
    render_finalize();
    audio_finalize();
  }
  else
    if(!strcmp(argv[1],"-e")&&argc>=3)
    {
      char filename[256]="";
      strncpy(filename,argv[2],255);
      filename[255]=0;
      state_init();
      render_init();
      printf("initializing ...\n");
      architect_init(&ar,1,filename);
      printf("done!\n");
      event_keyrepeat(120,80);
      edit_level(&ar);
      architect_finalize(&ar);
      render_finalize();
    }
    else
      printf("usage: ... \n");

  {
    int mleak=xcheck();
    if (mleak!=0)
      printf("woah! memory leak = %d\n",mleak); 
  }
  fclose(logfile);
  exit(EXIT_SUCCESS);
}

