/* map.c */
#include <string.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "map.h"
#include "memory.h"
#include "algebra.h"
#include "cam.h"
#include "render.h"
#include "scene.h"
#include "globals.h"
#include "main.h"
#include "colour.h"

#define _DEBUG_MAP

struct mapel
__mel_empty()
{
  struct mapel el={EMPTY_EL,0};
  return el;
}

/****************** boundary *********************/

static
int
_bd_in_range(struct map*m,int iface,int i0,int i1,int dr)
{
  assert(m);
  assert(iface>=0&&iface<6);
  assert(dr==0||dr==1);
  return(i0>=0&&i0<m->brg[iface/2][0]&&i1>=0&&i1<m->brg[iface/2][1]);
}

#ifdef UNUSED
static
int
_is_bd(struct map*m,int iface,int i0,int i1,int dr)
{
  struct boundary**bd;
  assert(m);
  assert(iface>=0&&iface<6);
  assert(_bd_in_range(m,iface,i0,i1,dr));
  assert(dr==0||dr==1);
  /** this is where the relevant struct boundary pointer is */
  bd = m->bd[iface] + (i0*m->brg[iface/2][1]*2+i1*2+dr);
  if(*bd) 
    assert((*bd)->refcount>0);
  return *bd!=NULL;
}
#endif

static
void
_unset_bd(struct map*m,int iface,int i0, int i1, int dr)
{
  struct boundary**bd;
  assert(m);
  assert(iface>=0&&iface<6);
  assert(_bd_in_range(m,iface,i0,i1,dr));
  assert(dr==0||dr==1);
  bd = m->bd[iface] + (i0*m->brg[iface/2][1]*2+i1*2+dr);
#ifdef BD_DEBUG
  fprintf(logfile,
    " map.c:_unset_bd : refcount=%d,iface=%d,i0=%d,i1=%d,dr=%d\n",
    (*bd)?(*bd)->refcount:0,iface,i0,i1,dr);
#endif
  if(*bd)
  {
    assert((*bd)->refcount>0);
    if((*bd)->refcount==1)
    {
      /** OK, turn it off */
      scene_del_draw_item(m->s,(*bd)->d);
      (*bd)->d=NULL;
      xfree(*bd);
      *bd=NULL;
    }
    else
      (*bd)->refcount--;
  }
}

static
int
_cidx(int iface, int idx)
{
  int cidx[3][2] = 
  { 
    {1,2}, /** j and k */
    {0,2}, /** i and k */
    {0,1}  /** i and j */
  };
  assert(iface>=0&&iface<6);
  assert(idx==0||idx==1);
  return cidx[iface/2][idx];
}

static
void
_set_bd(struct map*m,int iface,int i0,int i1,int dr)
{
  struct boundary**bd;
  assert(m);
  assert(iface>=0&&iface<6);
  assert(_bd_in_range(m,iface,i0,i1,dr));
  assert(dr==0||dr==1);
  bd = m->bd[iface] + (i0*m->brg[iface/2][1]*2+i1*2+dr);
  if(*bd==NULL)
  {
    /** OK make line */
    float _x[2*3*2*3]=
    {
      0.0,0.0,0.0, 0.0,1.0,0.0, 
      0.0,0.0,0.0, 0.0,0.0,1.0, /** for both ifaces 0 and 1 */

      0.0,0.0,0.0, 1.0,0.0,0.0, 
      0.0,0.0,0.0, 0.0,0.0,1.0, /** iface 2 and 3 */

      0.0,0.0,0.0, 1.0,0.0,0.0, 
      0.0,0.0,0.0, 0.0,1.0,0.0  /** iface 4 and 5 */
    };
    int idx;
    float *x=_x+2*2*3*(iface/2);
    /** put line onto correct face */
    if( iface%2 )
      for(idx=0;idx<4;idx++)
        x[idx*3+iface/2]+=m->bsz[iface/2];
    /** now perform translation */
    for(idx=0;idx<4;idx++)
    {
      x[idx*3+_cidx(iface,0)]+=i0;
      x[idx*3+_cidx(iface,1)]+=i1;
    }
    /** OK we are ready */
    *bd=(struct boundary*)xmalloc(sizeof(struct boundary));
    (*bd)->d=scene_new_draw_item(
      m->s,x+6*dr,2,
      colour_lookup(BORDER_CID,0),
      render_line,IS_DYNAMIC);
    main_set_draw();
    (*bd)->refcount=1;
  }
  else
  {
    (*bd)->refcount++;
    assert((*bd)->refcount<127);
  }
#ifdef BD_DEBUG
  fprintf(logfile,
    " map.c:_set_bd : refcount=%d,iface=%d,i0=%d,i1=%d,dr=%d\n",
    (*bd)->refcount,iface,i0,i1,dr);
#endif
}

static
struct bcluster*
_alloc_init_bcluster(
  struct map*m,void*owner,int iface, int i0, int i1)
{
  int i,j;
  struct bcluster*bcl=
    (struct bcluster*)xmalloc(sizeof(struct bcluster));
  bcl->next=NULL;
  bcl->owner=owner;
  bcl->iface=iface;
  bcl->i0=i0;
  bcl->i1=i1;
  for(i=0;i<BCL_SZ;i++)
    for(j=0;j<BCL_SZ;j++)
    {
      bcl->mask[i][j][0]=!(rand()%5);
      bcl->mask[i][j][1]=!(rand()%5);
    }
  return bcl;
}

static
void
_free_bcluster(struct bcluster*bcl)
{
  xfree(bcl);
}

static
void
_set_bcluster(struct map*m,struct bcluster*bcl)
{
  int i0, i1,dr;
  for(i0=0;i0<BCL_SZ;i0++)
    for(i1=0;i1<BCL_SZ;i1++)
      for(dr=0;dr<2;dr++)
       if(bcl->mask[i0][i1][dr]
         &&_bd_in_range(
           m,bcl->iface,i0-BCL_SZ/2+bcl->i0,i1-BCL_SZ/2+bcl->i1,dr))
         _set_bd(m,bcl->iface,i0-BCL_SZ/2+bcl->i0,i1-BCL_SZ/2+bcl->i1,dr);
}

static
void
_unset_bcluster(struct map*m,struct bcluster*bcl)
{
  int i0, i1,dr;
  for(i0=0;i0<BCL_SZ;i0++)
    for(i1=0;i1<BCL_SZ;i1++)
      for(dr=0;dr<2;dr++)
       if(bcl->mask[i0][i1][dr]
         &&_bd_in_range(
           m,bcl->iface,i0-BCL_SZ/2+bcl->i0,i1-BCL_SZ/2+bcl->i1,dr))
         _unset_bd(m,bcl->iface,i0-BCL_SZ/2+bcl->i0,i1-BCL_SZ/2+bcl->i1,dr);
}

static
int _dr[6][3]=
{
  {-1,0,0},
  {1,0,0},
  {0,-1,0},
  {0,1,0},
  {0,0,-1},
  {0,0,1}
};

static
int
_dr_idx(int di,int dj,int dk)
{
  int idx;
  for(idx=0;idx<6;idx++)
    if( _dr[idx][0]==di&&_dr[idx][1]==dj&&_dr[idx][2]==dk )
      return idx;
  fprintf(logfile," bad _dr vector: %d,%d,%d\n",di,dj,dk);
  assert(0);
  return -69969;
}

static
int
_get_iface(struct map*m, int i, int j, int k, int di, int dj, int dk)
{
  int idx=_dr_idx(-di,-dj,-dk);
  if(i==0&&idx==0) 		return 0;
  if(i==m->bi-1&&idx==1) 	return 1;
  if(j==0&&idx==2) 		return 2;
  if(j==m->bj-1&&idx==3) 	return 3;
  if(k==0&&idx==4) 		return 4;
  if(k==m->bk-1&&idx==5) 	return 5;
  assert(0);
  return -969699;
}

static
int
_on_bd(struct map*m,int i,int j,int k,int di,int dj,int dk)
{
  if(i==0 || i==m->bi-1 || j==0 || j==m->bj-1 || k==0 || k==m->bk-1)
  {
    int idx=_dr_idx(-di,-dj,-dk);
    if(i==0&&idx==0) 		return 1;
    if(i==m->bi-1&&idx==1) 	return 1;
    if(j==0&&idx==2) 		return 1;
    if(j==m->bj-1&&idx==3)	return 1;
    if(k==0&&idx==4) 		return 1;
    if(k==m->bk-1&&idx==5) 	return 1;
  }
  return 0;
}

/** public */
void
map_notify_object(
  struct map*m,void*owner,struct ivec iv)
{
  int i=iv.x[0];
  int j=iv.x[1];
  int k=iv.x[2];
  int di=iv.d[2];
  int dj=iv.d[5];
  int dk=iv.d[8];
  struct bcluster*bcl;
  struct bcluster*_bcl=NULL;
  int iface, i0, i1;
  assert(m);
  assert(owner);
  assert(map_in_range(m,i,j,k));

  /** well, we can be a bit lazy about this... */
  if(rand()%3)
    return;

  if(_on_bd(m,i,j,k,di,dj,dk))
  {
    /** make new bcluster */
    int ci[3]={i,j,k};
    iface=_get_iface(m,i,j,k,di,dj,dk);
    i0=ci[_cidx(iface,0)];
    i1=ci[_cidx(iface,1)];
    _bcl=_alloc_init_bcluster(m,owner,iface,i0,i1);
    /** insert onto head of list */
    _bcl->next=m->bcl->next;
    m->bcl->next=_bcl;
    /** turn it on! */
    _set_bcluster(m,_bcl);
  }

  /** does this owner already have a bcluster? */
  for(bcl=m->bcl;bcl->next;bcl=bcl->next)
    if(bcl->next->owner==owner
      &&bcl->next!=_bcl/** we just made this one*/)
      break;

  if(bcl->next)
  {
    /** OK, unset old one */
    _unset_bcluster(m,bcl->next);
    /** remove it, free it */
    _bcl=bcl->next;
    bcl->next=_bcl->next;
    _free_bcluster(_bcl);
  }
}

/** public **/
void
map_notify_object_removal(struct map*m,void*owner)
{
  struct bcluster*bcl,*_bcl;
  /** does this owner have a bcluster? */
  for(bcl=m->bcl;bcl->next;bcl=bcl->next)
    if(bcl->next->owner==owner)
      break;

  if(bcl->next)
  {
    /** OK, unset it */
    _unset_bcluster(m,bcl->next);
    /** remove it, free it */
    _bcl=bcl->next;
    bcl->next=_bcl->next;
    _free_bcluster(_bcl);
  }
}

/************* init ************************/
  
void
map_init(struct map*m,struct scene*s,int bi,int bj,int bk)
{
  int idx;
  m->s=s;
  m->bi=bi;
  m->bj=bj;
  m->bk=bk;
  m->bsz[0]=bi;
  m->bsz[1]=bj;
  m->bsz[2]=bk;
  m->sq=(struct draw_item**)
    xmalloc(sizeof(struct draw_item*)*(bi+1)*(bj+1)*(bk+1)*3);
  for(idx=0;idx<(bi+1)*(bj+1)*(bk+1)*3;idx++)
    m->sq[idx]=NULL;
  m->el=(struct mapel*)xmalloc(sizeof(struct mapel)*bi*bj*bk);
  for(idx=0;idx<bi*bj*bk;idx++)
    m->el[idx]=__mel_empty();
  m->pa=(struct parcel*)xmalloc(sizeof(struct parcel));
  m->pa->next=NULL;
  m->n_parcels=0; /* no parcels */
  m->ho=(struct hole*)xmalloc(sizeof(struct hole));
  m->ho->next=NULL;
  /** initialize boundary indicators */
  {
    int fsz[3]={(bj+1)*(bk+1),(bi+1)*(bk+1),(bi+1)*(bj+1)};
    int brg[3][2]={ {bj+1,bk+1} , {bi+1,bk+1} , {bi+1,bj+1} };
    int iface;
    memcpy(m->fsz,fsz,sizeof(int)*3);
    memcpy(m->brg,brg,sizeof(int)*6);
    for(iface=0;iface<6;iface++)
    {
      m->bd[iface]=
        (struct boundary**)xmalloc(sizeof(struct boundary*)*fsz[iface/2]*2);
      for(idx=0;idx<fsz[iface/2]*2;idx++)
        m->bd[iface][idx]=NULL;
    }
    for(iface=0;iface<6&&0;iface++)
    {
      int i0, i1, on=0;
      /** fill some of the boundary lines */
      for(i0=0;i0<brg[iface/2][0];i0++)
        for(i1=0;i1<brg[iface/2][1];i1++)
        {
          if((!on&&rand()%16==0) || (on&&rand()%4==0))
            on=!on;
          if(on)
            {
              _set_bd(m,iface,i0,i1,0);
              _set_bd(m,iface,i0,i1,1);
            }
        }
    }
  }
  /** initialize cluster list */
  m->bcl=_alloc_init_bcluster(m,NULL,0,0,0); /** sentinel */
  /***** portal *****/
  m->piv=NULL;
  /** m->portal_on_sq=0; ** redundant */
}

void
map_finalize(struct map*m)
{
  int i;
  assert(m->ho);
  while(m->ho)
  {
    struct hole*ho=m->ho->next;
    xfree(m->ho);
    m->ho=ho;
  }
  assert(m->pa);
  while(m->pa)
  {
    struct parcel*pa=m->pa->next;
    xfree(m->pa);
    m->pa=pa;
  }
  xfree(m->sq);
  m->sq=NULL;
  xfree(m->el);
  m->el=NULL;
  for(i=0;i<6;i++)
  {
    int idx;
    for(idx=0;idx<m->fsz[i/2]*2;idx++)
      if( m->bd[i][idx] )
      {
        assert( m->bd[i][idx]->refcount > 0 );
        xfree(m->bd[i][idx]);
      }
    xfree(m->bd[i]);
    m->bd[i]=NULL;
  }
  while(m->bcl)
  {
    struct bcluster*bcl=m->bcl->next;
    _free_bcluster(m->bcl);
    m->bcl=bcl;
  }
  if(m->piv)
    xfree(m->piv);
  m->piv=NULL;
}

struct mapel*
_mel(struct map*m,int i,int j,int k)
{
  /** yay Mel! */
  return m->el+k+j*m->bk+i*m->bk*m->bj;
}

void
map_print(struct map*m)
{
  int i,j,k;
  for(i=0;i<m->bi;i++)
  {
    for(j=0;j<m->bj;j++)
    {
      for(k=0;k<m->bk;k++)
        if(map_is_brick(m,i,j,k))
          printf("#");
        else
          printf("0");
      printf("\n");
    }
    printf("\n");
  }
}

/****************** query ************************/

void
map_compile(struct map*m)
{
  int i=0,j=0,k=0,l=0,siz=0;
  int*n;
  /** these two vars index _dr */
  int align; /** which way is my head pointing? */
  int goal; /** which way do i want to go? */
  for(i=0;i<m->bi;i++)
  for(j=0;j<m->bj;j++)
  for(k=0;k<m->bk;k++)
  {
    if( map_is_brick(m,i-1,j,k)
     || map_is_brick(m,i+1,j,k)
     || map_is_brick(m,i,j-1,k)
     || map_is_brick(m,i,j+1,k)
     || map_is_brick(m,i,j,k-1)
     || map_is_brick(m,i,j,k+1))
      siz++;
    for(align=0;align<6;align++)
      for(goal=0;goal<6;goal++)
        _mel(m,i,j,k)->u[align][goal]=0; /** no solution */
  }
#ifdef DEBUG_MAP
  fprintf(logfile,"compile siz=%d\n",siz);
#endif
  n=(int*)xmalloc(sizeof(int)*siz*3);
  l=0;
  for(i=0;i<m->bi;i++)
  for(j=0;j<m->bj;j++)
  for(k=0;k<m->bk;k++)
  {
    assert(l<siz);
    if( map_is_brick(m,i-1,j,k)
     || map_is_brick(m,i+1,j,k)
     || map_is_brick(m,i,j-1,k)
     || map_is_brick(m,i,j+1,k)
     || map_is_brick(m,i,j,k-1)
     || map_is_brick(m,i,j,k+1))
    {
      n[3*l]=i;
      n[3*l+1]=j;
      n[3*l+2]=k;
      l++;
    }
  }
  for(l=0;l<siz;l++)
  {
    for(align=0;align<6;align++)
    {
      i=n[3*l]; j=n[3*l+1]; k=n[3*l+2];
      /** check that this align is valid */
      if(map_is_brick(m,i-_dr[align][0],j-_dr[align][1],k-_dr[align][2]))
      {
        /** okay, we are on solid ground */
        for(goal=0;goal<6;goal++)
        {
          /** try going in direction di,dj,dk */
          /** a cyclic permutation is orthogonal */
          int di=_dr[align][1],dj=_dr[align][2],dk=_dr[align][0];
          int t;
          for(t=0;t<4;t++)
          {
            int _di,_dj,_dk; /** temps */
            /** do we flip? */
            if(map_is_brick(m,i+di,j+dj,k+dk))
            {
              if(goal==align)
              {
                /** want to go up */
                /** okay, go this way! */
                _mel(m,i,j,k)->u[align][goal]|=1<<_dr_idx(di,dj,dk);
              }
            }
            else
            /** do we fall? */
            if(!map_is_brick(m,
              i+di-_dr[align][0],j+dj-_dr[align][1],k+dk-_dr[align][2]))
            {
              if(  ((align%2) && goal==align-1)
                || (!(align%2) && goal==align+1))
                /** want to go down, ok! */
                _mel(m,i,j,k)->u[align][goal]|=1<<_dr_idx(di,dj,dk);
            }
            else
            {
              /** ok, can walk this way, do we want to? */
              if(goal==_dr_idx(di,dj,dk))
                _mel(m,i,j,k)->u[align][goal]|=1<<_dr_idx(di,dj,dk);
            }
            /* turn */
            _di=_dr[align][1]*dk-_dr[align][2]*dj;
            _dj=_dr[align][2]*di-_dr[align][0]*dk;
            _dk=_dr[align][0]*dj-_dr[align][1]*di;
            di=_di;
            dj=_dj;
            dk=_dk;
          } /* for t<4 */
          /** assert we are back where we started */
          assert(di==_dr[align][1]);
          assert(dj==_dr[align][2]);
          assert(dk==_dr[align][0]);
        } /* for goal */
      } /* if valid align */
    } /* for align */
  } /* for l */
  xfree(n);
}

int
sgn(int j)
{if(j>0)return 1;if(j<0)return -1;return 0;}

/** how do we get from s_iv to t_iv ? */
struct ivec
map_query(
  struct map*m,
  struct ivec s_iv,
  struct ivec t_iv )
{
  struct ivec iv; /** return value */
  iv=s_iv; 
  /*** chug chug **/
  if(ivec_xcmp(s_iv,t_iv))
  {
    /** we haven't reached our target */
    int do_something=1; /** assume we need to change direction */
    int sol[6],solc=0; /** all solutions */
    int dx[3],goal,align,n,l;
    align=_dr_idx(s_iv.d[2],s_iv.d[5],s_iv.d[8]); /** the up vector */
    dx[0]=sgn(t_iv.x[0]-s_iv.x[0]);
    dx[1]=sgn(t_iv.x[1]-s_iv.x[1]);
    dx[2]=sgn(t_iv.x[2]-s_iv.x[2]);
    for(n=0;n<3&&do_something;n++)
      if(dx[n])
      {
        if(n==0) goal=_dr_idx(dx[0],0,0);
        else if(n==1) goal=_dr_idx(0,dx[1],0);
        else goal=_dr_idx(0,0,dx[2]);
        for(l=0;l<6&&do_something;l++)
          if
          (
            (1<<l)
            &_mel(m,s_iv.x[0],s_iv.x[1],s_iv.x[2])->u[align][goal]
          )
          {
            if( align==l ) 
              /** OK, we are on track */
              do_something=0;
            else
              sol[solc++]=l; /** remember solution */
          }
      }
    if(do_something)
    {
      /** okay, is there a solution ? */
      if(solc)
      {
        int k=0;
        if(solc>1 && 0 /** be decisive! */ )
        {
          k=rand()%(solc-1);
          fprintf(logfile,"map found %d solutions\n",solc);
        }
        /** great, go this way then */
        iv.d[0]=_dr[sol[k]][0];
        iv.d[3]=_dr[sol[k]][1];
        iv.d[6]=_dr[sol[k]][2]; /** iv.d[0,3,6] is forward */
        /** iv.d[2,5,8] is up, is preserved */
        /** iv.d[1,4,7] is left=up cross forward */
        iv.d[1]=iv.d[5]*iv.d[6]-iv.d[8]*iv.d[3];
        iv.d[4]=iv.d[8]*iv.d[0]-iv.d[2]*iv.d[6];
        iv.d[7]=iv.d[2]*iv.d[3]-iv.d[5]*iv.d[0];
      }
    }
  }
  assert(ivec_xcmp(iv,s_iv)==0);
#ifdef DEBUG_MAP
  fprintf(logfile,"old ivec ------------*\n");
  ivec_print(s_iv);
  fprintf(logfile,"new ivec-------------*\n");
  ivec_print(iv);
  fprintf(logfile,"---------------------*\n");
#endif
  return iv;
}

/************************* sq *********************/

int
_sq_in_range(struct map*m,int i,int j,int k, int l)
{
  return 
    l>=0&&l<3 && k>=0 
    && j>=0 && i>=0  
    && k<m->bk+1 && j<m->bj+1 
    && i<m->bi+1;
}

struct draw_item**
_get_sq(struct map*m, int i, int j, int k, int l)
{
  assert(_sq_in_range(m,i,j,k,l));
  return m->sq+l+k*3+j*(m->bk+1)*3+i*(m->bk+1)*(m->bj+1)*3;
}

int
_is_sq(struct map*m,int i, int j,int k,int l)
{
  assert(_sq_in_range(m,i,j,k,l));
  return *_get_sq(m,i,j,k,l)!=NULL;
}

void
_make_sq_coords(struct map*m,int i,int j,int k,int l,float x[12*3])
{
  int _i;
  float _x[12*3]=
  {
    /** the three orientations of a square, specified by l */
    0.0,0.0,0.0,
    0.0,1.0,0.0,
    0.0,0.0,1.0,
    0.0,1.0,1.0,

    0.0,0.0,0.0,
    0.0,0.0,1.0,
    1.0,0.0,0.0,
    1.0,0.0,1.0,

    0.0,0.0,0.0,
    0.0,1.0,0.0,
    1.0,0.0,0.0,
    1.0,1.0,0.0,
  };
  memcpy(x,_x,sizeof(float)*12*3);
  /* translate it */
  for(_i=0;_i<12;_i++)
  {
    x[_i*3]+=i;
    x[_i*3+1]+=j;
    x[_i*3+2]+=k;
  }
}

void
_set_sq(struct map*m, int i, int j, int k, int l,int solid)
{
  struct draw_item**dp;
  float x[12*3];
  assert(_sq_in_range(m,i,j,k,l));
  dp=_get_sq(m,i,j,k,l);
  assert(*dp==NULL);
  _make_sq_coords(m,i,j,k,l,x);
  /* make it */
  *dp=scene_new_draw_item(m->s,x+l*12,4,
    colour_lookup(solid?SOLID_CID:BRICK_CID,l),
    render_quad,IS_STATIC); 
}

void
_unset_sq(struct map*m, int i, int j, int k, int l)
{
  struct draw_item**dp;
  assert(_sq_in_range(m,i,j,k,l));
  dp=_get_sq(m,i,j,k,l);
  assert(*dp);
  scene_del_draw_item(m->s,*dp);
  *dp=NULL;
}

void
_toggle_sq(struct map*m, int i, int j, int k, int l,int solid)
{
  assert(_sq_in_range(m,i,j,k,l));
  if(_is_sq(m,i,j,k,l))
    _unset_sq(m,i,j,k,l);
  else
    _set_sq(m,i,j,k,l,solid);
}

/*********************************************/

void
map_cull_backfaces(struct map*m,struct cam*c)
{
  int idx;
  struct mapel*el=m->el;
  for(idx=0; idx<m->bi*m->bj*m->bk; idx++,el++)
    if(el->b&BRICK_EL)
    {
      int i,j,k;
      int iface;
      i=idx/( m->bj*m->bk );
      j=( idx-i*m->bj*m->bk )/m->bk;
      k=idx-i*m->bj*m->bk-j*m->bk;
      for(iface=0;iface<6;iface++)
      {
        if( _dr[iface][0]*(c->s[0]-i)
          +_dr[iface][1]*(c->s[1]-j)
          +_dr[iface][2]*(c->s[2]-k) < 0.0 )
        {
          struct draw_item**dp;
          if(iface%2)
          {
            if(iface==1) i++;
            if(iface==3) j++;
            if(iface==5) k++;
            iface--;
            dp=m->sq+iface/2+k*3+j*(m->bk+1)*3+i*(m->bk+1)*(m->bj+1)*3;
            if(*dp)
              scene_set_draw_item_skip(m->s,*dp,1);
            iface++;
            if(iface==1) i--;
            if(iface==3) j--;
            if(iface==5) k--;
          }
          else
          {
            dp=m->sq+iface/2+k*3+j*(m->bk+1)*3+i*(m->bk+1)*(m->bj+1)*3;
            if(*dp)
              scene_set_draw_item_skip(m->s,*dp,1);
          }
        }
      }
    }
}

int
map_in_range(struct map*m,int i, int j, int k)
{
  assert(m);
  if(i<0||j<0||k<0)
    return 0;
  if(i>=m->bi||j>=m->bj||k>=m->bk)
    return 0;
  return 1;
}

/*************** lock ******************/

int
map_is_lock(struct map*m,int i, int j, int k)
{
  assert(m);
  if(!map_in_range(m,i,j,k))
    return 1; /** outside of the world is locked */
  return m->el[ k + j*m->bk + i*m->bk*m->bj ].l;
}

int n_locks=0;
void
map_set_lock(struct map*m,int i, int j, int k)
{
  assert(m);
  if(!map_in_range(m,i,j,k))
    return;
  assert(map_is_brick(m,i,j,k));
  m->el[k+j*m->bk+i*m->bk*m->bj].l++;
  /** mmm, is this correct? */
  assert(m->el[k+j*m->bk+i*m->bk*m->bj].l<=6); 
  assert(map_is_lock(m,i,j,k));
  n_locks++;
#ifdef DEBUG_MAP
  fprintf(
    logfile,
    "map_set_lock %d %d %d, total=%d\n",
    i,j,k,n_locks);
#endif
}

void
map_unset_lock(struct map*m,int i, int j, int k)
{
  assert(m);
  if(!map_in_range(m,i,j,k))
    return;
  assert(map_is_brick(m,i,j,k));
  assert(map_is_lock(m,i,j,k));
  m->el[k+j*m->bk+i*m->bk*m->bj].l--;
  n_locks--;
  assert(n_locks>=0);
#ifdef DEBUG_MAP
  fprintf(logfile,"map_unset_lock %d %d %d, total=%d\n",i,j,k,n_locks);
#endif
}

/************ brick *************************/

int
map_is_brick(struct map*m,int i, int j, int k)
{
  assert(m);
  if(!map_in_range(m,i,j,k))
    return 1; /* a world encased in brick */
  return m->el[k+j*m->bk+i*m->bk*m->bj].b & BRICK_EL;
}

int
map_is_solid(struct map*m,int i, int j, int k)
{
  assert(m);
  if(!map_in_range(m,i,j,k))
    return 0;
  return map_is_brick(m,i,j,k)
    &&m->el[k+j*m->bk+i*m->bk*m->bj].b & SOLID_EL;
}

void map_set_brick(struct map*m,int i,int j,int k,int solid)
{
  assert(m);
  assert(map_in_range(m,i,j,k));
  assert(!map_is_brick(m,i,j,k));
  assert(!map_is_lock(m,i,j,k));
#define _DEBUG_MAP
#ifdef DEBUG_MAP
  printf("map_set_brick(%d,%d,%d,%d)\n",i,j,k,solid);
#endif  
  {
    int u[3]={i,j,k},l;
    for(l=0;l<3;l++)
    {
      _toggle_sq(m,u[0],u[1],u[2],l,solid);
      u[l]++;
      _toggle_sq(m,u[0],u[1],u[2],l,solid);
      u[l]--;
    }
/*
    _toggle_sq(m,i,j,k,0,solid);
    _toggle_sq(m,i,j,k,1,solid);
    _toggle_sq(m,i,j,k,2,solid);
    _toggle_sq(m,i+1,j,k,0,solid);
    _toggle_sq(m,i,j+1,k,1,solid);
    _toggle_sq(m,i,j,k+1,2,solid);
*/
    m->el[k+j*m->bk+i*m->bk*m->bj].b |= BRICK_EL;
    if(solid)
    {
      m->el[k+j*m->bk+i*m->bk*m->bj].b |= SOLID_EL;
/**
      map_set_lock(m,i,j,k);
**/
    }
  }
}

void map_unset_brick(struct map*m,int i,int j,int k)
{
  assert(m);
  assert(map_in_range(m,i,j,k));
  assert(map_is_brick(m,i,j,k));
  assert(!map_is_lock(m,i,j,k));
  {
    int u[3]={i,j,k},l;
    for(l=0;l<3;l++)
    {
      int solid;
      u[l]--;
      solid=map_is_solid(m,u[0],u[1],u[2]);
      u[l]++;
      _toggle_sq(m,u[0],u[1],u[2],l,solid);
      u[l]++;
      solid=map_is_solid(m,u[0],u[1],u[2]);
      _toggle_sq(m,u[0],u[1],u[2],l,solid);
      u[l]--;
    }
/*
    _toggle_sq(m,i,j,k,0,0); 
    _toggle_sq(m,i,j,k,1,0);
    _toggle_sq(m,i,j,k,2,0);
    _toggle_sq(m,i+1,j,k,0,0);
    _toggle_sq(m,i,j+1,k,1,0);
    _toggle_sq(m,i,j,k+1,2,0);
*/
    m->el[k+j*m->bk+i*m->bk*m->bj].b &= ~BRICK_EL;
    m->el[k+j*m->bk+i*m->bk*m->bj].b &= ~SOLID_EL;
  }
}

/*************** hole *******************/

int
map_is_hole(struct map*m,int i, int j,int k)
{
  return( m->el[k+j*m->bk+i*m->bk*m->bj].b&HOLE_EL );
}

#ifdef USEC
void
map_set_hole(struct map*m,int i, int j, int k,struct timeval tv)
{
  struct hole*ho;
  assert(m);
  map_unset_brick(m,i,j,k);
  main_set_draw();
  assert(!map_is_hole(m,i,j,k));
  m->el[k+j*m->bk+i*m->bk*m->bj].b |= HOLE_EL;
  ho=(struct hole*)xmalloc(sizeof(struct hole));
  ho->i=i;
  ho->j=j;
  ho->k=k;
  ho->tv=tv;
  ho->tv.tv_usec+=HOLE_FRAMES*usec_per_frame;
  ho->tv.tv_sec+=(ho->tv.tv_usec/1000000);
  ho->tv.tv_usec%=1000000;
  ho->next=m->ho->next;
  m->ho->next=ho;
}

/** deal with holes filling in */
void
map_poll(struct map*m,struct timeval tv)
{
  struct hole*ho;
  for(ho=m->ho;ho->next;ho=ho->next)
    if(timercmp(&ho->next->tv,&tv,<))
    {
      /* time is up, fill the hole */
      struct hole*_ho;
      _ho=ho->next;
      ho->next=ho->next->next;
      assert(map_is_hole(m,_ho->i,_ho->j,_ho->k));
      m->el[_ho->k+_ho->j*m->bk+_ho->i*m->bk*m->bj].b -= HOLE_EL;
      map_set_brick(m,_ho->i,_ho->j,_ho->k,0);
      main_set_draw();
      xfree(_ho);
      if(!ho->next)
        break;
    }
}
#else /* USEC */
void
map_set_hole(struct map*m,int i, int j, int k,int msec)
{
  struct hole*ho;
  assert(m);
  map_unset_brick(m,i,j,k);
  main_set_draw();
  assert(!map_is_hole(m,i,j,k));
  m->el[k+j*m->bk+i*m->bk*m->bj].b |= HOLE_EL;
  ho=(struct hole*)xmalloc(sizeof(struct hole));
  ho->i=i;
  ho->j=j;
  ho->k=k;
  ho->msec=msec+HOLE_FRAMES*msec_per_frame;
  ho->next=m->ho->next;
  m->ho->next=ho;
}

/** deal with holes filling in */
void
map_poll(struct map*m,int msec)
{
  struct hole*ho;
  for(ho=m->ho;ho->next;ho=ho->next)
    if( ho->next->msec<msec )
    {
      /* time is up, fill the hole */
      struct hole*_ho;
      _ho=ho->next;
      ho->next=ho->next->next;
      assert(map_is_hole(m,_ho->i,_ho->j,_ho->k));
      m->el[_ho->k+_ho->j*m->bk+_ho->i*m->bk*m->bj].b -= HOLE_EL;
      map_set_brick(m,_ho->i,_ho->j,_ho->k,0);
      main_set_draw();
      xfree(_ho);
      if(!ho->next)
        break;
    }
}
#endif /* USEC */

/**************** parcel ************************/

int
map_is_parcel(struct map*m,int i, int j, int k)
{
  assert(m);
  if(!map_in_range(m,i,j,k))
  {
    assert(0);
    return 0;
  }
  return m->el[k+j*m->bk+i*m->bk*m->bj].b & PARCEL_EL;
}

void
map_set_parcel(struct map*m,int i,int j,int k)
{
  float x[3*8]=
  {
    0.0, 0.0, 0.0,
    0.0, 1.0, 0.0,
    0.0, 0.0, 1.0,
    0.0, 1.0, 1.0,
    1.0, 0.0, 0.0,
    1.0, 1.0, 0.0,
    1.0, 0.0, 1.0,
    1.0, 1.0, 1.0,
  };
  struct parcel*pa;
  int col, idx;
  assert(m);
  assert(map_in_range(m,i,j,k));
  assert(!map_is_parcel(m,i,j,k));
  m->el[k+j*m->bk+i*m->bk*m->bj].b |= PARCEL_EL;
  pa=(struct parcel*)xmalloc(sizeof(struct parcel));
  pa->i=i;
  pa->j=j;
  pa->k=k;
  for(idx=0;idx<3*8;idx++)
  {
    /** scale down */
    x[idx]*=0.4;
    /** center */
    x[idx]+=0.3;
  }
  for(idx=0;idx<8;idx++)
  {
    x[3*idx]+=i;
    x[3*idx+1]+=j;
    x[3*idx+2]+=k;
  }
/*
  col=rand()%256;
  fprintf(logfile,"parcel col=%d\n", col);
*/
  col=70;
  pa->d=scene_new_draw_item(m->s,x,8,
    colour_lookup(PARCEL_CID,0),
    render_parcel,IS_STATIC);
  pa->next=m->pa->next;
  m->pa->next=pa;
  m->n_parcels++;
}

void
map_unset_parcel(struct map*m,int i,int j,int k)
{
  struct parcel*pa;
  assert(m);
  assert(map_in_range(m,i,j,k));
  assert(map_is_parcel(m,i,j,k));
  m->el[k+j*m->bk+i*m->bk*m->bj].b -= PARCEL_EL;
  for(pa=m->pa;pa->next;pa=pa->next)
    if( pa->next->i==i&&pa->next->j==j&&pa->next->k==k )
    {
      struct parcel*_pa=pa->next;
      pa->next=pa->next->next;
      scene_del_draw_item(m->s,_pa->d);
      xfree(_pa);
      m->n_parcels--;
      assert(m->n_parcels>=0);
      return;
    }
  assert(0); /** could not find parcel */
}

void
map_toggle_parcel(struct map*m,int i, int j, int k)
{
  if(map_is_brick(m,i,j,k))
  {
    printf("not going to embed parcel in a brick");
    printf(" %d,%d,%d\n",i,j,k);
    return;
  }
  if(map_is_parcel(m,i,j,k))
    map_unset_parcel(m,i,j,k);
  else
    map_set_parcel(m,i,j,k);
}

int
map_n_parcels(struct map*m)
{
  return m->n_parcels;
}

/**************** portal ***************/

int
map_is_portal(struct map*m,struct ivec iv)
{
  int i=iv.x[0];
  int j=iv.x[1];
  int k=iv.x[2];
  assert(m);
#ifdef DEBUG_MAP
  fprintf(logfile,"map_is_portal\n");
#endif
  if(!map_in_range(m,i,j,k))
  {
    ivec_print(iv);
    assert(0);
    return 0;
  }
  if( m->piv )
  {
    int r=!ivec_ucmp(*m->piv,iv);
    fprintf(logfile,"map_is_portal r=%d\n",r);
    return r;
  }
  else
    return 0;
}

void
map_unset_portal(struct map*m,struct ivec iv)
{
  int i=iv.x[0];
  int j=iv.x[1];
  int k=iv.x[2];
  int di=-iv.d[2]; /** the    */
  int dj=-iv.d[5]; /** down   */
  int dk=-iv.d[8]; /** vector */
  int iface;
  assert(m);
#ifdef DEBUG_MAP
  fprintf(logfile,"map_unset_portal:\n");
  ivec_print(iv);
#endif
  assert(map_in_range(m,i,j,k));
  assert(map_is_portal(m,iv));
  if(map_in_range(m,i+di,j+dj,k+dk))
  {
    assert(map_is_lock(m,i+di,j+dj,k+dk));
    map_unset_lock(m,i+di,j+dj,k+dk);
  }
  iface=_dr_idx(di,dj,dk);
  if(iface==1) i++;
  if(iface==3) j++;
  if(iface==5) k++;
  _unset_sq(m,i,j,k,iface/2);
  xfree(m->piv);
  m->piv=NULL;
  if(iface==1) i--;
  if(iface==3) j--;
  if(iface==5) k--;
  if(!map_is_brick(m,i,j,k)&&map_in_range(m,i+di,j+dj,k+dk))
  {
    if(iface==1) i++;
    if(iface==3) j++;
    if(iface==5) k++;
    _set_sq(m,i,j,k,iface/2,0);  /* FIXME: solid? */
  }
#ifdef DEBUG_MAP
  else
    fprintf(logfile,"removing hidden portal\n");
#endif
}

void
map_set_portal(struct map*m,struct ivec iv)
{
  int i=iv.x[0];
  int j=iv.x[1];
  int k=iv.x[2];
  int di=-iv.d[2];
  int dj=-iv.d[5];
  int dk=-iv.d[8]; /** down vector **/
  int iface;
  struct draw_item**dp;
  float x[12*3];
  assert(m);
#ifdef DEBUG_MAP
  fprintf(logfile,"map_set_portal:\n");
  ivec_print(iv);
#endif
  assert(map_in_range(m,i,j,k));
  if(!map_is_brick(m,i+di,j+dj,k+dk))
    /** must have portal on a brick */
    return;
  if(map_is_brick(m,i,j,k))
    /** no hidden portals */
    return;
  if(map_in_range(m,i+di,j+dj,k+dk))
    /** lock the brick that supports this portal */
    map_set_lock(m,i+di,j+dj,k+dk);
  iface=_dr_idx(di,dj,dk);
  if(iface==1) i++;
  if(iface==3) j++;
  if(iface==5) k++;
  if(m->piv)
    /** remove old portal */
    map_unset_portal(m,*m->piv);
  /** make new portal */
  m->piv=(struct ivec*)xmalloc(sizeof(struct ivec));
  *m->piv=iv;
  assert(_sq_in_range(m,i,j,k,iface/2));
  dp=_get_sq(m,i,j,k,iface/2);
  if(*dp)
    /** remove offending square */
    _unset_sq(m,i,j,k,iface/2);
  assert(*dp==NULL);
  _make_sq_coords(m,i,j,k,iface/2,x);
  /** make it */
  *dp=scene_new_draw_item(
    m->s,x+(iface/2)*12,4,
    colour_lookup(PORTAL_CID,0),
    render_quad,IS_STATIC);
}

void
map_toggle_portal(struct map*m,struct ivec iv)
{
  if(map_is_portal(m,iv))
    map_unset_portal(m,iv);
  else
    map_set_portal(m,iv);
}

/*****************************************/

float _center[3];
float*
map_get_center(struct map*m)
{
  _center[0]=((float)m->bi)/2.0;
  _center[1]=((float)m->bj)/2.0;
  _center[2]=((float)m->bk)/2.0;
  return _center;
}

