#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include "render.h"
#include "globals.h"
#include "memory.h"

int _malloced=0;
int mark;

#define MEM_MAGIC 0xFEF5
#define MEM_FREED 0xDDE1

#ifdef MEMORY_CHECK
void*
xmalloc(size_t s)
{
  int*p=(int*)malloc(s+2*sizeof(int));
  if (p==NULL)
  {
    render_finalize();
    fprintf(logfile,"malloc bomb.\n");
    assert(0);
    exit(EXIT_FAILURE);
  }
  p[0]=MEM_MAGIC;
  assert(p[0]==MEM_MAGIC);
  p[1]=s;
  _malloced+=s;
#ifdef DEBUG_MEM
  fprintf(logfile,"malloc(%d)=%p, total=%d\n",s,p,_malloced);
#endif
  return (void*)(p+2);
}

void*
xcalloc(size_t s)
{
  int*p=(int*)calloc(1,s+2*sizeof(int));
  if(p==NULL)
  {
    render_finalize();
    fprintf(logfile,"malloc bomb.\n");
    assert(0);
    exit(EXIT_FAILURE);
  }
  p[0]=MEM_MAGIC;
  assert(p[0]==MEM_MAGIC);
  p[1]=s;
  _malloced+=s;
#ifdef DEBUG_MEM
  fprintf(logfile,"malloc(%d)=%p, total=%d\n",s,p,_malloced);
#endif
  return (void*)(p+2);
}

void*
xrealloc(void*p,size_t s)
{
  p=((int*)p)-2;
  assert(((int*)p)[0]==MEM_MAGIC);
  _malloced-=((int*)p)[1];
  p=realloc(p,s+2*sizeof(int));
  if (p==NULL)
  {
    render_finalize();
    fprintf(logfile,"realloc bomb.\n");
    assert(0);
    exit(EXIT_FAILURE);
  }
  ((int*)p)[1]=s;
  _malloced+=s;
#ifdef DEBUG_MEM
  fprintf(logfile,"realloc(?,%d)=%p, total=%d\n",s,p,_malloced);
#endif
  p=((int*)p)+2;
  return p;
}

void
xmark()
{
  mark=_malloced;
#ifdef DEBUG_MEM
  fprintf(logfile," memory.c xmark() : %d malloced\n", _malloced );
#endif
}

int
xcheck()
{
#define _DEBUG_MEM
#ifdef DEBUG_MEM
  fprintf(logfile," memory.c xcheck() : %d malloced\n", _malloced );
#endif
  return _malloced-mark;
}

void
xfree(void*p)
{
  p=((int*)p)-2;
#ifdef DEBUG_MEM
  fprintf(logfile,"free(%p)",p);
  fflush(logfile);
#endif
  if(((int*)p)[0]!=MEM_MAGIC)
  {
    render_finalize();
    fprintf(logfile,"xfree bomb.\n");
    if(((int*)p)[0]==MEM_FREED)
      fprintf(logfile,"mem freed!\n");
    assert(0);
    exit(EXIT_FAILURE);
  }
  _malloced-=((int*)p)[1];
  ((int*)p)[0]=MEM_FREED;
#ifdef DEBUG_MEM
  fprintf(logfile," -%d, total=%d\n",((int*)p)[1],_malloced);
#endif
  free(p);
}

#endif
