#ifdef PC_SPEAKER
#include <sys/ioctl.h>
#include <linux/kd.h>
#endif
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <assert.h>
#include "globals.h"
#include "scene.h"
#include "people.h"
#include "render.h"
#include "main.h"
#include "frame.h"
#include "colour.h"
#include "mtime.h"
#include "audio.h"

#define _DEBUG_TASK

#define MIVECX(iv)	(iv).x[0],(iv).x[1],(iv).x[2]
#define MIVEC(iv)	(iv).x[0],(iv).x[1],(iv).x[2],(iv).d[2],iv.d[5],iv.d[8]

/** private */
static
void
_clear_redraw(struct person*pn)
{
  pn->redraw=0;
}

/** private */
static
void
_set_redraw(struct person*pn)
{
  pn->redraw=1;
}

static
void
_mark_free(struct person*pn)
{
  pn->free=1;
#ifdef DEBUG_TASK
  fprintf(logfile," pn %p is marked\n",pn);
#endif
}

static
void
__make_pn_coords(struct person*pn,float x[18])
{
  int i;
  struct ivec iv;
  assert(pn->frame-_pn_walk<3*N_FRAMES*PN_POINTS);
  assert(pn->frame-_pn_walk>=0);
  memcpy(x,pn->frame,sizeof(float)*3*6);
  iv=ivec_interpolate(
    pn->iv,
    pn->_iv,
    pn->toff/(float)(pn->ta->next->ti->toff));
#ifdef DEBUG_PEOPLE
  fprintf(logfile,"--------->__make_pn_coords\n");
  ivec_print(iv);
  fprintf(logfile,"<---------\n");
#endif
  for(i=0;i<6;i++)
    ivec_operate(iv,x+3*i);
  _clear_redraw(pn);
}

void
pn_print(struct person*pn)
{
  struct task * ta;
  struct lock*lk;
  fprintf(logfile,"  %p: ",pn);
  if(pn->is_hero)
    fprintf(logfile, "hero ");
  else
    fprintf(logfile, "robot ");
  fprintf(logfile, "%.0f,%.0f,%.0f", MIVECX(pn->iv));
  if(pn->lk)
  {
    fprintf(logfile," lk:[");
    for(lk=pn->lk->next;lk;lk=lk->next)
      fprintf(logfile, "{%d,%d,%d} ", lk->i,lk->j,lk->k );
    fprintf(logfile,"]");
  }
  for(ta=pn->ta->next;ta;ta=ta->next)
  {
    fprintf(logfile,"[");
    if(ta->stat==PN_STAT_ALLOC)
      fprintf(logfile,"#");
    if(ta->stat==PN_STAT_INIT||ta->stat==PN_STAT_SEND)
      fprintf(logfile,">");
    fprintf(logfile,"%s",ta->ti->name);
    if(ta->stat==PN_STAT_FINALIZE||ta->stat==PN_STAT_SEND)
      fprintf(logfile,"<");
    fprintf(logfile,"] ");
  }
  fprintf(logfile,"\n");
  ivec_print(pn->iv);
  ivec_print(pn->_iv);
}

static
int
_is_lock(struct person*pn,int i, int j, int k)
{
  struct lock*lk;
  assert(pn);
  assert(pn->lk);
  for(lk=pn->lk;lk->next;lk=lk->next)
    if( lk->next->i==i && lk->next->j==j && lk->next->k==k )
    {
      assert(map_is_lock(pn->m,i,j,k));
      return 1;
    }
  return 0;
}

static
void
_set_lock(struct person*pn,int i,int j,int k)
{
  struct lock*lk;
  assert(pn);
  assert(pn->lk);
  lk=(struct lock*)xmalloc(sizeof(struct lock));
  lk->i=i;
  lk->j=j;
  lk->k=k;
  lk->next=pn->lk->next;
  pn->lk->next=lk;
  map_set_lock(pn->m,i,j,k);
  assert(_is_lock(pn,i,j,k));
}

static
void
_unset_lock(struct person*pn,int i, int j, int k)
{
  struct lock*lk;
  assert(pn);
  assert(pn->lk);
  for(lk=pn->lk;lk->next;lk=lk->next)
    if( lk->next->i==i && lk->next->j==j && lk->next->k==k )
    {
      /** bingo */
      struct lock*_lk=lk->next;
      map_unset_lock(pn->m,_lk->i,_lk->j,_lk->k);
      lk->next=lk->next->next;
      xfree(_lk);
      assert(!_is_lock(pn,i,j,k));
      return;
    }
  fprintf(logfile," %p could not find lock on brick %d %d %d\n",
    pn,i,j,k);
  assert(0); /** bungo */
}

static
int
_n_locks(struct person*pn)
{
  int n=0;
  struct lock*lk;
  assert(pn);
  assert(pn->lk);
  for(lk=pn->lk;lk->next;lk=lk->next)
    n++;
  return n;
}

static
struct person*
_is_robot_at(struct people*pl,struct ivec iv)
{
  int i;
  for(i=1;i<MAX_PERSON;i++)
    if(pl->pn[i])
      if(ivec_xcmp(pl->pn[i]->iv,iv)==0
/***
        || ivec_xcmp(pn->pn[i]->_iv,iv)==0
***/
      )
        return pl->pn[i];
  return NULL;
}

static
int
_is_hero_at(struct people*pl,struct ivec iv)
{
  if(pl->pn[0])
    if(ivec_xcmp(pl->pn[0]->iv,iv)==0)
      return 1;
  return 0;
}

static
void
_check_parcel(struct person*pn)
{
  if(
    map_is_parcel(pn->m,MIVECX(pn->_iv))
    && pn->toff>pn->ta->next->ti->toff/3 )
  {
    if(pn->is_hero||!pn->n_parcels)
    {
      /** acquire parcel */
      pn->n_parcels++;
      map_unset_parcel(pn->m,MIVECX(pn->_iv));
      if(pn->is_hero)
        audio_start_parcel();
    }
  }
}

static
void
_robot_command(struct person*);

/******       *********       ******       *******/
/******       *********       ******       *******/
/******       *********       ******       *******/
/******          task stuff                *******/
/******       *********       ******       *******/

struct task_info pn_task_stop=
  {
    PN_TOFF_STOP, 
    pn_init_stop,
    pn_send_stop, 
    pn_finalize_stop, 
    pn_request_stop,
    "stop",
  };

struct task_info pn_task_walk=
  {
    PN_TOFF_WALK, 
    pn_init_walk,
    pn_send_walk,
    pn_finalize_walk,
    pn_request_walk,
    "walk",
  };
struct task_info pn_task_flip=
  {
    PN_TOFF_FLIP, 
    pn_init_flip, pn_send_flip, pn_finalize_flip, pn_request_default,
    "flip",
  };
struct task_info pn_task_turn_left=
  {
    PN_TOFF_TURN, 
    pn_init_turn_left, pn_send_turn_left, pn_finalize_turn_left,
    pn_request_default,
    "turn left",
  };
struct task_info pn_task_turn_right=
  {
    PN_TOFF_TURN, /** time to turn left == time to turn right */
    pn_init_turn_right, pn_send_turn_right, pn_finalize_turn_right, 
    pn_request_default,
    "turn right",
  };
struct task_info pn_task_turn_back=
  {
    PN_TOFF_TURN,
    pn_init_turn_back, pn_send_turn_back, pn_finalize_turn_back, 
    pn_request_default,
    "turn back",
  };
struct task_info pn_task_dig=
  {
    PN_TOFF_DIG, 
    pn_init_dig, pn_send_dig, pn_finalize_dig, pn_request_default,
    "dig",
  };
struct task_info pn_task_fall=
  {
    PN_TOFF_FALL, 
    pn_init_fall, pn_send_fall, pn_finalize_fall, pn_request_default,
    "fall",
  };
struct task_info pn_task_stuck=
  {
    PN_TOFF_STUCK, 
    pn_init_stuck, pn_send_stuck, pn_finalize_stuck, pn_request_default,
    "stuck",
  };
struct task_info pn_task_climb=
  {
    PN_TOFF_CLIMB, 
    pn_init_climb, pn_send_climb, pn_finalize_climb, pn_request_default,
    "climb",
  };

int
_is_stuck_robot(struct people*pl,struct ivec iv)
{
  struct person*pn=_is_robot_at(pl,iv);
  if(pn)
  {
    assert(pn->ta);
    assert(pn->ta->next);
    assert(pn->ta->next->ti);
    if(pn->ta->next->ti==&pn_task_stuck)
      return 1;
  }
  return 0;
}

/** returns NULL if empty */
struct task*
_pop_task(struct task*ta)
{
  struct task*_ta;
  assert(ta);
  _ta=ta->next;
  if(_ta)
    ta->next=ta->next->next;
  return _ta;
}

struct task*
_alloc_task(struct task_info*ti)
{
  struct task*ta=xmalloc(sizeof(struct task));
  ta->ti=ti;
  ta->next=NULL;
  ta->stat=PN_STAT_ALLOC;
  return ta;
}

void
_free_task(struct task*ta)
{
  assert(ta);
  assert(ta->stat==PN_STAT_FINALIZE||ta->stat==PN_STAT_ALLOC);
  xfree(ta);
}

void
_push_task(struct task*ta,struct task_info*ti)
{
  struct task*_ta;
  assert(ta);
  assert(ti);
  _ta=_alloc_task(ti);
  _ta->next=ta->next;
  ta->next=_ta;
}

/** guaranteed to return 1 */
int
_request_task(struct person*pn,struct task*ta, struct task_info*ti)
{
  struct task*_ta;
  assert(ta);
  for(_ta=ta;_ta->next;_ta=_ta->next)
  {
    assert(_ta->next->ti);
    assert(_ta->next->ti->request);
    if((*_ta->next->ti->request)(pn,_ta,ti))
      /** OK request granted */
      return 1;
  }
  /** OK we stick this one on the end */
  _ta->next=_alloc_task(ti);
  return 1;
} 

static
void
__pn_frame_inc(struct person*,int);

int
_send_check_bury(struct person*pn)
{
  if(map_is_brick(pn->m,MIVECX(pn->iv)))
  {
     /** we been buried */
     pn->ta->next->stat=PN_STAT_SEND;
     fprintf(logfile," %p is buried\n",pn);
     _mark_free(pn);
     return 1;
  }
  return 0;
}

int
_finalize_check_bury(struct person*pn)
{
  assert(pn);
  assert(!pn->is_hero);
  if(map_is_brick(pn->m,MIVECX(pn->iv)))
  {
    /***** we been buried *******/
    struct ivec iv;
    if(pn->pl->iv==NULL)
    {
      /** there is no robot portal! */
      _mark_free(pn);
      return 1;
    }
    fprintf(logfile," resetting robot %p\n",pn);
    pn->iv=*pn->pl->iv;
    pn->_iv=pn->iv;
    iv=ivec_below(pn->iv);
    if(map_is_brick(pn->m,MIVECX(iv)))
      _push_task(pn->ta->next,&pn_task_walk);
    else
      _push_task(pn->ta->next,&pn_task_fall);
    pn->ta->next->stat=PN_STAT_FINALIZE;
    return 1;
  }
  return 0;
}

/*******************************************************/
/*******************************************************/
/************     task routines     ********************/
/************                       ********************/

void
pn_init_default(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti);
  pn->toff=0;
/*
  pn->_iv=pn->iv;
*/
  pn->ta->next->stat=PN_STAT_INIT;
}

void
_send_lock(struct person*pn)
{
  struct ivec iv;
  assert(pn);
  assert(pn->lk);
  iv=ivec_below(pn->iv);
  if(map_is_brick(pn->m,MIVECX(iv)))
    if(!_is_lock(pn,MIVECX(iv)))
      _set_lock(pn,MIVECX(iv));
  iv=ivec_below(pn->_iv);
  if(map_is_brick(pn->m,MIVECX(iv)))
    if(!_is_lock(pn,MIVECX(iv)))
      _set_lock(pn,MIVECX(iv));
  assert(_n_locks(pn)<=2);
}

int
pn_send_default(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti);
  assert(pn->ta->next->stat==PN_STAT_INIT
    || pn->ta->next->stat==PN_STAT_SEND );
  pn->ta->next->stat=PN_STAT_SEND;
  if(pn->toff<pn->ta->next->ti->toff)
  {
    pn->toff++;
    return 1; /** did something */
  }
  else
    return 0; /** nothing to do */
}

void
pn_finalize_default(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti);
  assert(pn->ta->next->stat==PN_STAT_SEND);
  pn->iv=pn->_iv;
/*
  pn->toff=0;
*/
  pn->ta->next->stat=PN_STAT_FINALIZE;
}

void
pn_finalize_goal(struct person*pn)
{
  if(pn->toff==pn->ta->next->ti->toff)
  {
    /** we made it */
    if(!pn->is_hero)
    {
      struct ivec iv,_iv;
      iv=ivec_below(pn->iv);
      _iv=ivec_below(pn->_iv);
      if(ivec_xcmp(iv,_iv))
        /** release old lock */
        if(map_is_brick(pn->m,MIVECX(iv)))
          _unset_lock(pn,MIVECX(iv));
    }
    pn_finalize_default(pn); 
  }
  else
  {
    /** ... something else has come up */
    if(!pn->is_hero)
    {
      struct ivec iv,_iv;
      iv=ivec_below(pn->iv);
      _iv=ivec_below(pn->_iv);
      if(ivec_xcmp(iv,_iv))
        if(map_is_brick(pn->m,MIVECX(_iv)))
          _unset_lock(pn,MIVECX(_iv));
      pn->_iv=pn->iv;
    }
    pn->ta->next->stat=PN_STAT_FINALIZE;
  }
}

void
pn_finalize_continue(struct person*pn)
{
  struct ivec iv;
  /** continue walking this way... */
  iv=ivec_below(pn->iv);
  /** are we on solid ground, or on a robot? */
  if(map_is_brick(pn->m,MIVECX(iv))
    ||_is_stuck_robot(pn->pl,iv))
  {
    /** OK, let's move this way */
    iv=ivec_forward(pn->iv);
    if( map_is_brick(pn->m,MIVECX(iv))
      || ( !pn->is_hero&&map_is_hole(pn->m,MIVECX(iv))))
    {
      /** flip up */
      _request_task(pn,pn->ta,&pn_task_flip);
    }
    else
    {
      /** all clear */
      _request_task(pn,pn->ta,&pn_task_walk);
    }
  }
  else
  {
    /** whoa, fell of the edge */
    _request_task(pn,pn->ta,&pn_task_fall);
    /** queue walking for when we finish the fall **/
    _request_task(pn,pn->ta,&pn_task_walk);
  }
}

int
_finalize_check_fall(struct person*pn)
{
  struct ivec iv;
  iv=ivec_below(pn->iv);
  if(!map_is_brick(pn->m,MIVECX(iv))
    &&!_is_stuck_robot(pn->pl,iv))
  {
    _push_task(pn->ta->next,&pn_task_fall);
    return 1;
  }
  return 0;
}

void
pn_finalize_motion(struct person*pn)
{
  _finalize_check_fall(pn);
  if(pn->ta->next->next)
    return; /** we already have something to do... */
  if(!pn->is_hero)
  {
    _robot_command(pn);
    if(pn->ta->next->next)
      return; /** OK we have something to do... */
  }
  pn_finalize_continue(pn);
}

/** return 1 if request is handled */
int
pn_request_default(
  struct person*pn,
  /** requestee */
  struct task*ta,
  /** request name */
  struct task_info*ti )
{
#ifdef DO_THIS_BIT
  if( ta->next->ti == ti )
    return 1; /** swallow ourself; contraction */
#endif
  return 0; /** can't handle it */
}

/************                       ********************/
/************                       ********************/
/*******************************************************/
/************ stop task routines    ********************/
/************                       ********************/

void
pn_init_stop(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_stop);
  pn_init_default(pn);
  if(!pn->is_hero)
  {
    struct ivec iv=ivec_below(pn->iv);
    assert(_is_lock(pn,MIVECX(iv)));
  }
  map_notify_object(pn->m,pn,pn->iv);
}

int
pn_send_stop(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_stop);
  assert(pn->ta->next->stat==PN_STAT_INIT
    || pn->ta->next->stat==PN_STAT_SEND );
  if(_send_check_bury(pn))
    return 0;
  pn->ta->next->stat=PN_STAT_SEND;
  if( !pn->ta->next->next )
  {
    if(!pn->is_hero)
      _robot_command(pn);
    return 1;
  }
  else
    return 0; /** we have something to do next */
}

/** this is confusing: it's the finalize routine of the stop task */
void
pn_finalize_stop(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_stop);
  pn_finalize_default(pn);
  if(!pn->is_hero)
    if(_finalize_check_bury(pn))
      return;
}

/** NB: this is not a request to stop! */
int
pn_request_stop(
  struct person*pn, struct task*ta, struct task_info*ti )
{
#ifdef STOP_CONSUME_WALK
  if( ti==&pn_task_walk )
    return 1;
#endif
  if( !pn_request_default(pn,ta,ti))
    _push_task(ta->next,ti);
  return 1;
}

/************                       ********************/
/************                       ********************/
/*******************************************************/
/************ walk task routines    ********************/
/************                       ********************/

void
pn_init_walk(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_walk);
  pn_init_default(pn);
  pn->_iv=ivec_forward(pn->iv);
  if( !pn->is_hero )
    _send_lock(pn);
  map_notify_object(pn->m,pn,pn->iv);
}

int
pn_send_walk(struct person*pn)
{
  struct ivec iv;
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_walk);
  if(_send_check_bury(pn))
    return 0;
  if(map_is_brick(pn->m,MIVECX(pn->_iv)))
  {
    /** can't walk this way */
    pn->ta->next->stat=PN_STAT_SEND;
    return 0;
  }
  _check_parcel(pn);
  if(!pn->is_hero)
    _send_lock(pn);
#define SERIOUS_HACKAGE
#ifdef SERIOUS_HACKAGE
  /** try stop robots from being in the same place at the same time */
  if( !pn->is_hero 
    && 
    (
      _is_robot_at(pn->pl,pn->_iv)
      || (_is_robot_at(pn->pl,pn->iv) && _is_robot_at(pn->pl,pn->iv)!=pn)
    )
    && rand()%3==0 /** hack, hack */ )
  {
    /** can't walk this way */
    pn->ta->next->stat=PN_STAT_SEND;
    return 0;
  }
#endif
#undef SERIOUS_HACKAGE
  iv=ivec_below(pn->iv);
  if( map_is_brick(pn->m,MIVECX(iv)) /** walking on brick */
    || _is_stuck_robot(pn->pl,iv)  /** walking on a robot */ 
  )
  {
    /** OK */
  }
  else
  if(!pn->is_hero)
  {
    /** assert( robot just climbed out of hole ) **/
    fprintf(logfile," NOTE: %p just climbed out of a hole.\n", pn );
  }
  else
  {
    fprintf(logfile," WARNING: %p walking on a live robot?\n", pn );
    printf(" WARNING: %p walking on a live robot?\n", pn );
  } 
  /** assert we are not in a brick */
  assert(!map_is_brick(pn->m,MIVECX(pn->iv)));
  if(pn_send_default(pn))
  {
    __pn_frame_inc(pn,1);
    return 1;
  }
  return 0;
}

void
pn_finalize_walk(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_walk
    ||pn->ta->next->ti==&pn_task_flip );
  if(!pn->is_hero)
    if(_finalize_check_bury(pn))
      return;
  pn_finalize_goal(pn);
  pn_finalize_motion(pn);
  if(!pn->is_hero)
    if(_finalize_check_bury(pn))
      return;
}

/** return 1 if request is handled */
int
pn_request_walk(
  struct person*pn,
  /** requestee */
  struct task*ta,
  /** request name */
  struct task_info*ti )
{
#ifdef DO_THIS_BIT
  if( ta->next->ti == ti )
    return 1; /** swallow ourself; ala contraction */
#endif
  
  return 0; /** can't handle it */
}

/************                       ********************/
/************                       ********************/
/*******************************************************/
/************ flip task routines    ********************/
/************                       ********************/

void
pn_init_flip(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_flip);
  pn_init_default(pn);
  pn->_iv=ivec_turn_up(pn->iv);
  if( !pn->is_hero )
    _send_lock(pn);
}

int
pn_send_flip(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_flip);
  if(_send_check_bury(pn))
    return 0;
  if(!pn->is_hero)
    _send_lock(pn);
  if(pn_send_default(pn))
  {
    __pn_frame_inc(pn,1);
    return 1;
  }
  return 0;
}

void
pn_finalize_flip(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_flip);
  pn_finalize_walk(pn); /** works aswell */
}

/************                       ********************/
/************                       ********************/
/*******************************************************/
/************ turn left task routines ******************/
/************                       ********************/

void
pn_init_turn_left(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_turn_left);
  pn_init_default(pn);
  pn->_iv=ivec_turn_left(pn->iv);
}

int
pn_send_turn_left(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_turn_left);
  if(_send_check_bury(pn))
    return 0;
  if(!pn->is_hero)
    _send_lock(pn);
  if(pn_send_default(pn))
  {
    __pn_frame_inc(pn,1);
    return 1;
  }
  return 0;
}

void
pn_finalize_turn_left(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_turn_left);
  pn_finalize_goal(pn);
  pn_finalize_motion(pn);
  if(!pn->is_hero)
    if(_finalize_check_bury(pn))
      return;
/*
  if(!pn->is_hero)
    _robot_command(pn);
*/
}

/************                       ********************/
/************                       ********************/
/*******************************************************/
/************ turn right task routines *****************/
/************                       ********************/

void
pn_init_turn_right(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_turn_right);
  pn_init_default(pn);
  pn->_iv=ivec_turn_right(pn->iv);
}

int
pn_send_turn_right(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_turn_right);
  if(_send_check_bury(pn))
    return 0;
  if(!pn->is_hero)
    _send_lock(pn);
  if(pn_send_default(pn))
  {
    __pn_frame_inc(pn,1);
    return 1;
  }
  return 0;
}

void
pn_finalize_turn_right(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_turn_right);
  pn_finalize_goal(pn);
  pn_finalize_motion(pn);
  if(!pn->is_hero)
    if(_finalize_check_bury(pn))
      return;
/*
  if(!pn->is_hero)
    _robot_command(pn);
*/
}

/************                       ********************/
/************                       ********************/
/*******************************************************/
/************ turn back task routines *****************/
/************                       ********************/

void
pn_init_turn_back(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_turn_back);
  pn_init_default(pn);
  pn->_iv=ivec_turn_back(pn->iv);
}

int
pn_send_turn_back(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_turn_back);
  if(_send_check_bury(pn))
    return 0;
  if(!pn->is_hero)
    _send_lock(pn);
  if(pn_send_default(pn))
  {
    __pn_frame_inc(pn,1);
    return 1;
  }
  return 0;
}

void
pn_finalize_turn_back(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_turn_back);
  pn_finalize_goal(pn);
  pn_finalize_motion(pn);
  if(!pn->is_hero)
    if(_finalize_check_bury(pn))
      return;
/*
  if(!pn->is_hero)
    _robot_command(pn);
*/
}

/************                       ********************/
/************                       ********************/
/*******************************************************/
/************ dig task routines     ********************/
/************                       ********************/

void
pn_init_dig(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_dig);
  assert(pn->is_hero);
  pn_init_default(pn);
  pn->_iv=pn->iv;
  /** TODO: init dig animation */
  audio_start_dig();
}

int
pn_send_dig(struct person*pn)
{
  struct ivec iv;
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_dig);
  assert(pn->is_hero);
  assert(pn->ta->next->stat==PN_STAT_INIT
    ||pn->ta->next->stat==PN_STAT_SEND);
  if(_send_check_bury(pn))
    return 0;
  iv=ivec_forward(pn->iv);
  if(map_is_brick(pn->m,MIVECX(iv)))
  {
    pn->ta->next->stat=PN_STAT_SEND;
    return 0; /* can't dig here */
  }
  iv=ivec_below(ivec_forward(pn->iv));
  if(!map_is_brick(pn->m,MIVECX(iv)))
  {
    pn->ta->next->stat=PN_STAT_SEND;
    return 0; /* no brick to dig */
  }
  if(map_is_lock(pn->m,MIVECX(iv)))
  {
    pn->ta->next->stat=PN_STAT_SEND;
    return 0; /* can't dig a locked brick */
  }
  if(map_is_solid(pn->m,MIVECX(iv)))
  {
    pn->ta->next->stat=PN_STAT_SEND;
    return 0; /* can't dig a solid brick */
  }
  if(pn_send_default(pn))
  {
    /** TODO: dig animation */
    return 1;
  }
  return 0;
}

void
pn_finalize_dig(struct person*pn)
{
  struct ivec iv;
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_dig);
  assert(pn->is_hero);
  /** TODO: dig animation */
  if(pn->toff==pn->ta->next->ti->toff)
  {
    /** hole was completed */
    iv=ivec_forward(pn->iv);
    iv=ivec_below(iv);
    assert(map_is_brick(pn->m,MIVECX(iv)));
    if(!map_is_lock(pn->m,MIVECX(iv)))
#ifdef USEC
      map_set_hole(pn->m,MIVECX(iv),*pn->tv);
#else
      map_set_hole(pn->m,MIVECX(iv),pn->msec);
#endif
  }
  pn_finalize_default(pn);
  audio_stop_dig();
}

/************                       ********************/
/************                       ********************/
/*******************************************************/
/************ fall task routines    ********************/
/************                       ********************/

void
pn_init_fall(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_fall);
  pn_init_default(pn);
  pn->_iv=ivec_below(pn->iv);
  if( pn->is_hero )
    audio_start_fall();
  assert(!map_is_brick(pn->m,MIVECX(pn->_iv)));
}

int
pn_send_fall(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_fall);
  if(map_is_hole(pn->m,MIVECX(pn->_iv))&&!pn->is_hero&&pn->n_parcels)
  {
    /** robot with parcel is falling into hole; drop parcel */
    assert(!map_is_parcel(pn->m,MIVECX(pn->iv)));
    map_set_parcel(pn->m,MIVECX(pn->iv));
    pn->n_parcels--;
    assert(pn->n_parcels>=0);
  }
  else
  {
    if(!map_is_hole(pn->m,MIVECX(pn->_iv))||pn->is_hero)
      _check_parcel(pn);
    if(_send_check_bury(pn))
      return 0;
  }    
  if(pn_send_default(pn))
  {
/* if pn->is_hero send_fall_sound */
    _set_redraw(pn);
    return 1;
  }
  return 0;
}

void
pn_finalize_fall(struct person*pn)
{
  assert(pn);
  assert(pn->ta);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_fall);
  if(!pn->is_hero)
    if(_finalize_check_bury(pn))
      return;
  pn_finalize_default(pn);
  if(!pn->is_hero)
    if(_finalize_check_bury(pn))
      return;
  if(map_is_hole(pn->m,MIVECX(pn->iv))&&!pn->is_hero)
    /** robot is stuck in hole */
    _push_task(pn->ta->next,&pn_task_stuck);
  else
  {
    struct ivec iv;
    iv=ivec_below(pn->iv);
    if(map_is_brick(pn->m,MIVECX(iv))||_is_stuck_robot(pn->pl,iv))
    {
      struct task*ta;
      /** we have landed */
      if( pn->is_hero )
        audio_stop_fall();
      if(!pn->is_hero)
      {
        _set_lock(pn,MIVECX(iv));
        /***** flush all tasks ****/
        ta=pn->ta->next;
        assert(ta);
        if(ta->next)
          while(ta->next)
            _free_task(_pop_task(ta));
        /*** work out what to do next ***/
        _robot_command(pn);
      }
      else
      {
/*
        if(pn->is_hero)
          _finalize_fall_sound();
*/

        /** flush everything except for the last request */
        ta=pn->ta->next;
        assert(ta);
        if(ta->next)
          while(ta->next->next)
            _free_task(_pop_task(ta));
      }
    }
    else
      /** continue falling */
      _push_task(pn->ta->next,&pn_task_fall);
  }
  map_notify_object(pn->m,pn,pn->iv);
}

/************                       ********************/
/************                       ********************/
/*******************************************************/
/************ stuck task routines   ********************/
/************                       ********************/

void
pn_init_stuck(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_stuck);
  assert(!pn->is_hero);
  pn_init_default(pn);
  pn->_iv=pn->iv;
#ifdef LOCK_BRICK_WHEN_STUCK
  {
    struct ivec iv;
    iv=ivec_below(pn->iv);
    if(map_is_brick(pn->m,MIVECX(iv)))
      _set_lock(pn,MIVECX(iv));
  }
#endif
  /** flush the task buffer */
  while(pn->ta->next->next)
    _free_task(_pop_task(pn->ta->next));
}

int
pn_send_stuck(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_stuck);
  assert(!pn->is_hero);
#ifdef DONT_WAIT_FOR_STUCK_WHEN_SWALLOWED
  if(_send_check_bury(pn))
    return 0;
#endif
  return pn_send_default(pn);
}

void
pn_finalize_stuck(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_stuck);
  assert(!pn->is_hero);
  pn_finalize_default(pn);
#ifdef LOCK_BRICK_WHEN_STUCK
  {
    struct ivec iv;
    iv=ivec_below(pn->iv);
    if(map_is_brick(pn->m,MIVECX(iv)))
      _unset_lock(pn,MIVECX(iv));
  }
#endif
  if( !_finalize_check_bury(pn))
    _push_task(pn->ta->next,&pn_task_climb);
}

/************                       ********************/
/************                       ********************/
/*******************************************************/
/************ climb task routines   ********************/
/************                       ********************/

void
pn_init_climb(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_climb);
  assert(!pn->is_hero);
  pn_init_default(pn);
  pn->_iv=ivec_above(pn->iv);
}

int
pn_send_climb(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_climb);
  assert(!pn->is_hero);
  if(_send_check_bury(pn))
    return 0;
  if(pn_send_default(pn))
  {
    __pn_frame_inc(pn,1);
    return 1;
  }
  return 0;
}

void
pn_finalize_climb(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti==&pn_task_climb);
  assert(!pn->is_hero);
  if(!_finalize_check_bury(pn)) /** untested check **/
  {
    pn_finalize_default(pn);
    if(!pn->ta->next->next)
      _robot_command(pn);
    _check_parcel(pn);
  }
}

/************                      *********************/
/***       ** end of task routines ***       ***********/
/*******************************************************/
/***       **********       **********       ***********/
/***       **********       **********       ***********/

/* private */
static
void
__init_pn_state(struct person*pn)
{
  struct ivec iv;
  iv=ivec_below(pn->iv);
  if(map_is_brick(pn->m,MIVECX(iv)))
  {
    if(!pn->is_hero)
      _set_lock(pn,MIVECX(iv));
    _push_task(pn->ta,&pn_task_stop);
  }
  else
    _push_task(pn->ta,&pn_task_fall);
  /** init task */
  assert(pn->ta->next);
  (*pn->ta->next->ti->init)(pn);
  _set_redraw(pn);
}


/* private */
static
void
_init_pn(
  struct person*pn,
  struct scene*s,
  struct map*m,
  struct people*pl,
  struct ivec iv,
  int is_hero,
  int col, int msec )
{
  float x[18];
  assert(pn);
  assert(s);
  assert(m);
  pn->s=s;
  pn->m=m;
  pn->pl=pl;
  pn->is_hero=is_hero;
  pn->n_parcels=0;
  pn->frame=_pn_walk;
#ifdef USEC
  pn->tv=NULL; /** no time yet */
#else
  pn->msec=msec;
  pn->msec=0;
#endif
  pn->toff=0;
  pn->iv=iv;
  pn->_iv=iv;
  pn->free=0;
  pn->ta=(struct task*)xmalloc(sizeof(struct task));
  pn->ta->next=NULL;
  pn->ta->ti=NULL; /** sentinel */
  if(!pn->is_hero)
  {
    pn->lk=(struct lock*)xmalloc(sizeof(struct lock));
    pn->lk->next=NULL; /** sentinel */
  }
  else
    pn->lk=NULL;
  _clear_redraw(pn);
  __init_pn_state(pn);
  __make_pn_coords(pn,x);
#ifdef PERSON_DEBUG
  print_x(x,6,3);
#endif
  pn->d=scene_new_draw_item(
    s,
    x,
    6   /* points */,
    col /* colour */,
    render_person,
    IS_DYNAMIC);
}

static
int
_is_hero_caught(struct person*pn)
{
  assert(pn);
  assert(pn->is_hero);
  {
    int i;
    for(i=1;i<MAX_PERSON;i++)
      if(pn->pl->pn[i])
      {
        struct person*_pn=pn->pl->pn[i];
        struct ivec iv, _iv;
        iv=ivec_interpolate(
          pn->iv,
          pn->_iv,
          pn->toff/(float)(pn->ta->next->ti->toff));
        _iv=ivec_interpolate(
          _pn->iv,
          _pn->_iv,
          _pn->toff/(float)(_pn->ta->next->ti->toff));
        if( ivec_xmetric(iv,_iv) < 0.7 )
        {
          _mark_free(pn); /** hero is finished */
          return 1;
        }
      }
  }
  return 0;
}

static
void
_check_hero_finished(struct person*pn)
{
#ifdef DEBUG_PEOPLE
  fprintf(logfile,"check_hero_finished\n");
#endif
  assert(pn->is_hero);
  if( map_n_parcels(pn->m) == 0 )
    if( map_is_portal(pn->m,pn->iv))
    {
      int i;
      for(i=1;i<MAX_PERSON;i++)
        if(pn->pl->pn[i])
          if(pn->pl->pn[i]->n_parcels)
            return; /** a robot still has a parcel... */
      printf("check_hero_finished: yes\n");
      main_set_level_done();
    }
#ifdef DEBUG_PEOPLE
  else
    fprintf(logfile,"more parcels to get!\n");
#endif
} 

/** private */
static
int
_pn_inc(struct person*pn)
{
  assert(pn);
  assert(pn->ta->next);
  assert(pn->ta->next->ti);
  assert(pn->ta->next->ti->send);
#ifdef DEBUG_TASK
  pn_print(pn);
#endif
  while(!(*pn->ta->next->ti->send)(pn)&&!pn->free)
  {
#ifdef DEBUG_TASK
    pn_print(pn);
#endif  
    /** okay finalized task */
    assert(pn->ta->next->ti->finalize);
    (*pn->ta->next->ti->finalize)(pn);
#ifdef DEBUG_TASK
    pn_print(pn);
#endif
    /** consume task */
    _free_task(_pop_task(pn->ta));
#ifdef DEBUG_TASK
    pn_print(pn);
#endif
    if(pn->free)
      break;
    if(!pn->ta->next)
    {
#ifdef DEBUG_TASK
      fprintf(logfile,"pn %p stopping here\n",pn);
#endif
      _request_task(pn,pn->ta,&pn_task_stop);
    }
    assert(pn->ta->next);
    assert(pn->ta->next->ti);
#ifdef DEBUG_TASK
    pn_print(pn);
#endif
    assert(pn->ta->next->ti->init);
    /** init next task */
    (*pn->ta->next->ti->init)(pn);
    assert(pn->ta->next->ti->send);
#ifdef DEBUG_TASK
    pn_print(pn);
#endif
  }
#ifdef DEBUG_TASK
  pn_print(pn);
#endif
  if(pn->is_hero)
  {
    if(_is_hero_caught(pn))
      return 0;
    _check_hero_finished(pn);
  }
  if(pn->free)
    return 0;
  return 1; /** we had a time consuming send */
}

static
void _del_person(struct people*,struct person*);

static
void
__pn_frame_inc(struct person*pn,int f_inc)
{
  assert(pn);
  pn->frame+=f_inc*PN_POINTS*3;
  while(pn->frame - _pn_walk  >= N_FRAMES*3*PN_POINTS )
    pn->frame-=N_FRAMES*PN_POINTS*3;
  while(pn->frame - _pn_walk < 0 )
    pn->frame+=N_FRAMES*PN_POINTS*3;
  _set_redraw(pn);
}

#ifdef USEC
static
int
person_poll(struct person*pn,struct timeval*tv)
{
  assert(pn);
  assert(tv);
  if( pn->tv )
  {
    /*** how many frames do we progress ? ***/
    int f_inc;
#define PN_TIMERS
#ifdef PN_TIMERS
    int pn_usec_inc;
    int pn_sec_inc;
    f_inc=
        ((tv->tv_sec-pn->tv->tv_sec)*1000000
         +(tv->tv_usec-pn->tv->tv_usec))
        /usec_per_frame;
    assert(f_inc>=0);
    if( f_inc == 0 )
      return 1; /* nothing changes */

    /*** increment time by one unit ***/
    if(!pn->is_hero)
      pn_usec_inc=1*usec_per_frame*ROBOT_LAG;
    else
      pn_usec_inc=1*usec_per_frame;
    pn_sec_inc=pn_usec_inc/1000000;
    pn_usec_inc%=1000000;

    pn->tv->tv_usec+=pn_usec_inc;
    pn->tv->tv_sec+=pn_sec_inc+(pn->tv->tv_usec/1000000);
    pn->tv->tv_usec%=1000000;
#else
    *(pn->tv)=*tv;
    f_inc=1;
#endif
    if(_pn_inc(pn))
    {
      /** pn still valid */
      /**map_poll(pn->m,*pn->tv);**/ /** slight overkill */
      if( f_inc == 1 )
        return 1; /** person up to date */
      else
        return 0; /** person needs more polling */
    }
    if( pn->free )
    {
      if( pn->is_hero || !pn->pl->iv )
      {
        _del_person(pn->pl,pn);
        if( pn->is_hero )
          main_set_caught();
      }
      else
        _del_person(pn->pl,pn);
    }
    return 1; /** we are done with this person */
  }
  else
  {
    /**** okay, we start here ***/
    pn->tv=xmalloc(sizeof(struct timeval));
    *pn->tv=*tv;
    assert(pn->toff==0);
    return 1; /** person up to date */
  }
}

#else /* USEC */

static
int
person_poll(struct person*pn,int msec)
{
  /*** how many frames do we progress ? ***/
  int f_inc;
  assert(pn);
  if(pn->msec==0)
  {
    /* ok start here */
    pn->msec=msec;
    return 1;
  }
  f_inc=(msec-pn->msec)/msec_per_frame;
  assert(f_inc>=0);
  if( f_inc == 0 )
    return 1; /* nothing changes */

  /*** increment time by one unit ***/
  if(!pn->is_hero)
    pn->msec+=1*msec_per_frame*ROBOT_LAG;
  else
    pn->msec+=1*msec_per_frame;

  if(_pn_inc(pn))
  {
    /** pn still valid */
    /**map_poll(pn->m,*pn->tv);**/ /** slight overkill */
    if( f_inc == 1 )
      return 1; /** person up to date */
    else
      return 0; /** person needs more polling */
  }
  if( pn->free )
  {
    if( pn->is_hero || !pn->pl->iv )
    {
      _del_person(pn->pl,pn);
      if( pn->is_hero )
        main_set_caught();
    }
    else
      _del_person(pn->pl,pn);
  }
  return 1; /** we are done with this person */
}
#endif /* USEC */

/** return 0 when we don't need to be called again */
int
update_people_draw(struct people*pl)
{
  int i;
  for(i=0;i<MAX_PERSON;i++)
  {
    if(pl->pn[i])
      if(pl->pn[i]->redraw)
      {
        float x[3*6];
        int col=  colour_lookup(
          ((i==0)?HERO_CID:ROBOT_CID),0);
        __make_pn_coords(pl->pn[i],x);
        scene_update_draw_item(
          pl->pn[i]->s,pl->pn[i]->d,x,col,render_person);
        return 1;
      }
  }
  return 0;
}

/** keep on polling the persons until they are up to date */
void
#ifdef USEC
update_people_state(struct people*pl,struct timeval*tv)
#else
update_people_state(struct people*pl,int msec)
#endif
{
  int i;
  int done;
  int pdone[MAX_PERSON];
  assert(pl);
#ifdef USEC
  assert(tv);
#endif
  for(i=0;i<MAX_PERSON;i++)
    pdone[i]=0;
  for(i=0;i<MAX_PERSON;i++)
    if(pl->pn[i])
      assert((pl->pn[i]->is_hero!=0)==(i==0));
#ifdef USEC
  map_poll(pl->m,*tv);
#else
  map_poll(pl->m,msec);
#endif
  for(done=0;!done;)
  {
    done=1;
    for(i=0;i<MAX_PERSON;i++)
      if( pl->pn[i] && !pdone[i] )
      {
#ifdef USEC
        pdone[i]=person_poll(pl->pn[i],tv);
#else
        pdone[i]=person_poll(pl->pn[i],msec);
#endif
        if(!pdone[i])
          done=0;
      }
  }
}

float _target[3];

float*
pn_get_target(struct people*pl)
{
  if(pl->pn[0])
    /** look at the hero */
    return pl->pn[0]->iv.x;
  else
  {
    _target[0]=pl->m->bi/2;
    _target[1]=pl->m->bj/2;
    _target[2]=pl->m->bk/2;
    return _target;
  }
}

/* private */
void
_finalize_person(struct people*pl,struct person*pn)
{
  int idx;
  assert(pl);
  assert(pn);
#ifdef DEBUG_TASK
  fprintf(logfile,"deleting pn \n");
#endif
  scene_del_draw_item(pn->s,pn->d);
  pn->d=NULL;
  while(pn->ta->next)
    xfree(_pop_task(pn->ta));
  xfree(pn->ta);
  while(pn->lk)
  {
    struct lock*lk=pn->lk->next;
    if(lk)
      map_unset_lock(pn->m,lk->i,lk->j,lk->k);
    xfree(pn->lk);
    pn->lk=lk;
  }
#ifdef USEC
  if( pn->tv )
    xfree(pn->tv);
#endif
  for(idx=0;idx<MAX_PERSON;idx++)
    if(pl->pn[idx]==pn)
      pl->pn[idx]=NULL;
}

/** private */
void
_del_person(struct people*pl,struct person*pn)
{
  assert(pl);
  assert(pn);
  map_notify_object_removal(pn->m,pn);
  _finalize_person(pl,pn);
  xfree(pn);
}

/******************* public routines *******************/

void
people_init(struct people*pl,struct scene*s,struct map*m)
{
  int idx;
  pl->s=s;
  pl->m=m;
/*
  pl->hl=NULL;
*/
  for(idx=0;idx<MAX_PERSON;idx++)
    pl->pn[idx]=NULL;
  pl->iv=NULL;
}

void
_robot_command(struct person*pn)
{
  struct ivec iv;
  assert(pn);
  assert(!pn->is_hero);
  assert(!pn->ta->next->next);
  if(pn->pl->pn[0])
  {
    /** okay, find the hero! */
    iv=map_query(pn->m,pn->iv,pn->pl->pn[0]->iv);
    if(ivec_cmp(iv,pn->iv)
#ifdef THIS_WAY
      && pn->ta->next->ti!=&pn_task_walk
      && pn->ta->next->ti!=&pn_task_stuck
#else
      ||pn->ta->next->ti==&pn_task_stop
      ||pn->ta->next->ti==&pn_task_climb
#endif
      )
    {
      /** better do something! */
      if(!ivec_cmp(iv,ivec_turn_left(pn->iv)))
        /* OK turn left */
        _request_task(pn,pn->ta,&pn_task_turn_left);
      else
      if(!ivec_cmp(iv,ivec_turn_right(pn->iv)))
        /* OK turn right */
        _request_task(pn,pn->ta,&pn_task_turn_right);
      else
      if(!ivec_cmp(iv,ivec_turn_back(pn->iv)))
      {
        /* OK turn back */
        _request_task(pn,pn->ta,&pn_task_turn_right);
        _request_task(pn,pn->ta,&pn_task_turn_right);
      }
      else
      {
        /** OK walk this way */
        _request_task(pn,pn->ta,&pn_task_walk);
      }
    }
/*****
    else
      fprintf(logfile," robot on target\n");
******/
  }
}

void
person_stop(struct person*pn)
{
  assert(pn);
  assert(pn->is_hero);
  _request_task(pn,pn->ta,&pn_task_stop);
}

void
person_walk_forward(struct person*pn)
{
  assert(pn);
  assert(pn->is_hero);
  _request_task(pn,pn->ta,&pn_task_walk);
}

void
person_walk_back(struct person*pn)
{
  assert(pn);
  assert(pn->is_hero);
  _request_task(pn,pn->ta,&pn_task_turn_back);
/*********
  _request_task(pn,pn->ta,&pn_task_turn_left);
  _request_task(pn,pn->ta,&pn_task_turn_left);
*********/
}

void
person_walk_left(struct person*pn)
{
  assert(pn);
  assert(pn->is_hero);
  _request_task(pn,pn->ta,&pn_task_turn_left);
}

void
person_walk_right(struct person*pn)
{
  assert(pn);
  assert(pn->is_hero);
  _request_task(pn,pn->ta,&pn_task_turn_right);
}

void
person_dig(struct person*pn)
{
  assert(pn);
  assert(pn->is_hero);
  _request_task(pn,pn->ta,&pn_task_dig);
}

struct person*people_get_hero(struct people*pl)
{ assert(pl); assert(pl->pn[0]); return pl->pn[0]; }

void
people_set_hero(struct people*pl,struct ivec iv )
{
  int i;
  assert(pl);
  for(i=1;i<MAX_PERSON;i++)
    if(pl->pn[i])
      if( ivec_xcmp(iv,pl->pn[i]->iv)==0 )
      {
        /** can't do this to the hero! */
        fprintf(logfile,
          "user error: can't put the hero on a robot.\n");
        return;
      }
  if(pl->pn[0])
  {
    /*** already have a hero ****/
    if( ivec_xcmp(iv,pl->pn[0]->iv)==0 )
      /** toggle hero off */
      _del_person(pl,pl->pn[0]);
    else
    {
      /*** move hero */
      float x[3*6];
      pl->pn[0]->iv= iv;
      __init_pn_state(pl->pn[0]);
      __make_pn_coords(pl->pn[0],x);
      scene_update_draw_item(
        pl->pn[0]->s,pl->pn[0]->d,x,
         colour_lookup(HERO_CID,0),render_person);
    }
  }
  else
  {
    /*** make the hero !  */
    pl->pn[0]=(struct person*)xmalloc(sizeof(struct person));
/*
    pl->pn[0]->offset_scale=HERO_SCALE;
*/
    _init_pn( pl->pn[0], pl->s, pl->m, pl, iv, 1, 
      colour_lookup(HERO_CID,0),mtime_get());
  }
  main_set_people();
}

int
people_toggle_robot(struct people*pl,struct ivec iv)
{
  int i;
  assert(pl);
  if(pl->pn[0])
    if(ivec_cmp(iv,pl->pn[0]->iv) == 0 )
    {
      fprintf(logfile,"user error: can't put a robot on the hero!\n");
      return 0;
    }
  for(i=1;i<MAX_PERSON;i++)
    if(pl->pn[i])
      if(ivec_xcmp(iv,pl->pn[i]->iv) == 0 )
      {
        _del_person(pl,pl->pn[i]);
        pl->pn[i]=NULL;
        main_set_people();
        return 1;
      }
  for(i=1;i<MAX_PERSON;i++)
    if(!pl->pn[i])
    {
      /** make the robot */
      pl->pn[i]=(struct person*)xmalloc(sizeof(struct person));
/*
      pl->pn[i]->offset_scale=ROBOT_SCALE;
*/
      _init_pn(pl->pn[i],pl->s,pl->m,pl,iv,0,
        colour_lookup(ROBOT_CID,0),mtime_get());
      main_set_people();
      return 1;
    }
  fprintf(logfile,"user error: Too many robots!\n");
  return 0;
}

void
people_set_robot_portal(struct people*pl,struct ivec iv)
{
#ifdef PEOPLE_DEBUG
  fprintf(logfile,"people_set_robot_portal\n");
#endif
  if(!pl->iv)
  {
    pl->iv=(struct ivec*)xmalloc(sizeof(struct ivec));
  }
  *pl->iv=iv;
  /* robot portal has no draw item */
}

void
people_finalize(struct people*pl)
{
  int idx;
  assert(pl);
  for(idx=0;idx<MAX_PERSON;idx++)
    if(pl->pn[idx])
    {
#ifdef DEBUG_TASK
      fprintf(logfile,"del person %p\n",pl->pn[idx]);
#endif
      fflush(logfile);
      _del_person(pl,pl->pn[idx]);
      pl->pn[idx]=NULL;
    }
  if( pl->iv )
    xfree(pl->iv);
  pl->iv=NULL;
}

