#ifndef __PEOPLE_H
#define __PEOPLE_H
#include <unistd.h>
#include "ivec.h"

#define MAX_PERSON 16

struct scene;

#define PN_POINTS 6

#define PN_TOFF_STOP		1
#define PN_TOFF_WALK 		6
#define PN_TOFF_FLIP 		3
#define PN_TOFF_TURN		3
#define PN_TOFF_DIG 		7
#define PN_TOFF_FALL 		4
#define PN_TOFF_STUCK		25
#define PN_TOFF_CLIMB		5

struct person;
struct task;

struct task_info
{
  int toff; 
  /** consumes zero time */
  void (*init)(struct person*); 

  /** consumes 1 unit of time, 
      returns 0 when nothing was done */
  int (*send)(struct person*);

  /** consumes zero time */
  void (*finalize)(struct person*);

  /** do we need to ? */
  int (*request)(
    struct person*,struct task*,struct task_info*);

  char *name; /** for debugging */
};

#define PN_TASK_STOP		0
#define PN_TASK_WALK 		1
#define PN_TASK_FLIP		2
#define PN_TASK_TURN_LEFT	3
#define PN_TASK_TURN_RIGHT	4
#define PN_TASK_DIG 		5 /** only for the hero */
#define PN_TASK_FALL		6
#define PN_TASK_STUCK   	7 /** only for a robot in a hole */
#define PN_TASK_CLIMB		8 /** for robot climbing out of hole */

/** these are for task.stat below */
#define PN_STAT_ALLOC		0  /** ready for init */
#define PN_STAT_INIT		1  /** ready for send */
#define PN_STAT_SEND		2  /** has been sent */
#define PN_STAT_FINALIZE	3  /** ready for free */

struct task
{
  int stat;
  struct task_info*ti;
  /** linked list */
  struct task*next;
};

#define ROBOT_LAG		(1.3)

struct lock
{
  int i,j,k;
  struct lock*next;
};

struct person
{
  struct scene*s;
  struct map*m;
  /** need to know about the others... */
  struct people*pl;
  /** true for the hero only */
  int is_hero;
  /** how many parcels do we have ? */
  int n_parcels;

  /*********** state information ************************/
  /** task buffer **/
  struct task*ta; /** sentinel */
  /** last valid position, set by "finalize" function */
  struct ivec iv; 
  /** task specific; set by "init" function, sometimes undefined */
  struct ivec _iv;
 
  /** a parameter for interpolating between iv and _iv */
  /** is 0 to max_offset-1 */
  int toff; 

  /** what time this person is at */
  int msec;

  /** delete me */
  int free;

  /** for robots only */
  struct lock*lk; /** locks we own */

  /*********** graphics information **********************/
  /** two legs and a trunk */
#define N_FRAMES 12
  float* frame; 
  struct draw_item*d; 
  /** a flag; the need to call update_draw_item */
  int redraw;
};

struct people
{
  struct scene*s;
  struct map*m;
  /** the hero is always at pn[0] */
  struct person*pn[MAX_PERSON];
  /** the robot portal */
  struct ivec*iv;
};

/*******************************************************/
void
pn_init_default(struct person*pn);
int
pn_send_default(struct person*pn);
int
pn_request_default(struct person*pn,struct task*,struct task_info*);
void
pn_end_default(struct person*pn);
void
pn_init_stop(struct person*pn);
int
pn_send_stop(struct person*pn);
void
pn_finalize_stop(struct person*pn);
int
pn_request_stop(struct person*pn,struct task*,struct task_info*);
void
pn_init_walk(struct person*pn);
int
pn_send_walk(struct person*pn);
void
pn_finalize_walk(struct person*pn);
int
pn_request_walk(
  struct person*pn,
  /** requestee */
  struct task*ta,
  /** request name */
  struct task_info*ti );
void
pn_init_flip(struct person*pn);
int
pn_send_flip(struct person*pn);
void
pn_finalize_flip(struct person*pn);
void
pn_init_turn_left(struct person*pn);
int
pn_send_turn_left(struct person*pn);
void
pn_finalize_turn_left(struct person*pn);
void
pn_init_turn_right(struct person*pn);
int
pn_send_turn_right(struct person*pn);
void
pn_finalize_turn_right(struct person*pn);
void
pn_init_turn_back(struct person*pn);
int
pn_send_turn_back(struct person*pn);
void
pn_finalize_turn_back(struct person*pn);
void
pn_init_dig(struct person*pn);
int
pn_send_dig(struct person*pn);
void
pn_finalize_dig(struct person*pn);
void
pn_init_fall(struct person*pn);
int
pn_send_fall(struct person*pn);
void
pn_finalize_fall(struct person*pn);
void
pn_init_stuck(struct person*pn);
int
pn_send_stuck(struct person*pn);
void
pn_finalize_stuck(struct person*pn);
void
pn_init_climb(struct person*pn);
int
pn_send_climb(struct person*pn);
void
pn_finalize_climb(struct person*pn);
/*******************************************************/

float*
pn_get_target(struct people*);
void
people_set_robot_portal(struct people*,struct ivec);
int
update_people_draw(struct people*pl);

struct person*people_get_hero(struct people*pl);

void
update_people_state(struct people*pl,int msec);

void
people_init(struct people*pl,struct scene*s,struct map*m);

void
people_finalize(struct people*);

void
people_set_hero(struct people*pl,struct ivec);

int
people_toggle_robot(struct people*,struct ivec);

void
person_dig(struct person*pn);

void
person_walk_right(struct person*pn);

void
person_walk_left(struct person*pn);

void
person_walk_back(struct person*pn);

void
person_walk_forward(struct person*pn);

void
person_stop(struct person*pn);

#endif  /* __PEOPLE_H */
