#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>

#include "render.h"
#include "memory.h"
#include "algebra.h"
#include "globals.h"
#include "colour.h"
#include "draw.h"
#include "mtime.h"

#define USE_ANTI_ALIASING

static int frames=0;
static int msec;
static float x_0,y_0,sx,sy;
static int status_height=20;

#define MAX(a,b) ((a)>(b)?(a):(b))
#define SGN(x) ((x)>0?1:((x)==0?0:(-1)))
#define ABS(x) ((x)>0?(x):(-x))

void text_init();
void text_banner(char*message);
void text_status(char*,char*,char*,int);
void text_put(char*,int,int,int,int,int,int);
void text_finalize();

void
render_init()
{
  draw_init();
  colour_init();
  text_init();

  x_0=width/2;
  y_0=height/2;
  sx=width/10;
  sy=-height/10;
  msec=mtime_get();
  frames=0;
/*
  fprintf(logfile," render: %f %f %f %f \n",x_0,y_0,sx,sy );
*/
}

void render_finalize()
{
  /** how many frames per second ? */
  int _msec;
  _msec=mtime_get();
  printf("render_finalize(): %.1f fps\n", 
    ((float)(frames)*1000.0)/_msec);

  text_finalize();
  colour_finalize();
  draw_finalize();
}

static
int 
bounded(float x,float y)
{
  return( (int)x>=0
    &&(int)y>=0
    &&(int)x<width
    &&(int)y<height-status_height);
}

static
void 
_scale(float*x,int nx)
{
  int i;
  for(i=0;i<nx;i++)
  {
    x[i*2]*=sx;
    x[i*2+1]*=sy;
    x[i*2]+=x_0;
    x[i*2+1]+=y_0;
  }
}

#define SWP(a,b,i)  (i)=(a);(a)=(b);(b)=(i);
#define _HLINE(x0,y,x1,c) draw_hline(x0,y,x1,c)

void
render_quad(float*_x,int c)
{
  int x[8]; /** 4 points; numbered 0 to 3 below **/
  int i,j;
  int i0, i1; /** start, end of scanline **/
  int di0, dj0, di1, dj1;
  int _i0, _i1; /** remainders */
#ifdef COMMENT
  char*comment;
#endif
  assert(_x);
  /** count 0 to 3 */
  for(i=0;i<4;i++)
  {
    x[2*i]=_x[2*i]*sx+x_0;
    x[2*i+1]=_x[2*i+1]*sy+y_0;
  }
  /** 0 and 3 are opposite */
  /** 1 and 2 are opposite */

  if(x[0]<0)return;
  if(x[2]<0)return;
  if(x[4]<0)return;
  if(x[6]<0)return;
  if(x[1]<0)return;
  if(x[3]<0)return;
  if(x[5]<0)return;
  if(x[7]<0)return;
  if(x[0]>=width)return;
  if(x[2]>=width)return;
  if(x[4]>=width)return;
  if(x[6]>=width)return;
  if(x[1]>=height-status_height)return;
  if(x[3]>=height-status_height)return;
  if(x[5]>=height-status_height)return;
  if(x[7]>=height-status_height)return;
  
  if( x[7]>x[3] )
  {
    /** swap 1 and 3 */
    SWP(x[2],x[6],i); 
    SWP(x[3],x[7],i);
    /** swap 0 and 2 */
    SWP(x[0],x[4],i);
    SWP(x[1],x[5],i);
  }
  /******** 1 is above 3 ********/
  if( x[1]>x[3] )
  {
    /** swap 0 and 1 */
    SWP(x[0],x[2],i);
    SWP(x[1],x[3],i);
    /** swap 2 and 3 */
    SWP(x[4],x[6],i);
    SWP(x[5],x[7],i);
  }
  /******** 1 is above 0 ********/
  /** take cross product: (3-1)X(0-1) */
  if((x[2]-x[0])*(x[3]-x[7]) < (x[3]-x[1])*(x[2]-x[6]))
  {
    SWP(x[0],x[6],i);
    SWP(x[1],x[7],i);
  }
  /** 0 is "left" of 3 */
#ifdef DEBUG_RENDER
  for(i=0;i<4;i++)
    fprintf(logfile,"(%d,%d)",x[2*i]-150,x[2*i+1]-150);
  fprintf(logfile,"\n");
#endif
  di0=x[2]-x[0];
  dj0=x[3]-x[1];
  di1=x[6]-x[2];
  dj1=x[7]-x[3];
  if( x[1]>x[7] )
  {
    /** 0 is above 3 */
#ifdef COMMENT
    comment="0 is above 3";
#endif
    i0=x[2];
    i1=x[2];
    _i0=0;
    _i1=0;
    j=x[3];
    if( dj0 && dj1 )
    {
      for(;j>x[1];j--)
      {
        _HLINE(i0,j,i1,c);
        i0-=(di0+_i0)/dj0;
        _i0=(di0+_i0)%dj0;  /** left over */
        i1-=(di1+_i1)/dj1;
        _i1=(di1+_i1)%dj1;  /** left over */
      }
    }
    i0=x[0];
    di0=x[4]-x[0];
    dj0=x[5]-x[1];
    if( dj0 && dj1 )
      for(;j>x[7];j--)
      {
        _HLINE(i0,j,i1,c);
        i0-=(di0+_i0)/dj0;
        _i0=(di0+_i0)%dj0;  /** left over */
        i1-=(di1+_i1)/dj1;
        _i1=(di1+_i1)%dj1;  /** left over */
      }
    i1=x[6]; /** necessary? */
    di1=x[4]-x[6];
    dj1=x[5]-x[7];
    if(dj0&&dj1)
      for(;j>x[5];j--)
      {
        _HLINE(i0,j,i1,c);
        i0-=(di0+_i0)/dj0;
        _i0=(di0+_i0)%dj0;  /** left over */
        i1-=(di1+_i1)/dj1;
        _i1=(di1+_i1)%dj1;  /** left over */
      }

  }
  else /** x[1] <= x[7] */
  {
    /** 3 is above 0 */  
#ifdef COMMENT
    comment="3 is above 0";
#endif
    i0=x[2];
    i1=x[2];
    _i0=0;
    _i1=0;
    if( dj0 && dj1 )
      for(j=x[3];j>x[7];j--)
      {
        _HLINE(i0,j,i1,c);
        i0-=(di0+_i0)/dj0;
        _i0=(di0+_i0)%dj0;  /** left over */
        i1-=(di1+_i1)/dj1;
        _i1=(di1+_i1)%dj1;  /** left over */
      }
    i1=x[6];
    di1=x[4]-x[6];
    dj1=x[5]-x[7];
    if(dj0&&dj1)
      for(j=x[7];j>x[1];j--)
      {
        _HLINE(i0,j,i1,c);
        i0-=(di0+_i0)/dj0;
        _i0=(di0+_i0)%dj0;  /** left over */
        i1-=(di1+_i1)/dj1;
        _i1=(di1+_i1)%dj1;  /** left over */
      }
    i0=x[0]; /** necessarry ? */
    di0=x[4]-x[0];
    dj0=x[5]-x[1];
    if( dj0 && dj1 )
      for(j=x[1];j>x[5];j--)
      {
        _HLINE(i0,j,i1,c);
        i0-=(di0+_i0)/dj0;
        _i0=(di0+_i0)%dj0;  /** left over */
        i1-=(di1+_i1)/dj1;
        _i1=(di1+_i1)%dj1;  /** left over */
      }
  }
}

#ifdef USE_ANTI_ALIASING
#define _LINE(i,j) \
   draw_wu_line(x[2*(i)],x[2*(i)+1],x[2*(j)],x[2*(j)+1],c,0,256,8)
#else
#define _LINE(i,j) \
   draw_line(x[2*(i)],x[2*(i)+1],x[2*(j)],x[2*(j)+1],c)
#endif

void
render_line(float *_x,int c)
{
  float x[4];
  assert(_x);
  memcpy(x,_x,sizeof(float)*4);
  _scale(x,2);
#ifdef DEBUG_RENDER
  fprintf(logfile,"render_line: ");
  print_x(x,2,2);
#endif
  if(bounded(x[0],x[1])&&bounded(x[2],x[3]))
  {
#ifdef DRAW
    _LINE(0,1);
#endif
  }
}

void
render_wire_quad(float *_x,int c)
{
  float x[8];
  assert(_x);

#ifdef PRINT_STUFF  
  int i;
  for(i=0;i<4;i++)
    fprintf(logfile,"(%.2f,%.2f)",_x[2*i],_x[2*i+1]);
  fprintf(logfile,"\n");
#endif

  memcpy(x,_x,sizeof(float)*8);
  _scale(x,4);
  if( bounded(x[0],x[1]) && bounded(x[2],x[3])
    && bounded(x[4],x[5]) && bounded(x[6],x[7]) )
  {
#ifdef DRAW
    _LINE(0,1);
    _LINE(0,2);
    _LINE(2,3);
    _LINE(1,3);
    return;
#endif
  }
}

void
render_dalek(float*_x,int c)
{
  int i;
  float x[20];
  float *y,*z;
  assert(_x);
  y=x+16;
  z=x+18;
  memcpy(x,_x,16*sizeof(float));
  _scale(x,8);
  for(i=0;i<8;i++)
    if( !bounded(x[2*i],x[2*i+1]))
      return;
#ifdef DRAW
  _LINE(0,1);
  _LINE(0,4);
  _LINE(1,5);
  _LINE(4,5);
  x[16]=(x[4]+x[6]+x[12]+x[14])/4.0;
  x[17]=(x[5]+x[7]+x[13]+x[15])/4.0;
  _LINE(0,8);
  _LINE(1,8);
  _LINE(4,8);
  _LINE(5,8);
  x[18]=(x[0]+x[2]+x[8]+x[10])/4.0;
  x[19]=(x[1]+x[3]+x[9]+x[11])/4.0;
  x[16]=(x[16]+x[18])/2.0;
  x[17]=(x[17]+x[19])/2.0;
  x[18]=(x[8]+x[12]+x[14]+x[10])/4.0;
  x[19]=(x[9]+x[13]+x[15]+x[11])/4.0;
  _LINE(8,9);
#endif
}

void
render_box(float*_x,int c)
{
  int i;
  float x[16];
  assert(_x);
  memcpy(x,_x,16*sizeof(float));
  _scale(x,8);
  for(i=0;i<8;i++)
    if( !bounded(x[2*i],x[2*i+1]))
      return;
#ifdef DRAW
  _LINE(0,1);
  _LINE(0,2);
  _LINE(2,3);
  _LINE(1,3);
  _LINE(0,4);
  _LINE(1,5);
  _LINE(3,7);
  _LINE(4,6);
  _LINE(4,5);
  _LINE(6,7);
  _LINE(5,7);
  _LINE(2,6);
#endif
}

void
render_person(float*_x,int c)
{
  float x[12]; 
  int i;
  assert(_x);
  memcpy(x,_x,sizeof(float)*12);
  _scale(x,6);
  for(i=0;i<6;i++)
    if(!bounded(x[2*i],x[2*i+1]))
      return;
#ifdef PRINT_STUFF
  fprintf(logfile,"render_person:");
  print_x(_x,6,2);
  print_x(x,6,2);
  fprintf(logfile,"\n");
#endif
#ifdef DRAW
  _LINE(0,1);
  _LINE(1,2);
  _LINE(2,3);
  _LINE(1,4);
  _LINE(4,5);
#endif
}

void
render_parcel(float*_x,int c)
{
  float x[40];
  float *y=x+16; /* mid points */
  int i;
  assert(_x);
  memcpy(x,_x,sizeof(float)*16);
  _scale(x,8);
  for(i=0;i<8;i++)
    if(!bounded(x[2*i],x[2*i+1]))
      return;
  for(i=0;i<24;i++)
    y[i]=0;
  for(i=0;i<4;i++)
  {
    int j;
    if(i<2) j=i; else j=i+2;
    y[2*i   ]=x[4*i  ]+x[4*i+2]; 
    y[2*i+1 ]=x[4*i+1]+x[4*i+3];
    y[2*i+8 ]=x[2*i  ]+x[2*i+8];
    y[2*i+9 ]=x[2*i+1]+x[2*i+9];
    y[2*i+16]=x[2*j  ]+x[2*j+4]; 
    y[2*i+17]=x[2*j+1]+x[2*j+5];
  }
  for(i=0;i<24;i++)
    y[i]/=2.0;
#ifdef DRAW
  _LINE(0,1);
  _LINE(4,5);
  _LINE(2,3);
  _LINE(6,7);
  _LINE(12,13);
  _LINE(14,15);
  _LINE(16,17);
  _LINE(18,19);
  _LINE(0,2);
  _LINE(4,6);
  _LINE(1,3);
  _LINE(5,7);
  _LINE(8,9); 
  _LINE(10,11);
  _LINE(12,14);
  _LINE(13,15);
  _LINE(0,4);
  _LINE(2,6);
  _LINE(1,5);
  _LINE(3,7);
  _LINE(8,10); 
  _LINE(9,11); 
  _LINE(16,18);
  _LINE(17,19);
#endif
}

void
render_instructions(char**lines,int n_lines,int n)
{
  int i;
  int dy=height/n_lines;
  int y=0;
  render_clear_all();
  /*printf("\n");*/
  for(i=0;i<n_lines;i++)
  {
    char*line=lines[(i+n)%n_lines];
    int len;
    len=strlen(line);
    if(len)
    {
      if(line[len-1]=='\n')
        line[len-1]=0;
      /*printf("%s\n",line);*/
      text_put(line,width/10,y,width*9/10,y+dy/2,_white,0);
    }
    y+=dy;
  }
  render_update();
}


void
render_banner(char*text)
{
#ifdef DRAW
  text_banner(text);
  render_update();
#endif
  fprintf(logfile,"render_banner(%s)\n",text);
}

void
render_status(char*left,char*center,char*right)
{
  text_status(left,center,right,status_height);
}

void
render_clear(float*_x)
{
  float x[8];
  memcpy(x,_x,sizeof(float)*8);
  _scale(x,4);
  if(x[0]>x[2]) {float tmp=x[0];x[0]=x[2];x[2]=tmp;}
  if(x[1]>x[3]) {float tmp=x[3];x[3]=x[1];x[1]=tmp;}
  if(x[0]<0.0) x[0]=0;
  if(x[1]<0.0) x[1]=0;
  if(x[2]<0.0) x[2]=0;
  if(x[3]<0.0) x[3]=0;
  if(x[0]>width-1) x[0]=width-1;
  if(x[1]>height-1) x[1]=height-1;
  if(x[2]>width-1) x[2]=width-1;
  if(x[3]>height-1) x[3]=height-1;
#ifdef DRAW
  #ifdef COLOUR_REDRAW
    draw_fillbox(
      x[0],x[1],
      ceil(1+x[2]-x[0]),ceil(1+x[3]-x[1]),rand());
  #else
    draw_fillbox(
      x[0],x[1],
      ceil(1+x[2]-x[0]),ceil(1+x[3]-x[1]),_black);
  #endif
#endif
}

void
render_colours()
{
  int i, j, di=width/16, dj=height/16;
#ifdef DRAW
  for(i=0;i<16;i++)
    for(j=0;j<16;j++)
      draw_fillbox(i*di,j*dj,di-1,dj-1,i+j*16);
#endif
}

void
render_clear_all()
{
#ifdef DRAW
  draw_fillbox(0,0,width-1,height-status_height,_black);
#endif
}

void render_update()
{
  draw_update();
  frames++;
}



