#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include "algebra.h"
#include "cam.h"
#include "render.h"
#include "memory.h"
#include "globals.h"
#include "scene.h"
#include "bsp.h"

void
scene_print_draw_list(struct dlink*_dl)
{
  struct dlink*dl;
  fprintf(logfile,"<");
  for(dl=_dl;dl->next;dl=dl->next)
    fprintf(logfile,"%.2f,",dl->next->d->im->depth);
  fprintf(logfile,">\n");
}

void
scene_init(struct scene*s,struct map*m,struct cam*c)
{
  s->c=*c;
  s->p=bsp_alloc_init(0,m->bi/2);
  s->dl=dlink_alloc_init();
  s->m=m;
  s->_redraw_me=1;
}

void
_old_add_draw_item(struct dlink*dl,struct draw_item*d)
{
  /****** insert in descending order ******/
  for(;dl->next;dl=dl->next)
    if(dl->next->d->im->depth < d->im->depth )
      break;
  dlink_push(dl,d);
}

void
_add_draw_item(struct dlink*dl,struct draw_item*d)
{
  dlink_push(dl,d);
}

void
scene_invalidate(struct scene*s)
{
  s->_redraw_me=1;
}

int
_remove_draw_item(struct dlink*dl,struct draw_item*d)
{
  for(;dl->next;dl=dl->next)
    if(dl->next->d==d)
    {
      dlink_pop(dl);
      return 1;
    }
  return 0;
}

void
_init_source_item(
  struct source_item*so,
  float*x,
  int nx,
  int c,
  void (*render)(float*,int))
{
  so->x=(float*)xmalloc(sizeof(float)*nx*3);
  memcpy(so->x,x,nx*3*sizeof(float));
  so->c=c;
  so->render=render;
}

/** public */
struct draw_item*
scene_new_draw_item(
  struct scene*s,
  float*x,
  int nx,
  int c,
  void (*render)(float*,int),
  int dynamic )
{
  /* float x[3*4] */
  struct draw_item*d;

#define _DEBUG_SCENE
#ifdef DEBUG_SCENE
  fprintf(logfile,"new_draw_item:\n");
  print_x(x,nx,3);
  fflush(logfile);
#endif

  assert(nx>1);
  d=(struct draw_item*)xmalloc(sizeof(struct draw_item));
  d->nx=nx;
  d->del_me=0;
  d->so=(struct source_item*)xmalloc(sizeof(struct source_item));
  _init_source_item(d->so,x,nx,c,render);
  d->im=(struct image_item*)xmalloc(sizeof(struct image_item));
  d->im->x=(float*)xmalloc(sizeof(float)*nx*2);
  d->im->depth=0.0; /* a depth coordinate */
  cam_image_source_item(&s->c,d->im,d->so,d->nx);
  scene_invalidate(s);
  _add_draw_item(s->dl,d);
  bsp_insert_draw_item(s->p,d);

  return d;
}

/** public */
void
scene_update_draw_item(
  struct scene*s,
  struct draw_item*d,
  float*x,
  int c,
  void (*render)(float*,int))
{
  /* item has changed ! */
#ifdef DEBUG_SCENE
  fprintf(logfile,"update;;;;;;;;;;;;;\n");
  scene_print_draw_list(s->dl);
#endif

  if(!_remove_draw_item(s->dl,d)) 
    assert(0);
  if(!bsp_remove_draw_item(s->p,d))
    assert(0);
  d->so->c=c;
  scene_invalidate(s);
  memcpy(d->so->x,x,3*d->nx*sizeof(float));
  d->so->render=render;
  cam_image_source_item(&s->c,d->im,d->so,d->nx);
  _add_draw_item(s->dl,d);
  bsp_insert_draw_item(s->p,d);
}

void
scene_set_draw_item_colour(
  struct scene*s,
  struct draw_item*d,
  int c )
{
  assert(s);
  assert(d);
  d->so->c=c;
  scene_invalidate(s);
}

void
scene_set_draw_item_skip(
  struct scene*s,struct draw_item*d,int skip)
{
  assert(d);
  assert(s);
  assert(d->im);
  d->im->skip=skip;
  scene_invalidate(s);
}

void
scene_update_cam(struct scene*s,struct cam*c)
{
  struct dlink*dl;
  s->c=*c;
  for(dl=s->dl;dl->next;dl=dl->next)
  {
    assert(dl->next->d->im);
    cam_image_source_item(c,dl->next->d->im,dl->next->d->so,dl->next->d->nx);
  }
  map_cull_backfaces(s->m,c);
  scene_invalidate(s);
}

static int _clean=0, _no_clean=0;
int
scene_draw(struct scene*s)
{
  if(!s->_redraw_me)
  {
    _no_clean++;
    /**return 0;**/
  }
  else
    _clean++;
  s->_redraw_me=0;
  render_clear_all();

  bsp_draw_scene(s->p,&s->c);
  return 1;
}

void
_del_source_item(struct source_item*so)
{
  xfree(so->x);
  xfree(so);
}

void
_del_image_item(struct image_item*im)
{
  xfree(im->x);
  xfree(im);
}

void
_del_draw_item(struct draw_item*d)
{
#ifdef DEBUG_SCENE
  fprintf(logfile,"del item\n");
#endif
  _del_source_item(d->so);
  if(d->im)
    _del_image_item(d->im);
  xfree(d);
}

void
scene_del_draw_item(struct scene*s,struct draw_item*d)
{
  if(!_remove_draw_item(s->dl,d))
    assert(0);
  if(!bsp_remove_draw_item(s->p,d))
    assert(0);
  scene_invalidate(s);
  _del_draw_item(d);
}  

/* private */
void
_del_draw_items(struct scene*s)
{
  while(s->dl->next)
    _del_draw_item(dlink_pop(s->dl));
}

void
scene_finalize(struct scene*s)
{
#ifdef DEBUG_SCENE
  fprintf(logfile,"finalize_scene()\n");
#endif
  _del_draw_items(s);
#ifdef DEBUG_SCENE
  fprintf(logfile,"draw items deleted.\n");
#endif
  dlink_free(s->dl);
  bsp_finalize_free(s->p);
  printf("scene_finalize(): redraw=%d, idle=%d\n",
    _clean, _no_clean );
}
