#ifndef __SCENE_H
#define __SCENE_H
#include <stdio.h>
#include "map.h"
#include "cursor.h"
#include "globals.h"
#include "cam.h"
#include "dlink.h"
#include "people.h"

/*** viewpoint specific details ***/
struct image_item
{
  float *x; /* image coordinates */
  float depth; /* for back to front draw ordering */
  int skip; /* don't draw me */
};

/**** viewpoint independant details *****/
struct source_item
{
  float *x; /* coordinates */
  int c; /* colour */
  void (*render)(float*ix,int c);
};

#define IS_DYNAMIC 1
#define IS_STATIC 0

struct draw_item
{
  /*** the inhabitants of our scene *****/
  struct source_item *so;
  struct image_item *im; /** the present image */
  int nx; /* number of points */
  int del_me; /** flag */
};

struct scene
{
  struct cam c;
  struct bsp*p;
  /** list of all draw_item(s) */
  struct dlink *dl; 
  int _redraw_me;

  struct map*m; /** map knows things we don't */
};

void
scene_init(struct scene*s,struct map*,struct cam*);

struct draw_item*
scene_new_draw_item(
  struct scene*s,
  float*x,
  int nx,
  int c,
  void (*render)(float*,int),
  int dynamic);

/** item has changed */
void
scene_update_draw_item(
  struct scene*s,
  struct draw_item*d,
  float*x,
  int c,
  void (*render)(float*,int));

/** cam has changed */
void
scene_set_draw_item_colour(
  struct scene*s,
  struct draw_item*d,
  int c );

/** cam has changed */
void
scene_update_cam(
  struct scene*s,
  struct cam*c); 

void
scene_set_draw_item_skip(
  struct scene*s,struct draw_item*d,int skip);

int 
scene_poll_draw(
  struct scene*s,
  int t /* max number of items to draw */ );

void
scene_build(struct scene*s);

int
scene_draw(struct scene*s);

void
scene_del_draw_item(struct scene*s,struct draw_item*d);

void
scene_finalize(struct scene*s);

#endif



