#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>

#include <SDL/SDL.h>

#include "draw.h"
#include "memory.h"
#include "globals.h"

#define MAX(a,b) ((a)>(b)?(a):(b))
#define MIN(a,b) ((a)<(b)?(a):(b))
#define BND(a,b,c) MIN(MAX(a,b),c)

#define MAX_DAT 24
struct row
{
  int n;
  int dat[MAX_DAT]; /* a list of pairs; on/off columns */
};

struct char_data
{
  int w,h;
  struct row *code; /* length h */
  SDL_Surface*s;
};

static struct char_data font[128];

static
void
text_error(char*m)
{
  fprintf(stderr,"text_error(%s)\n",m);
  exit(1);
}

void
text_init()
{
  int i;
  for(i=0;i<128;i++)
  {
    font[i].s=NULL;
    font[i].code=NULL;
    font[i].w=0;
    font[i].h=0;
  }
}

#ifdef UNUSED
static
void
log_char(int a)
{
  int i,k;
  fprintf(logfile,"log_char %d\n", a);
  fprintf(logfile,"%d %d\n",font[a].w,font[a].h); 
  for(i=0;i<font[a].h;i++)
  {
    fprintf(logfile,"%d :",font[a].code[i].n);
    for(k=0;k<font[a].code[i].n;k++)
      fprintf(logfile,"%d ",font[a].code[i].dat[k]);
    fprintf(logfile,"\n");
  }
}
#endif

static int
load_char_data(int a)
{
  assert(0<=a&&a<128);
  if (font[a].code==NULL)
  {
    char filename[64];
    FILE*stream=NULL;
    char line[256];
    int i;
    strcpy(filename,"font-data/?.dat");
    filename[10]=a;
    stream=fopen(filename,"r");
    if(stream==NULL)
    {
      printf("cant open %s\n", filename);
      return 0;
    }
    fgets(line,256,stream);
    if( sscanf(line,"%d %d ",&font[a].w,&font[a].h) < 2 )
      assert(0);
    font[a].code=(struct row*)
      xmalloc(sizeof(struct row)*font[a].h);
#define _DEBUG_TEXT
#ifdef DEBUG_TEXT
    fprintf(logfile,"%s\n",filename);
    fprintf(logfile,"%d %d\n",font[a].w,font[a].h); 
#endif
    for(i=0;i<font[a].h;i++)
    {
      int j,k=0;
      int on,off;
      fgets(line,256,stream);
      for( j=0; 
        sscanf(line+k,"%d %d ", &on, &off)==2 && j<MAX_DAT; 
        j+=2 )
      {
        int l;
        for(l=2;l&&line[k];k++)
          if( line[k] == 32 )
          {
            while(line[k+1]&&line[k]==32) k++;
            l--;k--;
          }
        font[a].code[i].dat[j]=on;
        font[a].code[i].dat[j+1]=off;
      }
      assert(j<MAX_DAT);
      font[a].code[i].n=j;
#ifdef DEBUG_TEXT
      for(k=0;k<j;k++)
        fprintf(logfile,"%d ",font[a].code[i].dat[k]);
      if(k)
        fprintf(logfile,"\n");
#endif
    }
    fclose(stream);
  }
  return 1;
}

static
void
put_char(
  int a,
  int x, int y,
  float sx, float sy /*scale factor */,
  int fgc,
  int bgc, int bg_flag )
{
  int _y;
  assert(0<=a&&a<128);
  if(a==32) return;
  if(!load_char_data(a))
    text_error("can't find char data\n");

#define _DEBUG_TEXT
#ifdef DEBUG_TEXT
  printf("put_char(%c,%d,%d)\n",a,x,y);
  /*log_char(a);*/
#endif

  for(_y=0;_y<font[a].h;_y++)
  {
    struct row*r=&font[a].code[_y];
    int i,_x=0;
    for(i=0;i<r->n;i+=2)
    {
      _x+=r->dat[i]; /* on */
      if( bg_flag )
        draw_hline(
          BND(x+_x*sx-2, 0,width-1),
          BND(y+_y*sy, 0,height-1),
          BND(x+(_x+r->dat[i+1]-1)*sx+2, 0,width-1),
          bgc);
      draw_hline(
        BND(x+_x*sx, 0,width-1),
        BND(y+_y*sy, 0,height-1),
        BND(x+(_x+r->dat[i+1]-1)*sx, 0,width-1),
        fgc);
      _x+=r->dat[i+1]; /* off */
    }
  }
}

static int space_w=50;

static
void
get_char_size(int a,int *p_w,int *p_h)
{
  assert(0<=a&&a<128);
  if(a==32)
  {
    if(p_w!=NULL)*p_w=space_w;
    if(p_h!=NULL)*p_h=1;
  }
  else
  {
    if(!load_char_data(a))
      text_error("can't find char data\n");
    if(p_w!=NULL)*p_w=font[a].w;
    if(p_h!=NULL)*p_h=font[a].h;
  }
#ifdef DEBUG_TEXT
  printf("char_size %c %dx%d\n",(char)a,*p_w,*p_h);
#endif
}

void
get_text_size(char*text,int*p_w,int*p_h)
{
  char*p;
  int h=0,w=0;
  if(p_w!=NULL)*p_w=0;
  if(p_h!=NULL)*p_h=0;
  for(p=text;*p;p++)
  {
    get_char_size(*p,&w,&h);
    if(p_w!=NULL)*p_w+=w;
    if(p_h!=NULL)*p_h=MAX(h,*p_h);
  }
#ifdef DEBUG_TEXT
  printf("text_size %s %dx%d\n",text,*p_w,*p_h);
#endif
}

void
text_put(
  char*message,
  int x0,int y0,int x1,int y1,
  int colour,int bg_flag )
{
  char*p;
  int h,w;
  int x=0;
  float sx=1.0,sy=1.0;
#ifdef DEBUG_TEXT
  printf("text_box('%s')\n",message);
#endif
  get_text_size(message,&w,&h);
  x0=BND(x0,0,width-2);
  y0=BND(y0,0,height-2);
  x1=BND(x1,x0+1,width-1);
  y1=BND(y1,y0+1,height-1);
  if(w>x1-x0)
    sx=(float)(x1-x0)/(float)w;
  if(h>y1-y0)
    sy=(float)(y1-y0)/(float)h;
  for(p=message;*p;p++)
  {
    int _w,_h;
    get_char_size(*p,&_w,&_h);
    put_char(
      *p,
      x0+x*sx,y0+sy*(h-_h)/3,sx,sy,
      colour,_black,bg_flag);
    x+=_w;
  }
}

void
text_banner(char*message)
{
  int h,w;
  int y;
  int x;
#ifdef DEBUG_TEXT
  printf("text_banner('%s')\n",message);
#endif
  get_text_size(message,&w,&h);
  w=MIN(w,width-20);
  h=MIN(h,height-100);
  x=MAX((width-w)/2,10);
  y=MAX((height-h)/2,10);
/**
  draw_fillbox(0,y-4,width-1,y+h+4,_red);
  draw_fillbox(0,y-2,width-1,y+h+2,_black);
**/
  text_put(message,x,y,x+w,y+h,_red,1);
}

void
text_status(char*left,char*center,char*right,int h)
{
  /* show status line at bottom */
  int lw=1,lh=1,cw=1,ch=1,rw=1,rh=1;
  float sx=1.0,sy;
  if(left)get_text_size(left,&lw,&lh);
  if(center)get_text_size(center,&cw,&ch);
  if(right)get_text_size(right,&rw,&rh);
  
  if(2*MAX(lw,rw)+cw>=width)
    sx=(float)width/(float)(2*MAX(lw,rw)+cw);
  sy=(float)(h-2)/(float)MAX(lh,MAX(ch,rh));
  if( 2*sy<sx )
    sx=2*sy;
  draw_hline(0,height-h,width-1,_red);
  draw_hline(0,height-h+1,width-1,_red);
  draw_fillbox(0,height-h+2,width-1,height-1,_black);
  text_put(left,0,height-h+2,sx*lw,height-1,_red,0);
  text_put(center,
    (width-sx*cw)/2,height-h+2,
    (width+sx*cw)/2,height-1,_red,0);
  text_put(right,
    width-sx*rw-1,height-h+2,width-1,height-1,_red,0);
}

void
text_finalize()
{
  int i;
  for(i=0;i<128;i++)
  {
    if(font[i].s)
      SDL_FreeSurface(font[i].s);
    if(font[i].code)
      xfree(font[i].code);
  }
}


